/* Copyright (c) 2000-2001 ActiveState Tool Corporation.
   See the file LICENSE.txt for licensing information. */

/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

// The status bar API for Komodo.
//
// Here is an excerpt from the status bar requirements document:
//  --------------------------------------------------------------------------
//  |  <message>                  | <ck> | <enc.> | <pos.> | <lang.> | <ro> /,
//  --------------------------------------------------------------------------
//     ^--(a) Message                ^       ^        ^         ^       ^    ^
//        (b) Linter/Checker status -'       |        |         |       |    |
//        (c) Encoding ----------------------'        |         |       |    |
//        (d) Position (line and column) -------------'         |       |    |
//        (e) Language -----------------------------------------'       |    |
//        (f) Read-only ------------------------------------------------'    |
//        (z) Grippy --------------------------------------------------------'
//
// Implementation:
//
// The Status Bar observes various status notification messages (e.g.
// 'current_view_encoding_changed') via Mozilla's nsIObserver mechanism.
// Because some of the
// status bar segments are specific to the currently focused view ((b)-(f)),
// the preference is that notifications only be sent for changes to the current
// view. This is not always easy, so in some cases the logic in this module
// must ensure that notifications are for the current view.
//
// How to make a status message:
// Method 1.
//      If you are in JavaScript code you can just issue a
//      StatusBar_AddMessage(...) call. See that function for a description
//      of the arguments.
// Method 2.
//      You can send a 'status_message' notification via the nsIObserver
//      mechanism where the 'subject' argument of the Notify() is a
//      koIStatusMessage object. Here is an example in Python code:
//
//        observerSvc = components.classes["@mozilla.org/observer-service;1"]\
//                      .getService(components.interfaces.nsIObserverService)
//        sm = components.classes["@activestate.com/koStatusMessage;1"]\
//             .createInstance(components.interfaces.koIStatusMessage)
//        sm.category = "<some string>"
//        sm.msg = "<some string>"
//        sm.timeout = 0    # 0 for no timeout, else a number of milliseconds
//        sm.highlight = 0  # boolean, whether or not to highlight
//        try:
//            observerSvc.Notify(sm, 'status_message', None)
//        except COMException, e:
//            # do nothing: Notify sometimes raises an exception if (???)
//            # receivers are not registered?
//            pass
//
// TODO:
//  - double-click and right-click functionality as per the requirements doc
//  - read-only status icon
//  - better icons for showing check status (more obvious, the inprogress one
//    must at least be replaced)                                             


//---- globals

var gObserverSvc;
var gStatusBarObserver;
var gStatusBarPrefObserver;

var gStatusMessageStack = null;


//---- helper functions

function _StatusBar_UpdateEncoding(view) {
    var encoding = view.py_inst.encoding.short_encoding_name;
    var encodingWidget = document.getElementById('statusbar-encoding');
    encodingWidget.setAttribute("label", encoding);
}


function _StatusBar_ClearEncoding() {
    var encodingWidget = document.getElementById('statusbar-encoding');
    encodingWidget.removeAttribute("label");
}


function _StatusBar_UpdateLanguage(view) {
    language = view.py_inst.language;
    var languageWidget = document.getElementById('statusbar-language');
    languageWidget.setAttribute("label", language);
}


function _StatusBar_ClearLanguage() {
    var languageWidget = document.getElementById('statusbar-language');
    languageWidget.removeAttribute("label");
}

var _updateLineColTimer = null;
function _StatusBar_doUpdateLineCol(lineColText) {
    _updateLineColTimer = null;
    var lineColWidget = document.getElementById('statusbar-line-col');
    lineColWidget.setAttribute('label', lineColText);
}

function _StatusBar_UpdateLineCol(view) {
    if (!view || typeof(view)=="undefined") return;
    // Determine the current position.
    var scimoz = view._scimoz;
    var caretPos = scimoz.currentPos;
    var caretLine = scimoz.lineFromPosition(caretPos) + 1;
    var caretCol = scimoz.getColumn(caretPos) + 1;
    
    // Update line and column text.    
    var lineColText = "  Ln: " + caretLine + " Col: " + caretCol;    
    if (_updateLineColTimer) {
        clearTimeout(_updateLineColTimer);
    }
    setTimeout("_StatusBar_doUpdateLineCol('"+lineColText+"')",200);

    // If there are lint result messages for the current line then display
    // them.
    var lintResults = view.lintBuffer.lintResults;
    var lintError = view.lintBuffer.errorString;
    var lintMessage = null;
    if (!lintError && lintResults) {
        var resultsObj = new Object();
        var numResultsObj = new Object();
        lintResults.getResultsAtPosition(caretLine, caretCol, resultsObj,
                                         numResultsObj);
        var results = resultsObj.value;
        var numResults = numResultsObj.value;

        for (var i = 0; i < numResults; i++) {
            if (lintMessage) {
                lintMessage += "\n" + results[i].description;
            } else {
                lintMessage = results[i].description;
            }
        }
    }
    
    if (lintError) {
        StatusBar_AddMessage(lintError, "check", 0, true);
    } else if (lintMessage) {
        StatusBar_AddMessage(lintMessage, "check", 0, false);
    } else {
        StatusBar_AddMessage(null, "check", 0, false);
    }
}


function _StatusBar_ClearLineCol() {
    var lineColWidget = document.getElementById('statusbar-line-col');
    lineColWidget.removeAttribute("label");
    StatusBar_AddMessage(null, "check", 0, false);
}


function _StatusBar_ClearCheck() {
    var checkWidget = document.getElementById('statusbar-check');
    checkWidget.removeAttribute("src");
}


function _StatusBar_UpdateCheck(view) {
    // Update the status bar for the current check status.
    var checkWidget = document.getElementById('statusbar-check');

    // Is linting enabled?
    var checkingEnabled = view.prefs.getBooleanPref("editUseLinting");
    if (!checkingEnabled) {
        checkWidget.removeAttribute("src");
        checkWidget.setAttribute("tooltiptext", "Syntax Checker Status");
        return;        
    }

    // Is there an error in the linter?
    var lintError = view.lintBuffer.errorString;
    if (lintError) {
        checkWidget.setAttribute('src',
            "chrome://komodo/skin/images/icon_check_error.png");
        return;
    }

    // If so, get the lintResults.
    //   lintResults == null        -> inprogress
    //   lintResults.length == 0    -> ok
    //   lintResults.length > 0     -> errors/warnings
    var lintResults = view.lintBuffer.lintResults;
    if (!lintResults) {
        checkWidget.setAttribute('src',
            "chrome://komodo/skin/images/icon_check_inprogress.png");
        checkWidget.setAttribute("tooltiptext",
                                 "Syntax Checking Status: in progress");
    } else {
        var resultsObj = new Object();
        var numResultsObj = new Object();
        lintResults.getResults(resultsObj, numResultsObj);
        if (numResultsObj.value == 0) {
            checkWidget.setAttribute('src',
                "chrome://komodo/skin/images/icon_check_ok.png");
            checkWidget.setAttribute("tooltiptext",
                                     "Syntax Checking Status: ok");
        } else {
            checkWidget.setAttribute('src',
                "chrome://komodo/skin/images/icon_check_error.png");
            checkWidget.setAttribute("tooltiptext",
                "Syntax Checking Status: " +
                lintResults.getNumErrors() + " error(s), "+
                lintResults.getNumWarnings() + " warning(s)");
        }
    }
}

var _addMessageTimer = null;
function _StatusBar_AddMessageObject(sm)
{
    gStatusMessageStack.Push(sm);
    //dump("StatusBar: add message: current stack:\n");
    //gStatusMessageStack.Dump();
    
    // Get the latest message and show it in the UI.
    if (_addMessageTimer) {
        clearTimeout(_addMessageTimer);
    }
    _addMessageTimer = setTimeout("_StatusBar_UpdateMessage();", 200);
    // If this new message timesout then schedule an update for then as well.
    if (sm.timeout > 0) {
        // Allow for some inaccuracy in timeout scheduling to ensure that
        // the message has actually timed-out before updating.
        var epsilon = 300;
        setTimeout("_StatusBar_UpdateMessage();", sm.timeout+epsilon);
    }
}


function _StatusBar_UpdateMessage()
{
    _addMessageTimer = null;
    //dump("StatusBar: update message: current stack:\n");
    //gStatusMessageStack.Dump();

    // Get the latest message and show it in the UI.
    var sm = gStatusMessageStack.Top();
    var messageWidget = document.getElementById('statusbar-message');
    if (sm) {
        //dump("StatusBar: update message: top: msg='"+sm.msg+"', category='"+
        //     sm.category+"', timeout='"+sm.timeout+"', highlight='"+
        //     sm.highlight+"'\n");
        
        messageWidget.setAttribute("category", sm.category);
        messageWidget.setAttribute("label", sm.msg);
        messageWidget.setAttribute("tooltiptext", sm.msg);
        if (sm.highlight) {
            addStyle(messageWidget, "color", "white");
            addStyle(messageWidget, "background-color",
                     "#0a246a" /* windows title bar blue */);
        } else {
            //XXX How to just go back to defaults?
            addStyle(messageWidget, "color", "windowtext");
            addStyle(messageWidget, "background-color", "buttonface");
        }
    } else {
        messageWidget.removeAttribute("label");
        messageWidget.removeAttribute("tooltiptext");
        //XXX How to just go back to defaults?
        addStyle(messageWidget, "color", "windowtext");
        addStyle(messageWidget, "background-color", "buttonface");
    }
}



//---- local classes

function StatusBarObserver() {};
StatusBarObserver.prototype.Observe = function(subject, topic, data)
{
    // Unless otherwise specified the 'subject' and 'data' arguments are
    // expected to be empty for all notifications.
    //dump("StatusBar observed '"+topic+"': \n");
    var view = null;
    if (topic != 'status_message'
        && topic != 'scc_status_message'
       ) {
        // 'status_message' is the only field (currently) not requiring
        // a view.
        var vm = getFocusedViewManager();
        if (!vm) return;
        view = vm.getCurrentView();
        if (!view) return;
    }
    
    switch (topic) {
    case 'post_activate_view':
        _StatusBar_UpdateEncoding(view);
        _StatusBar_UpdateLanguage(view);
        _StatusBar_UpdateLineCol(view);
        _StatusBar_UpdateCheck(view);
        // register to receive changes in the current view pref set
        view.prefs.AddObserver(gStatusBarPrefObserver);
        break;
    case 'pre_deactivate_view':
        _StatusBar_ClearEncoding(view);
        _StatusBar_ClearLanguage(view);
        _StatusBar_ClearLineCol(view);
        _StatusBar_ClearCheck(view);
        // unregister observation of pref changes in the current view
        view.prefs.RemoveObserver(gStatusBarPrefObserver);
        break;
    case 'current_view_encoding_changed':
        _StatusBar_UpdateEncoding(view);
        break;
    case 'current_view_language_changed':
        _StatusBar_UpdateLanguage(view);
        break;
    case 'current_view_linecol_changed':
        _StatusBar_UpdateLineCol(view);
        break;
    case 'current_view_check_started':
        _StatusBar_UpdateCheck(view);
        break;
    case 'current_view_check_finished':
        _StatusBar_UpdateCheck(view);
        break;
    case 'scc_status_message':
        // fall through
    case 'status_message':
        // "subject" is expected to be a koIStatusMessage object.
        _StatusBar_AddMessageObject(subject);
        break;
    }
}


function StatusBarPrefObserver () {};
StatusBarPrefObserver.prototype.Observe = function(prefSet, prefSetID,
                                                   prefName)
{
    if (prefName == "editUseLinting") {
        //dump("StatusBar: observed pref '"
        //     + prefName + "' change (prefSet="
        //     + prefSet + ", prefSetID="
        //     + prefSetID + "\n");
        var view = getFocusedViewManager().getCurrentView();
        if (view) {
            _StatusBar_UpdateCheck(view);
        }
    }
};



//---- public functions

function StatusBar_AddMessage(msg, category, timeout, highlight)
{
    // Post a message to the status bar message area.
    // "msg" is the message string. An empty string or null indicates
    //      that the message (of the given category) should be cleared.
    // "category" is the message group to which the message belongs. It
    //      is an arbitrary string (it must be at least one character).
    // "timeout" is the amount of time, in milliseconds, that the message
    //      should appear. A value of 0 indicates that the message does
    //      not timeout.
    // "highlight" is a boolean indicating whether the message should be
    //      highlighted on the status bar.
    //
    // A structure similar to a stack of status messages is maintained.
    // The latest message is always shown. When/if it timesout then the
    // previous message is the stack is displayed. There can only be one
    // message per category, so reusing a category allows for removal of
    // status messages that are no longer appropriate.
    //
    // To add a message that does not timeout:
    //  StatusBar_AddMessage("hello there", "my_category", 0, false)
    // To remove that message:
    //  StatusBar_AddMessage(null, "my_category", 0, false)
    // To add a highlighted message for three seconds:
    //  StatusBar_AddMessage(null, "my_category", 3000, true)

    //dump("StatusBar: add message: msg='"+msg+"', category='"+category+
    //     "', timeout='"+timeout+"', highlight='"+highlight+"'\n");

    // create a status message component and insert it into the stack
    var sm = Components.classes["@activestate.com/koStatusMessage;1"].
          createInstance(Components.interfaces.koIStatusMessage);
    sm.msg = msg;
    sm.category = category;
    sm.timeout = timeout;
    sm.highlight = highlight;
    
    _StatusBar_AddMessageObject(sm);
}


function StatusBar_Init()
{
    gStatusBarObserver = new StatusBarObserver();
    gStatusBarPrefObserver = new StatusBarPrefObserver();

    gObserverSvc = Components.classes["@mozilla.org/observer-service;1"].
                       getService(Components.interfaces.nsIObserverService);
    gObserverSvc.AddObserver(gStatusBarObserver, 'post_activate_view');
    gObserverSvc.AddObserver(gStatusBarObserver, 'pre_deactivate_view');
    gObserverSvc.AddObserver(gStatusBarObserver, 'current_view_encoding_changed');
    gObserverSvc.AddObserver(gStatusBarObserver, 'current_view_language_changed');
    gObserverSvc.AddObserver(gStatusBarObserver, 'current_view_linecol_changed');
    gObserverSvc.AddObserver(gStatusBarObserver, 'current_view_check_started');
    gObserverSvc.AddObserver(gStatusBarObserver, 'current_view_check_finished');
    gObserverSvc.AddObserver(gStatusBarObserver, 'status_message');
    gObserverSvc.AddObserver(gStatusBarObserver, 'scc_status_message');

    globalPrefService.prefs.AddObserver(gStatusBarPrefObserver);

    gStatusMessageStack = Components.classes["@activestate.com/koStatusMessageStack;1"].
                          createInstance(Components.interfaces.koIStatusMessageStack);
}


function StatusBar_Fini()
{
    gObserverSvc.RemoveObserver(gStatusBarObserver, 'post_activate_view');
    gObserverSvc.RemoveObserver(gStatusBarObserver, 'pre_deactivate_view');
    gObserverSvc.RemoveObserver(gStatusBarObserver, 'current_view_encoding_changed');
    gObserverSvc.RemoveObserver(gStatusBarObserver, 'current_view_language_changed');
    gObserverSvc.RemoveObserver(gStatusBarObserver, 'current_view_linecol_changed');
    gObserverSvc.RemoveObserver(gStatusBarObserver, 'current_view_check_started');
    gObserverSvc.RemoveObserver(gStatusBarObserver, 'current_view_check_finished');
    gObserverSvc.RemoveObserver(gStatusBarObserver, 'status_message');
    gObserverSvc.RemoveObserver(gStatusBarObserver, 'scc_status_message');
    globalPrefService.prefs.RemoveObserver(gStatusBarPrefObserver);
    gStatusBarObserver = null;
    gStatusBarPrefObserver = null;
    gStatusMessageStack = null;
}

function Test_DumpStatusManager()
{
    // dumps to stdout
    gStatusMessageStack.Dump();
}
