/* Copyright (c) 2000-2001 ActiveState SRL. 
   See the file LICENSE.txt for licensing information. */

/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * Komodo's Find and Replace general functions
 *
 * Presumably these are to used by the editor for findNext shortcuts and
 * by the Find and Replace dialog.
 *
 * Usage:
 *      Find_FindNext(editor, context, pattern)
 *      Find_Replace(editor, context, pattern, replacement)
 *      Find_ReplaceAll(editor, context, pattern, replacement)
 * where:
 *      "editor" is generally the editor object on which getFocusedView() et al
 *          can be called;
 *      "pattern" is the pattern string to search for;
 *      "replacement" is the replacement string to use; and
 *      "context" is a koIFindContext component.
 */


//---- globals

var findSvc = null; // the find XPCOM service that does all the grunt work

// The find session is used to hold find results and to determine when to stop
// searching (i.e. when we have come full circle).
var gFindSession = Components.classes["@activestate.com/koFindSession;1"]
                   .getService(Components.interfaces.koIFindSession);

var gLastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
                    .getService(Components.interfaces.koILastErrorService);


//--- internal support routines

function _SetupAndFindNext(editor, context, pattern, mode)
{
    // Return a find result from using the Find Service. Return null if the
    // find session is complete (i.e. if the pattern could not be found or if
    // repeated finds fully wrapped around the find context.)
    // "mode" is either 'find' or 'replace'. The maintenance of a find session
    //      needs to know if a find-next operation is in the context of a
    //      finding or replacing.

    // abort if there is no current view
    var view = editor.gLastFocusedViewManager.getCurrentView();
    if (view == null) {
        return null;
    }

    // Get the necessary data (context may be other than the whole file).
    var scimoz = view._scimoz;
    var url = view.url;
    var text;
    var contextOffset;
    var startOffset;
    if (context.type == Components.interfaces.koIFindContext.FCT_CURRENT_DOC
        || context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS)
    {
        text = scimoz.text;
        contextOffset = 0;
        startOffset = scimoz.currentPos;
    }
    else if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION)
    {
        text = scimoz.getTextRange(context.startIndex, context.endIndex);
        contextOffset = context.startIndex;
        startOffset = scimoz.currentPos - context.startIndex;
    }
    else
    {
        throw("unexpected context: name='" + context.name + "' type=" + context.type);
    }

    // Repeated searching forward works fine because a forward search sets up
    // for the next search when it selects the find result, which, as a
    // side-effect, sets the cursor position to the end of the find result.
    // This does not work in reverse because you keep getting the same result.
    // If there is a current find result (i.e. a selection) then set the
    // startOffset to the *start* of the selection.
    if (findSvc.options.searchBackward && scimoz.selText != "") {
        startOffset -= scimoz.selectionEnd - scimoz.selectionStart;
    }

    // Special case: avoid duplicate matches for special patterns that match no
    // characters.
    // If there was a previous match in this URL that matched no characters
    // (e.g. "^", "$", "(...)*"), then advance the startOffset by one character
    // to ensure that this match does not happen again (because duplicate
    // matches end a find session).
    // XXX I *think* this is a valid fix, i.e. the result is what the user
    //     expects.
    var lastFindResult = gFindSession.GetLastFindResult();
    var unadjustedStartOffset = startOffset;
    if (lastFindResult && url == lastFindResult.url
        && lastFindResult.value.length == 0)
    {
        if (findSvc.options.searchBackward) {
            startOffset -= 1;
            if (startOffset < 0) {
                startOffset = text.length-1;
            }
        } else {
            startOffset += 1;
            if (startOffset > text.length-1) {
                startOffset = 0;
            }
        }
    }

    // find it
    gFindSession.StartFind(pattern, url, findSvc.options,
                           // Lie to the find session manager about the
                           // possible startOffset adjustment, else the find
                           // session will reset itself.
                           unadjustedStartOffset + contextOffset,
                           scimoz.selectionStart, scimoz.selectionEnd,
                           mode);
    var findResult = null;
    while (1) {  // while not done searching all file(s)
        findResult = findSvc.find(url, text, pattern, startOffset);
        // fix up the find result for context
        if (findResult != null) {
            findResult.start += contextOffset;
            findResult.end += contextOffset;
        }

        // if was not found then wrap (if have not already) and try again
        if (findResult == null && !gFindSession.wrapped) {
            gFindSession.wrapped = true;
            if (!findSvc.options.searchBackward) {
                startOffset = 0;
            } else {
                startOffset = text.length - 1;
            }
            continue;
        }

        // If finished searching the current document then:
        // - restore the cursor/selection state
        // - move to the next document (if we are searching more than one)
        if (findResult == null || gFindSession.WasAlreadyFound(findResult)
            || (mode == "replace" && gFindSession.IsRecursiveHit(findResult)))
        {
            // Switch to the next document to search, if there are any.
            if (context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS) {
                var oldView = view;
                if (findSvc.options.searchBackward) {
                    view = editor.gLastFocusedViewManager.getPreviousView();
                } else {
                    view = editor.gLastFocusedViewManager.getNextView();
                }
                if (view == null || view == oldView) {
                    break;  // Either there is no view or only the one.
                }
                url = view.url;
                // restore the cursor/selection state in current document
                scimoz.hideSelection(true);
                scimoz.currentPos = gFindSession.fileStartPos;
                scimoz.selectionStart = gFindSession.fileSelectionStartPos;
                scimoz.selectionEnd = gFindSession.fileSelectionEndPos;
                scimoz.hideSelection(false);
                scimoz.ensureVisibleEnforcePolicy(scimoz.lineFromPosition(scimoz.currentPos));
                // we are done if this document has already been searched
                if (gFindSession.HaveSearchedThisUrlAlready(url)) {
                    break;
                }
                // make the switch
                editor.gLastFocusedViewManager.setCurrentView(view);
                scimoz = view._scimoz;
            } else {
                break;
            }

            // Save the start information in this doc.
            gFindSession.fileStartPos = scimoz.currentPos;
            gFindSession.fileSelectionStartPos = scimoz.selectionStart;
            gFindSession.fileSelectionEndPos = scimoz.selectionEnd;

            // Setup for a search in this document. (Currently only "all open
            // documents" should reach this code.)
            text = scimoz.text;
            if (findSvc.options.searchBackward) {
                startOffset = text.length - 1;
            } else {
                startOffset = 0;
            }
            contextOffset = 0;

            continue;
        }
        
        // ignore results in hidden text if this option is specified
        if (findSvc.options.ignoreHiddenText) {
            if (scimoz.getLineVisible(scimoz.lineFromPosition(findResult.start)) &&
                scimoz.getLineVisible(scimoz.lineFromPosition(findResult.end))) {
                // do nothing, i.e. fall through
            } else {
                if (findSvc.options.searchBackward) {
                    startOffset = findResult.start - contextOffset;
                } else {
                    startOffset = findResult.end - contextOffset;
                }
                findResult = null;
                continue;
            }
        }

        // fix up the find result for context and return it
        if (findResult != null) {
            gFindSession.NoteFind(findResult);
            return findResult;
        }
    }

    // Restore the editor state to that when this find session started.
    scimoz.hideSelection(true);
    if (context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS) {
        // switch to the original url if it is still open
        var views = editor.gLastFocusedViewManager.getViews();
        var originalView = views[gFindSession.firstUrl];
        if (originalView != null) {
            editor.gLastFocusedViewManager.setCurrentView(originalView);
        }
        scimoz.currentPos = gFindSession.firstFileStartPos;
        scimoz.selectionStart = gFindSession.firstFileSelectionStartPos;
        scimoz.selectionEnd = gFindSession.firstFileSelectionEndPos;
    } else if (context.type == Components.interfaces.koIFindContext.FCT_CURRENT_DOC) {
        scimoz.currentPos = gFindSession.firstFileStartPos;
        scimoz.selectionStart = gFindSession.firstFileSelectionStartPos;
        scimoz.selectionEnd = gFindSession.firstFileSelectionEndPos;
    } else if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
        scimoz.selectionStart = context.startIndex;
        scimoz.selectionEnd = context.endIndex;
    } else {
        throw("unexpected context: name='" + context.name + "' type=" + context.type);
    }
    scimoz.hideSelection(false);
    scimoz.ensureVisibleEnforcePolicy(scimoz.lineFromPosition(scimoz.currentPos));
    return null;
}


function _SetupAndFindAll(editor, context, pattern)
{
    // Return all find results from using the Find Service. Return an empty
    // array if the pattern is not found.

    // abort if there is no current view
    var view = editor.gLastFocusedViewManager.getCurrentView();
    if (view == null) {
        return null;
    }

    // Get the necessary data (context may be other than the whole file).
    var scimoz = view._scimoz;
    var url = view.url;
    var text;
    var contextOffset;
    if (context.type == Components.interfaces.koIFindContext.FCT_CURRENT_DOC
        || context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS)
    {
        text = scimoz.text;
        contextOffset = 0;
    } else if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
        text = scimoz.getTextRange(context.startIndex, context.endIndex);
        contextOffset = context.startIndex;
    } else {
        throw("unexpected context: name='" + context.name + "' type=" + context.type);
    }

    //logDebug("Reset find session, because starting a 'FindAll' action.");
    gFindSession.Reset();
    gFindSession.StartFind(pattern, url, findSvc.options,
                           scimoz.currentPos,
                           scimoz.selectionStart, scimoz.selectionEnd,
                           'find');
    var i;
    var allFindResults = new Array();
    while (1) {  // while not done searching current all file(s)
        // Find all occurrences in the current document.
        var dummyCount = new Object();
        var findResults = new Array();
        findResults = findSvc.findall(url, text, pattern, dummyCount);

        // Skip results that are in folded sections.
        if (findSvc.options.ignoreHiddenText) {
            var visibleFindResults = new Array();
            for (i = 0; i < findResults.length; ++i) {
                var findResult = findResults[i];
                var start = findResult.start + contextOffset;
                var end = findResult.end + contextOffset;
                if (scimoz.getLineVisible(scimoz.lineFromPosition(start)) &&
                    scimoz.getLineVisible(scimoz.lineFromPosition(end))) {
                    visibleFindResults.push(findResult); 
                }
            }
            findResults = visibleFindResults;
        }
    
        // Fix up the results for context, fill in some data for display, and
        // shove them onto the master list.
        for (i = 0; i < findResults.length; ++i) {
            var fr = findResults[i];
            fr.start += contextOffset;
            fr.end += contextOffset;

            var startLineNum = scimoz.lineFromPosition(fr.start);
            var endLineNum = scimoz.lineFromPosition(fr.end);
            var contextStartPos = scimoz.positionFromLine(startLineNum);
            var contextEndPos = scimoz.getLineEndPosition(endLineNum);
            fr.context_ = scimoz.getTextRange(contextStartPos, contextEndPos);
            fr.line = startLineNum + 1;
            fr.column = fr.start - scimoz.positionFromLine(startLineNum);

            gFindSession.NoteFind(fr);
            allFindResults.push(fr);
        }
        gFindSession.NoteUrl(url);

        // If finished searching the current document then:
        // - restore the cursor/selection state
        // - move to the next document (if we are searching more than one)
        if (context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS) {
            // find the document to switch to
            var oldView = view;
            if (findSvc.options.searchBackward) {
                view = editor.gLastFocusedViewManager.getPreviousView();
            } else {
                view = editor.gLastFocusedViewManager.getNextView();
            }
            if (view == null || view == oldView) {
                break;  // Either there is no view or only the one.
            }
            url = view.url;
            // restore the cursor/selection state in current document
            scimoz.hideSelection(true);
            scimoz.currentPos = gFindSession.fileStartPos;
            scimoz.selectionStart = gFindSession.fileSelectionStartPos;
            scimoz.selectionEnd = gFindSession.fileSelectionEndPos;
            scimoz.hideSelection(false);
            scimoz.ensureVisibleEnforcePolicy(scimoz.lineFromPosition(scimoz.currentPos));
            // we are done if this document has already been searched
            if (gFindSession.HaveSearchedThisUrlAlready(url)) {
                break;
            }
            // make the switch
            editor.gLastFocusedViewManager.setCurrentView(view);
            scimoz = view._scimoz;
            text = scimoz.text;
            contextOffset = 0;

            // Save the start information in this doc.
            gFindSession.fileStartPos = scimoz.currentPos;
            gFindSession.fileSelectionStartPos = scimoz.selectionStart;
            gFindSession.fileSelectionEndPos = scimoz.selectionEnd;

        } else {
            break;
        }
    }

    // Restore the editor state to that when this find session started.
    scimoz.hideSelection(true);
    if (context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS) {
        // switch to the original url if it is still open
        var views = editor.gLastFocusedViewManager.getViews();
        var originalView = views[gFindSession.firstUrl];
        if (originalView != null) {
            editor.gLastFocusedViewManager.setCurrentView(originalView);
        }
        scimoz.currentPos = gFindSession.firstFileStartPos;
        scimoz.selectionStart = gFindSession.firstFileSelectionStartPos;
        scimoz.selectionEnd = gFindSession.firstFileSelectionEndPos;
    } else if (context.type == Components.interfaces.koIFindContext.FCT_CURRENT_DOC) {
        scimoz.currentPos = gFindSession.firstFileStartPos;
        scimoz.selectionStart = gFindSession.firstFileSelectionStartPos;
        scimoz.selectionEnd = gFindSession.firstFileSelectionEndPos;
    } else if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
        scimoz.selectionStart = context.startIndex;
        scimoz.selectionEnd = context.endIndex;
    } else {
        throw("unexpected context: name='" + context.name + "' type=" + context.type);
    }
    scimoz.hideSelection(false);
    scimoz.ensureVisibleEnforcePolicy(scimoz.lineFromPosition(scimoz.currentPos));

    return allFindResults;
}


function _ReplaceLastFindResult(editor, context, pattern, replacement)
{
    // Replace the last find result in the current find session, if there is
    // one and if it is genuine and current. No return value.

    var view = editor.gLastFocusedViewManager.getCurrentView();
    var scimoz = view._scimoz;
    var url = view.url;
    var findResult = gFindSession.GetLastFindResult();
    var replaceResult = null;
    if (findResult != null) {
        var findText = scimoz.getTextRange(findResult.start,
                                           findResult.end);
        var startOffset;
        var contextOffset = findResult.start;
        if (findSvc.options.searchBackward) {
            startOffset = findText.length;
        } else {
            startOffset = 0;
        }
        gFindSession.StartReplace(pattern, replacement, url, findSvc.options);
        replaceResult = findSvc.replace(url, findText, pattern,
                                        replacement, startOffset);
        if (replaceResult) {
            // fix it up
            replaceResult.start += contextOffset;
            replaceResult.end += contextOffset;
        }
        
        // - make the replacement (if the findResult is genuine)
        if (replaceResult &&
            findResult.start == replaceResult.start &&
            findResult.end == replaceResult.end &&
            findText == replaceResult.value)
        {
            //dump("replace result:" + replaceResult.start +
            //     "-" + replaceResult.end + ": '" + replaceResult.value +
            //     "' -> '" + replaceResult.replacement + "'\n");
            //XXX Scintilla Bug: Changing a line that is a fold line will open
            //    the fold. This makes it impossible to 'ignoreHiddenText'
            //    when replacements occur on fold lines.
            //WORKAROUND: If the replacement line was a collapsed fold line
            //    then re-collapse it if necessary.
            //LIMITATION: This re-collapsing will be undoable via Ctrl-Z.
            var line = scimoz.lineFromPosition(replaceResult.end);
            var wasCollapsed = (scimoz.getFoldLevel(line) &
                                scimoz.SC_FOLDLEVELHEADERFLAG) &&
                               !scimoz.getFoldExpanded(line);
            scimoz.hideSelection(true);
            scimoz.selectionStart = replaceResult.start;
            scimoz.selectionEnd = replaceResult.end;
            scimoz.replaceSel(replaceResult.replacement);
            scimoz.selectionStart = replaceResult.start;
            scimoz.selectionEnd = replaceResult.start +
                                  replaceResult.replacement.length;
            scimoz.hideSelection(false);
            if (wasCollapsed && scimoz.getFoldExpanded(line)) {
                scimoz.toggleFold(line);
            }
            // fix up the search context as appropriate
            if (context.type == Components.interfaces.koIFindContext.FCT_CURRENT_DOC
                || context.type == Components.interfaces.koIFindContext.FCT_ALL_OPEN_DOCS)
            {
                // nothing for now
            } else if (context.type == Components.interfaces.koIFindContext.FCT_SELECTION) {
                // adjust the context range for the change in length
                context.endIndex += replaceResult.replacement.length -
                                    replaceResult.value.length;
            } else {
                throw("unexpected context: name='" + context.name + "' type=" + context.type);
            }

            // Fill in some data for possible display in the "Replace Results"
            // tab and note the replacement.
            var startLineNum = scimoz.lineFromPosition(replaceResult.start);
            var endLineNum = scimoz.lineFromPosition(replaceResult.end);
            var contextStartPos = scimoz.positionFromLine(startLineNum);
            var contextEndPos = scimoz.getLineEndPosition(endLineNum);
            replaceResult.context_ = scimoz.getTextRange(contextStartPos, contextEndPos);
            replaceResult.line = startLineNum + 1;
            replaceResult.column = replaceResult.start -
                                   scimoz.positionFromLine(startLineNum);
            gFindSession.NoteReplace(replaceResult);
        }
    }
}


function _DisplayFindResult(editor, findResult)
{
    //dump("display find result:" + findResult.start + "-" +
    //     findResult.end + ": '" + findResult.value + "'\n");
    // Jump to the find result and select it.
    // Jump to the _end_ (i.e. the start if going backwards) of the
    // result so that repeated finds are non-overlapping.
    var scimoz = editor.gLastFocusedViewManager.getCurrentView()._scimoz;

    //XXX Setting the SciMoz selection has the side effect of moving
    //    the current cursor position to selectionEnd, so this gotoPos
    //    is effective useless.
    if (!findSvc.options.searchBackward) {
        scimoz.gotoPos(findResult.end);
    } else {
        scimoz.gotoPos(findResult.start);
    }
    scimoz.hideSelection(true);
    scimoz.selectionStart = findResult.start;
    scimoz.selectionEnd = findResult.end;
    scimoz.hideSelection(false);

    // Ensure folded lines are expanded. We *could* conceivably have search
    // results that span lines so should ensure that the whole selection is
    // visible. XXX should include every line in btwn here
    scimoz.ensureVisibleEnforcePolicy(scimoz.lineFromPosition(findResult.start))
    scimoz.ensureVisibleEnforcePolicy(scimoz.lineFromPosition(findResult.end))

    editor.DoTheScimozWiggle(scimoz);

    // Ensure that the find result is not obstructed by the find dialog.
    // - If then "editor" and "window" are the same object then there is no
    //   find dialog to occlude or move.
    if (editor == window) {
        //dump("There is find dialog to occlude the find result.\n");
        return;
    }

    // - If the find dialog is not in the editor region then it cannot be
    //   occluding.
    // XXX The Mozilla widget's boxObject dimensions do NOT include the Window
    //     Manager's decorations (e.g. title bar and border's).  Presumably
    //     these *could* be obtained from Win32/GTK+ calls, but hardcoding a
    //     small value will just be used (close enough I hope).
    var wmDecorationTop = 23;
    var wmDecorationBottom = 3;
    var wmDecorationLeft = 3;
    var wmDecorationRight = 3;
    var editorWidget = editor.document.getElementById("scin");
    var editorTop_Screen = editorWidget.boxObject.y + wmDecorationTop
                           + editor.screenY;
    var editorBottom_Screen = editorTop_Screen + editorWidget.boxObject.height;
    var editorLeft_Screen = editorWidget.boxObject.x + wmDecorationLeft
                            + editor.screenX;
    var editorRight_Screen = editorLeft_Screen + editorWidget.boxObject.width;
    //dump("editor region: left="+editorLeft_Screen
    //     +" right="+editorRight_Screen
    //     +" top="+editorTop_Screen
    //     +" bottom="+editorBottom_Screen+"\n");

    var findDialogTop_Screen = window.screenY; 
    var findDialogBottom_Screen = window.screenY + window.outerHeight; 
    var findDialogLeft_Screen = window.screenX; 
    var findDialogRight_Screen = window.screenX + window.outerWidth; 
    var findDialogHeight = window.outerHeight; 
    //dump("find dialog: left="+findDialogLeft_Screen
    //     +" right="+findDialogRight_Screen
    //     +" top="+findDialogTop_Screen
    //     +" bottom="+findDialogBottom_Screen+"\n");
    
    if (findDialogBottom_Screen <= editorTop_Screen
        || findDialogTop_Screen >= editorBottom_Screen
        || findDialogLeft_Screen >= editorRight_Screen
        || findDialogRight_Screen <= editorLeft_Screen)
    {
        //dump("Find dialog is not over editor region. Cannot be occluding.\n");
        return;
    }

    // - See if the horizontal space containing the find result hits the find
    //   dialog.
    var findResultTop_SciMoz = scimoz.pointYFromPosition(findResult.start);
    var findResultBottom_SciMoz = scimoz.pointYFromPosition(
        scimoz.positionFromLine(scimoz.lineFromPosition(findResult.end)+1));
    var findResultTop_Screen = findResultTop_SciMoz + editorTop_Screen;
    var findResultBottom_Screen = findResultBottom_SciMoz + editorTop_Screen;
    //dump("find result: top="+findResultTop_Screen
    //     +" bottom="+findResultBottom_Screen+"\n");

    if (findDialogBottom_Screen <= findResultTop_Screen
        || findDialogTop_Screen >= findResultBottom_Screen)
    {
        //dump("Find dialog is not over the horizontal region of the find result.\n");
        return;
    }

    // - Find dialog should be moved. Is there room above?
    if (findDialogHeight <=findResultTop_Screen) {
        // There is room on top, move it up.
        //XXX Could move it a bit higher if there is more than enough room
        //    (say, above the whole editor region).
        window.moveTo(findDialogLeft_Screen,
                      findResultTop_Screen-findDialogHeight);
        return;
    }

    // - No room above. Is there room below?
    if (findDialogHeight <=screen.availHeight-findResultBottom_Screen) {
        // There is room on top, move it up.
        //XXX Could move it a bit higher if there is more than enough room
        //    (say, above the whole editor region).
        window.moveTo(findDialogLeft_Screen,
                      findResultBottom_Screen);
        return;
    }

    // - There is no room to move it vertically out of the way. Give up.
}


function _DisplayFindResults(context, findResults, patternAlias)
{
    // Display the given find results in appropriate buffer.
    FindResultsTab_Clear(parentWindow);
    FindResultsTab_DescribeFind(parentWindow, gFindSession.GetPattern(),
                                context, findSvc.options, patternAlias);
    for (var i = 0; i < findResults.length; ++i) {
        var fr = findResults[i];
        var fileName = displayNameFromURL(fr.url);
        FindResultsTab_AddFindResult(parentWindow, fr.url, fr.start, fr.end,
                                     fr.value, fileName, fr.line, fr.column,
                                     fr.context_);
    }
    FindResultsTab_Show(parentWindow);
}


function _MarkFindResults(editor, context, findResults, patternAlias)
{
    // Mark all the given find results.
    var vm = editor.gLastFocusedViewManager;
    var view = vm.getCurrentView();
    var url = null;
    var scimoz = null;

    for (var i = 0; i < findResults.length; ++i) {
        var fr = findResults[i];
        // Switch to that file if it is not the current one.
        if (url != fr.url) {
            editor.doFileOpen(fr.url);
            url = fr.url;
            view = vm.getCurrentView();
            url = view.url;
            scimoz = view._scimoz;
        }

        //XXX 'mo' (Marker Order) in editor-debugger.js. Ugly.
        if (! (scimoz.markerGet(fr.line-1) & (1 << editor.mo.bm))) {
            scimoz.markerAdd(fr.line-1, editor.mo.bm);
        }
    }
}


function _DisplayReplaceResults(context, replaceResults)
{
    // Display the given find results in appropriate buffer.
    FindResultsTab_Clear(parentWindow);
    FindResultsTab_DescribeReplace(parentWindow,
                                   gFindSession.GetPattern(),
                                   gFindSession.GetReplacement(),
                                   context, findSvc.options);
    for (var i = 0; i < replaceResults.length; ++i) {
        var rr = replaceResults[i];
        var fileName = displayNameFromURL(rr.url);
        FindResultsTab_AddReplaceResult(parentWindow, rr.url, rr.start, rr.end,
                                        rr.value, rr.replacement, fileName,
                                        rr.line, rr.column, rr.context_);
    }
    FindResultsTab_Show(parentWindow);
}


function _UiForCompletedFindSession(editor, context)
{
    var numFinds = gFindSession.GetNumFinds();
    var numReplacements = gFindSession.GetNumReplacements();
    
    // Pop up an alert box with an appropriate message.
    var msg = "Komodo has finished searching " + context.name + ".";
    if (numFinds == 0) {
        msg += " The search item was not found.";
    } else if (numReplacements > 0) {
        msg += " " + numReplacements + " of " + numFinds +
               " occurrence(s) were replaced.";
    } else {
        msg += " " + numFinds + " occurrence(s) were found.";
    }
    alertDialog(msg);
}


function _UiForFindServiceError()
{
    var msg = "There was a syntax error in your search pattern: ";
    msg += gLastErrorSvc.getLastErrorMessage();
    msg += ".";
    alertDialog(msg);
}

function Find_FindNextInMacro(editor, contexttype, pattern, patternType, caseSensitivity, searchBackward, matchWord, ignoreHiddenText, mode, quiet) {
    // Called by macros -- only difference with Find_FindNext is that contexttype is
    // a constant, and the context needs to be created.
    // XXX the APIs should be factored out so that these "inmacro" versions aren't necessary.
    var old_patternType, old_caseSensitivity, old_searchBackward, old_matchWord, old_ignoreHiddenText;
    var context = Components.classes["@activestate.com/koFindContext;1"]
            .createInstance(Components.interfaces.koIFindContext);
    context.type = contexttype;
    
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"]
                  .getService(Components.interfaces.koIFindService);
    }

    // Stash old options
    old_patternType = findSvc.options.patternType;
    old_caseSensitivity = findSvc.options.caseSensitivity;
    old_searchBackward = findSvc.options.searchBackward;
    old_matchWord = findSvc.options.matchWord;
    old_ignoreHiddenText= findSvc.options.ignoreHiddenText;
    
    findSvc.options.patternType = patternType;
    findSvc.options.caseSensitivity = caseSensitivity;
    findSvc.options.searchBackward = searchBackward;
    findSvc.options.matchWord = matchWord;
    findSvc.options.ignoreHiddenText = ignoreHiddenText;

    Find_FindNext(editor, context, pattern, mode, quiet);
    
    // restore options
    findSvc.options.patternType = old_patternType;
    findSvc.options.caseSensitivity = old_caseSensitivity;
    findSvc.options.searchBackward = old_searchBackward;
    findSvc.options.matchWord = old_matchWord;
    findSvc.options.ignoreHiddenText = old_ignoreHiddenText;
}

function Find_FindAllInMacro(editor, contexttype, pattern, patternType, caseSensitivity, searchBackward, matchWord, ignoreHiddenText) {
    // Called by macros -- only difference with Find_FindAll is that contexttype is
    // a constant, the context needs to be created, options need to be set (and unset)
    // and that the parentWindow needs to be set
    // XXX the APIs should be factored out so that these "inmacro" versions aren't necessary.

    var old_patternType, old_caseSensitivity, old_searchBackward, old_matchWord, old_ignoreHiddenText;
    var context = Components.classes["@activestate.com/koFindContext;1"]
            .createInstance(Components.interfaces.koIFindContext);
    context.type = contexttype;
    
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"]
                  .getService(Components.interfaces.koIFindService);
    }

    // Stash old options
    old_patternType = findSvc.options.patternType;
    old_caseSensitivity = findSvc.options.caseSensitivity;
    old_searchBackward = findSvc.options.searchBackward;
    old_matchWord = findSvc.options.matchWord;
    old_ignoreHiddenText= findSvc.options.ignoreHiddenText;
    
    findSvc.options.patternType = patternType;
    findSvc.options.caseSensitivity = caseSensitivity;
    findSvc.options.searchBackward = searchBackward;
    findSvc.options.matchWord = matchWord;
    findSvc.options.ignoreHiddenText = ignoreHiddenText;

    parentWindow = window;
    Find_FindAll(editor, context, pattern);
    
    // restore options
    findSvc.options.patternType = old_patternType;
    findSvc.options.caseSensitivity = old_caseSensitivity;
    findSvc.options.searchBackward = old_searchBackward;
    findSvc.options.matchWord = old_matchWord;
    findSvc.options.ignoreHiddenText = old_ignoreHiddenText;
}

function Find_MarkAllInMacro(editor, contexttype, pattern, patternType, caseSensitivity, searchBackward, matchWord, ignoreHiddenText) {
    // Called by macros -- only difference with Find_MarkAll is that contexttype is
    // a constant, the context needs to be created, and that the parentWindow needs to be set
    // XXX the APIs should be factored out so that these "inmacro" versions aren't necessary.

    var old_patternType, old_caseSensitivity, old_searchBackward, old_matchWord, old_ignoreHiddenText;
    var context = Components.classes["@activestate.com/koFindContext;1"]
            .createInstance(Components.interfaces.koIFindContext);
    context.type = contexttype;
    
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"]
                  .getService(Components.interfaces.koIFindService);
    }

    // Stash old options
    old_patternType = findSvc.options.patternType;
    old_caseSensitivity = findSvc.options.caseSensitivity;
    old_searchBackward = findSvc.options.searchBackward;
    old_matchWord = findSvc.options.matchWord;
    old_ignoreHiddenText= findSvc.options.ignoreHiddenText;
    
    findSvc.options.patternType = patternType;
    findSvc.options.caseSensitivity = caseSensitivity;
    findSvc.options.searchBackward = searchBackward;
    findSvc.options.matchWord = matchWord;
    findSvc.options.ignoreHiddenText = ignoreHiddenText;

    parentWindow = window;
    Find_MarkAll(editor, context, pattern);
    
    // restore options
    findSvc.options.patternType = old_patternType;
    findSvc.options.caseSensitivity = old_caseSensitivity;
    findSvc.options.searchBackward = old_searchBackward;
    findSvc.options.matchWord = old_matchWord;
    findSvc.options.ignoreHiddenText = old_ignoreHiddenText;
}

//---- public functions

function Find_FindNext(editor, context, pattern, mode, quiet)
{
    if (typeof(mode) == 'undefined') {
        mode = 'find';
    }
    if (typeof(quiet) == 'undefined') {
        quiet = 0;
    }
    if (editor.gMacroMgr.recording) {
        editor.gMacroMgr.appendCode("Find_FindNextInMacro(window, " +
                                    context.type + ", '" + 
                                    pattern.replace('\'', '\\\'', 'g') + "', " + 
                                    findSvc.options.patternType + ", " +
                                    findSvc.options.caseSensitivity + ", " +
                                    findSvc.options.searchBackward + ", " +
                                    findSvc.options.matchWord + ", " +
                                    findSvc.options.ignoreHiddenText + ", " + 
                                    mode + ", " +
                                    quiet +
                                    ");\n"); 
    }
    // Find the next match given the current state of the editor buffer and
    // do all the UI work to display the found result.
    // "mode" is either 'find' or 'replace'. The maintenance of a find session
    //      needs to know if a find-next operation is in the context of a
    //      finding or replacing.
    // Returns true iff a result was found and displayed.
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"].
                  getService(Components.interfaces.koIFindService);
    }

    // Find the next "hit". If no result is returned then the find session is
    // complete.
    var findResult = null;
    try {
        findResult = _SetupAndFindNext(editor, context, pattern, mode);
    } catch(e) {
        if (!quiet)
            _UiForFindServiceError();
        return null;
    }
    MRU_add("find-patternMru", pattern, true);

    if (findResult) {
        _DisplayFindResult(editor, findResult);
    } else {
        if (!quiet)
            _UiForCompletedFindSession(editor, context);
        //logDebug("Reset find session, because 'FindNext' did not find "
        //         + "anything.");
        gFindSession.Reset();
    }
    return findResult != null;
}


function Find_FindAll(editor, context, pattern, patternAlias)
{
    // Returns true iff any matches were found and displayed.
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"]
                  .getService(Components.interfaces.koIFindService);
    }

    if (editor.gMacroMgr.recording) {
        editor.gMacroMgr.appendCode("Find_FindAllInMacro(window, " +
                                    context.type + ", '" + 
                                    pattern.replace('\'', '\\\'', 'g') + "', " + 
                                    findSvc.options.patternType + ", " +
                                    findSvc.options.caseSensitivity + ", " +
                                    findSvc.options.searchBackward + ", " +
                                    findSvc.options.matchWord + ", " +
                                    findSvc.options.ignoreHiddenText + 
                                    ");\n"); 
    }
    var findResults = null;
    try {
        findResults = _SetupAndFindAll(editor, context, pattern);
    } catch(e) {
        _UiForFindServiceError();
        return null;
    }
    MRU_add("find-patternMru", pattern, true);

    if (findResults.length > 0) {
        _DisplayFindResults(context, findResults, patternAlias);
    } else {
        _UiForCompletedFindSession(editor, context);
    }
    //logDebug("Reset find session, because finished a 'FindAll' action");
    gFindSession.Reset();

    return findResults.length > 0;
}


function Find_MarkAll(editor, context, pattern, patternAlias)
{
    // Returns true iff any matches were found and displayed.
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"]
                  .getService(Components.interfaces.koIFindService);
    }

    if (editor.gMacroMgr.recording) {
        editor.gMacroMgr.appendCode("Find_MarkAllInMacro(window, " +
                                    context.type + ", '" + 
                                    pattern.replace('\'', '\\\'', 'g') + "', " + 
                                    findSvc.options.patternType + ", " +
                                    findSvc.options.caseSensitivity + ", " +
                                    findSvc.options.searchBackward + ", " +
                                    findSvc.options.matchWord + ", " +
                                    findSvc.options.ignoreHiddenText + 
                                    ");\n"); 
    }

    var findResults = null;
    try {
        findResults = _SetupAndFindAll(editor, context, pattern);
    } catch(e) {
        _UiForFindServiceError();
        return null;
    }
    MRU_add("find-patternMru", pattern, true);

    if (findResults.length > 0) {
        _MarkFindResults(editor, context, findResults, patternAlias);
    } else {
        _UiForCompletedFindSession(editor, context);
    }
    //logDebug("Reset find session, because finished a 'MarkAll' action");
    gFindSession.Reset();

    return findResults.length > 0;
}


function Find_Replace(editor, context, pattern, replacement)
{
    // Replace the currently selected find result (if there is one) and find
    // the next occurrence of the current pattern. Returns true iff a next
    // occurence of the pattern was found and hilighted.
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"].
                  getService(Components.interfaces.koIFindService);
    }

    var view = editor.gLastFocusedViewManager.getCurrentView();
    var findResult = null;
    if (view != null) {
        // Here is how a replace works:
        //  - the first "Replace" will just identify the replacement to make
        //  - the second "Replace" will make that replacement and find the next
        //  - and so on...

        // replace the current find result if there is one
        var scimoz = view._scimoz;
        scimoz.beginUndoAction();
        _ReplaceLastFindResult(editor, context, pattern, replacement);
        scimoz.endUndoAction();
        
        // find the next one
        try {
            findResult = _SetupAndFindNext(editor, context, pattern,
                                           'replace');
        } catch(e) {
            _UiForFindServiceError();
            return null;
        }
        MRU_add("find-patternMru", pattern, true);
        view = editor.gLastFocusedViewManager.getCurrentView();

        // Do the appropriate UI work for results.
        if (findResult == null) {
            _UiForCompletedFindSession(editor, context);
            //logDebug("Reset find session, because 'Replace' action "
            //         + "returned no result.");
            gFindSession.Reset();
        } else {
            _DisplayFindResult(editor, findResult);
        }
    }
    return findResult != null;
}


function Find_ReplaceAll(editor, context, pattern, replacement)
{
    // Returns true iff a match was found and replaced.
    if (findSvc == null) {
        findSvc = Components.classes["@activestate.com/koFindService;1"].
                  getService(Components.interfaces.koIFindService);
    }

    var replacedSome = false;
    var view = editor.gLastFocusedViewManager.getCurrentView();
    var scimoz = view._scimoz;
    scimoz.beginUndoAction();
    scimoz.hideSelection(true);

    var findResult = null;
    if (scimoz.selText != "") {
        //XXX:HACK Presume any selection is the first findResult, should
        //XXX grab it from find session object. First have to extend the
        //XXX FindSession stuff to cache find results.
        findResult = new Object();
        findResult.value = scimoz.selText;
        findResult.start = scimoz.selectionStart;
        findResult.end = scimoz.selectionEnd;
    }

    try {
        while (1) {
            // replace the current find result if there is one
            _ReplaceLastFindResult(editor, context, pattern,
                                   replacement);

            // - find the next occurence
            findResult = _SetupAndFindNext(editor, context, pattern, 'replace');
            if (findResult == null) break;
        }
    } finally {
        MRU_add("find-patternMru", pattern, true);
        scimoz.hideSelection(false);
        scimoz.endUndoAction();
        if (gFindSession.GetNumReplacements() > 0) {
            var dummyCount = new Object();
            var replaceResults = new Object();
            replaceResults = gFindSession.GetReplacements(dummyCount);
            _DisplayReplaceResults(context, replaceResults);
            replacedSome = true;
        } else {
            _UiForCompletedFindSession(editor, context);
        }
        //logDebug("Reset find session, because finished a 'ReplaceAll' "
        //         + "action.");
        gFindSession.Reset();
    }

    return replacedSome;
}


