/* Copyright (c) 2000-2001 ActiveState Tool Corporation.
   See the file LICENSE.txt for licensing information. */

// Manage "project extensions" ('pe's for short).
//
//                  HOW TO WRITE A PROJECT EXTENSION
//
// A project extension is a singleton instance of a JS class. Currently (at
// least) pe's are defined in pe*.js files in the same directory as this file
// (e.g. the "command" pe is coded in "peCommand.js").
//
// A pe is registered and initialized as follows:
//  - The pe*.js file is included in komodo.xul like so:
//      <script type="application/x-javascript"
//              src="chrome://komodo/content/project/peCommand.js"/>
//  - Top level code in the pe*.js file adds a singleton instance of the class
//    to the global 'registeredExtensions' (defined here).
//      registeredExtensions[registeredExtensions.length] = new peCommand();
//  - The project extension manager's init() calls .init() on each registered
//    extension.
//
// A project extension's initialization can do the following:
//  - Register right-click context menu entries:
//      extensionManager.addCommandMenu(this,       // handler
//          Components.interfaces.koIPart_command,  // part-type interface
//          'Open Command',                         // text on right-click menu
//          'cmd_runCommand');      // command to run when selected
//  - Register self as a command handler for appropriate std commands:
//      extensionManager.registerCommand('cmd_cut', this);
//  - Register for the GUI events. E.g. 'ondblclick'. Most pe's will want to
//    register for the 'onUpdatePart' event. This event is sent to extensions
//    when a koIPart is wrapped -- which equates to whenever a koIPart is
//    created or updated. (XXX right? -- TrentM) For each registered event the
//    pe must define a .<eventname>() method.
//
// If a project extension registers to handle any commands it must implement
// the usual command interface: .supportsCommand(), .isCommandEnabled() and
// .doCommand() methods.
//

var extensionManager = null;
var registeredExtensions = [];

function projectExtensionManager() {
    /* maps an interface string to an array of menuItem XUL elements */
    this.menus = {};
    /* when we store an interface into a map, javascript turns it
       into a string.  Since we need the actual interface value, we need
       an additional map to map the interface string to the interface
       value.  interfaceMap maps the string to value.  menus (above) maps
       the interface string to the
    */
    this.interfaceMap = {};
    /* maps an event to an extension, the extension must implement the even handler */
    this.handlers = {};
    /* maps a command name to a function */
    this._commands = {};
    /* used to store regex matches on command names */
    this._commandsMatches = {};

    /* datapoints are elements that an extension might fill on in a wrapped part
       a view can use this list to create new columns for the data
       this is a PUBLIC property, but setting a value should use setDatapoint()
    */
    this.datapoints = {};
    
    /* the extension manager is the primary controller, and delegates to extensions */
    window.controllers.appendController(this);
    koRequestUpdateCommands("projectsChanged");
}

projectExtensionManager.prototype.constructor = projectExtensionManager;

projectExtensionManager.prototype.init = function() {
    var i;
    for (i = 0; i < registeredExtensions.length; i++)
    {
        try {
        registeredExtensions[i].init();
        registeredExtensions[i].registerCommands();
        } catch(e) { logError(e); }
    }
    for (i = 0; i < registeredExtensions.length; i++)
    {
        try {
        registeredExtensions[i].registerEventHandlers();
        } catch(e) { logError(e); }
    }
    for (i = 0; i < registeredExtensions.length; i++)
    {
        try {
        registeredExtensions[i].registerMenus();
        } catch(e) { logError(e); }
    }
    koRequestUpdateCommands("projectsChanged");
}

projectExtensionManager.prototype.setDatapoint = function(columnname,fieldname) {
    this.datapoints[columnname] = fieldname;
}

/* handles everything necessary to add a menu item and command */
projectExtensionManager.prototype.addCommandMenu = function(handler,
                                                            partInterface,
                                                            label,
                                                            commandName,
                                                            key /* =null */,
                                                            id /* =null */) {
    this.createMenuItem(partInterface,label,commandName,key,id);
    this.registerCommand(commandName,handler);
}

/* addCommand
 * adds a command element to the command controller */
projectExtensionManager.prototype.addCommand = function(commandName) {
    var commandList = document.getElementById('komodoProjectItems');
    if (!commandList) {
        logError('extension manager unable to get element for komodoProjectItems');
        return false;
    }
    
    // if it already exists, exit
    if (commandList.getElementsByAttribute('id', commandName).length > 0) return true;
    
    var commandEl = document.createElementNS(XUL_NS, 'command');
    commandEl.setAttribute('id',commandName);
    commandEl.setAttribute('oncommand','koGoDoCommand("'+commandName+'")');
    commandList.appendChild(commandEl);
    return true;
}

projectExtensionManager.prototype.createMenuItem = function(partInterface,label,commandName,key,id) {
    var mi =document.createElementNS(XUL_NS, 'menuitem');
    mi.setAttribute('label',label);
    mi.setAttribute('command',commandName);
    if (key) mi.setAttribute('accesskey',key);
    if (id) mi.setAttribute('id',id);
    return this.addMenuItem(partInterface, mi);
}

projectExtensionManager.prototype.addMenuItem = function(partInterface,newmenu) {
    var menupopup;
    var t = typeof(this.menus[partInterface]);
    if (t == 'undefined') {
        this.interfaceMap[partInterface] = partInterface;
        if (newmenu.nodeName == 'menupopup') {
            this.menus[partInterface] = newmenu.childNodes;
            return true;
        } else {
            // create a new menu for this node type
            this.menus[partInterface] = new Array();
            this.menus[partInterface][0] = newmenu;
        }
    } else {
        if (newmenu.nodeName == 'menupopup') {
            this.menus[partInterface] = this.menus[partInterface].concat(newmenu.childNodes);
        } else {
            this.menus[partInterface][this.menus[partInterface].length] = newmenu;
        }
    }
    return true;
}

projectExtensionManager.prototype._commonCommands = function(command, items) {
    for (var i = 0; i < items.length; i++) {
        if (!this._commands[command].supportsCommand(command, items[i])) return false;
    }
    return true;
}

projectExtensionManager.prototype._commonInterface = function(partInterface, items) {
    try {
        for (var i = 0; i < items.length; i++) {
            if (!items[i].part.QueryInterface(this.interfaceMap[partInterface])) return false;
        }
        return true;
    } catch(e) { }
    return false;
}

projectExtensionManager.prototype.updateMenuForInterface = function(partInterface, element) {
    var menuitems = this.menus[partInterface];
    var children = element.childNodes;
    var i;
    for (i = 0; i < children.length; i++) {
        element.removeChild(children[i]);
    }
    for (i = 0; i < menuitems.length; i++) {
        var command = menuitems[i].getAttribute('command');
        if (command && this._commands[command].supportsCommand(command))
            element.appendChild(menuitems[i].cloneNode(menuitems[i]));
    }
}

projectExtensionManager.prototype.updateMenu = function(items) {
    var primaryInterface = null;
    var primaryMenu = [];
    if (items.length == 1) {
        primaryInterface = items[0].part.primaryInterface;
    }
    // XXX we call on commandSupported here because currently
    // commandEnabled is not being called by mozilla for some reason
    // I am assuming it is something to do with the fact that this is a
    // context menu rather than a normal menu.
    var menuitems = [];
    var sep;
    for (var partInterface in this.interfaceMap) {
        if (typeof(this.menus[partInterface]) != 'undefined' &&
            this._commonInterface(this.interfaceMap[partInterface],items)) {
            if (primaryInterface == new String(this.interfaceMap[partInterface])) {
                primaryMenu = primaryMenu.concat(this.menus[partInterface]);
                continue;
            }
            if (menuitems.length > 0) {
                sep =document.createElementNS(XUL_NS, 'menuseparator');
                menuitems.push(sep);
            }
            menuitems = menuitems.concat(this.menus[partInterface]);
        }
    }
    if (primaryMenu.length > 0) {
        if (menuitems.length > 0) {
            sep =document.createElementNS(XUL_NS, 'menuseparator');
            primaryMenu.push(sep);
        }
        menuitems = primaryMenu.concat(menuitems);
    }
    if (menuitems.length > 0) {
        var popup = document.getElementById('projectContext');
        var children = popup.childNodes;
        var i;
        for (i = 0; i < children.length; i++) {
            popup.removeChild(children[i]);
        }
        var lastmenu = null;
        for (i = 0; i < menuitems.length; i++) {
            var command = menuitems[i].getAttribute('command');
            
            // don't do two seperators in a row
            if (menuitems[i].nodeName == 'menuseparator' && lastmenu && lastmenu.nodeName == 'menuseparator')
                continue;
            
            if (!command || this._commonCommands(command, items)) {
                lastmenu = menuitems[i];
                popup.appendChild(menuitems[i].cloneNode(menuitems[i]));
            }
        }
        return true;
    }
    return false;
}

projectExtensionManager.prototype.addEventHandler = function(partInterface,eventType,eventHandler)
{
    if (typeof(this.handlers[partInterface]) == 'undefined') {
        this.interfaceMap[partInterface] = partInterface;
        this.handlers[partInterface] = {};
    }
    if (typeof(this.handlers[partInterface][eventType]) == 'undefined') {
        this.handlers[partInterface][eventType] = new Array();
    }
    this.handlers[partInterface][eventType].push(eventHandler);
}

projectExtensionManager.prototype.callHandler = function(part,eventType,event) {
    for (var partInterface in this.interfaceMap) {
        try {
            if (part.part.QueryInterface(this.interfaceMap[partInterface]) &&
                typeof(this.handlers[partInterface]) != 'undefined' &&
                typeof(this.handlers[partInterface][eventType]) != 'undefined') {
                    var handlers = this.handlers[partInterface][eventType];
                    for (var i = 0; i < handlers.length; i++) {
                        try {
                            handlers[i][eventType].apply(handlers[i],[part,event]);
                        } catch(e) {
                            logDebug("exception calling "+eventType+"\n"+e+"\n");
                        }
                    }
            }
        } catch(e) { /* no interface */ }
    }
}

//----------------------------------------------------------------------------------------------
// Controller functions

projectExtensionManager.prototype.registerCommandMatch = function(command, match, extension) {
    this._commandsMatches[command] = match;
    this.registerCommand(command,extension);
}

projectExtensionManager.prototype.getCommandMatch = function(command) {
    for (var cmd in this._commandsMatches) {
        var groups = command.match(this._commandsMatches[cmd]);
        if(groups && groups[1]) {
            return cmd;
        }
    }
    return null;
}

projectExtensionManager.prototype.registerCommand = function(command, extension) {
    this.addCommand(command);
    this._commands[command] = extension;
}

projectExtensionManager.prototype.supportsCommand = function(command) {
    var cmd = command;
    if (typeof(this._commands[command]) == 'undefined') {
        cmd = this.getCommandMatch(command);
        if (!cmd) return false;
    }
    return this._commands[cmd].supportsCommand(command, null);
}

projectExtensionManager.prototype.isCommandEnabled = function(command) {
    var cmd = command;
    if (typeof(this._commands[command]) == 'undefined') {
        cmd = this.getCommandMatch(command);
        if (!cmd) return false;
    }
    //dump("isCommandEnabled command "+cmd+"\n");
    return this._commands[cmd].isCommandEnabled(command);
}

projectExtensionManager.prototype.doCommand = function(command) {
    var cmd = command;
    if (typeof(this._commands[command]) == 'undefined') {
        cmd = this.getCommandMatch(command);
        if (!cmd) return false;
    }
    //dump("doCommand ok "+command+"\n");
    return this._commands[cmd].doCommand(command);
}

projectExtensionManager.prototype.onEvent = function(event) {
    // XXX remove if possible
}

