/* Copyright (c) 2000-2002 ActiveState Tool Corporation.
   See the file LICENSE.txt for licensing information. */

// -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
//
// The interface for using the run output window (where run command output goes
// to, in the bottom pane of the Komodo workspace).
//
// Expected usage:
//  - Someone calls RunOutput_Init() at startup and RunOutput_Fini() at
//    shutdown.
//  - When a command is to be run in the output window do this:
//      RunOutput_StartSession();  // announce intention to start session
//      //... setup and start running the actual command calling
//      //    RunOutput_GetTerminal() and RunOutput_Show() as needed
//      RunOutput_SetProcessHandle(p);  // to allow user to kill process
//      //... setup RunOutput_EndSession() to be run when the process
//      //    terminates.
//


//---- globals

var _gTerminal = null;
var _gProcess = null;


//---- internal support routines

// Clear the terminal window and associated UI. This should be called prior to
// running a command in the output window.
function _RunOutput_Clear()
{
    _gTerminal.clear();
    var descWidget = document.getElementById("runoutput-desc");
    addStyle(descWidget, "color", "black");
    descWidget.removeAttribute("value");
    descWidget.removeAttribute("_command");
}



//---- public interface

function RunOutput_Init() {
    var scinWidget = document.getElementById("runoutput-scintilla");
    scinWidget.init();

    // Setup the style for the output window using the user's text settings
    scinWidget.ISciMoz.lexer = scinWidget.ISciMoz.SCLEX_NULL;
    var styleApplier = Components.classes['@activestate.com/koScintillaStyleApplier;1']
                       .createInstance(Components.interfaces.koIScintillaStyleApplier);
    // Force the font to be fixed width by making a preference item with that
    // single pref
    var fakeViewPrefs = Components.classes['@activestate.com/koPreferenceSet;1']
                        .createInstance(Components.interfaces.koIPreferenceSet);
    fakeViewPrefs.setBooleanPref('editUseFixedFont', true);
    // Set the style and apply it.
    styleApplier.init(scinWidget.ISciMoz, 'Text', fakeViewPrefs);
    styleApplier.applyStyles();

    // Create a terminal for the output window. The terminal has std handlers
    // that proxy read/write events between the scintilla and a spawned child
    // process.
    _gTerminal = Components.classes['@activestate.com/koRunTerminal;1']
                 .createInstance(Components.interfaces.koIRunTerminal);
    _gTerminal.init(scinWidget.ISciMoz);
    scinWidget.ISciMoz.hookEvents(_gTerminal);
    scinWidget.controllers.insertControllerAt(0, _gTerminal.controller);
    _RunOutput_Clear();
}


function RunOutput_Fini()
{
    var scinWidget = document.getElementById("runoutput-scintilla");
    scinWidget.controllers.removeController(_gTerminal.controller);
    //XXX Does the scinWidget have to be finalized?
}


function RunOutput_GetTerminal()
{
    return _gTerminal;
}


// Start a terminal session in the output window with the given command.
// This raises an exception if the run output window is currently busy.
function RunOutput_StartSession(command)
{
    if (_gTerminal.active) {
        throw("A previous command is currently running in the "+
              "Command Output window. That command must finish before "+
              "starting another.");
    }
    _RunOutput_Clear();
    _gTerminal.startSession();

    var descWidget = document.getElementById("runoutput-desc");
    //XXX Bug: This attribute change doesn't actually show up until the cursor
    //    is moved, triggering some gecko redraw. Cannot use
    //    window.sizeToContent() because that does bad (though non-fatal)
    //    things to komodo.xul.
    descWidget.setAttribute("value", "Running '" + command + "'...");
    // Store the command name for later use.
    descWidget.setAttribute("_command", command);
}


// Complete a terminal session. The command exited with the given value.
function RunOutput_EndSession(retval)
{
    //dump("XXX RunOutput_EndSession(retval="+retval+")\n");
    _gTerminal.endSession();
    
    var descWidget = document.getElementById("runoutput-desc");
    var command = descWidget.getAttribute("_command");
    //XXX Bug: This attribute change doesn't actually show up until the cursor
    //    is moved, triggering some gecko redraw. Cannot use
    //    window.sizeToContent() because that does bad (though non-fatal)
    //    things to komodo.xul.
    var msg = "'" + command + "' returned " + retval + ".";
    if (retval != 0) {
        addStyle(descWidget, "color", "#bb0000"); // dark red to not appear "neon" against grey.
    } else {
    }
    descWidget.setAttribute("value", msg);

    _gProcess = null;
    var closeButton = document.getElementById("runoutput-close-button");
    closeButton.setAttribute("disabled", "true");
}


function RunOutput_Show(editor)
{
    // open output pane in komodo.xul
    var splitterWidget = editor.document
                         .getElementById("redirect-splitter");
    splitterWidget.removeAttribute('state');
    
    // switch to the run output tab
    var tabsWidget = editor.document.getElementById("oTabbox");
    var tabWidget = editor.document.getElementById("runoutput_tab");
    tabsWidget.selectedTab = tabWidget; 
}


// Pass a koIRunProcess reference to the output window so it can manipulate the
// process that is being run in its terminal, if necessary.
function RunOutput_SetProcessHandle(process)
{
    if (_gTerminal.active) {
        _gProcess = process;
        var closeButton = document.getElementById("runoutput-close-button");
        closeButton.setAttribute("disabled", "false");
    }
}


// Kill the process currently running in the output window's terminal, if any.
function RunOutput_Kill(retval)
{
    if (_gProcess) {
        _gProcess.kill(retval);
    }
}

