/*** -*- Mode: Javascript; tab-width: 2;

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is Collabnet code.
The Initial Developer of the Original Code is Collabnet.

Portions created by Collabnet are
Copyright (C) 2000 Collabnet.  All
Rights Reserved.

Contributor(s): Pete Collins, Doug Turner, Brendan Eich, Warren Harris


*******
*
* JS File IO API (The purpose of this file is to make it a little easier to do file IO from js) 
*
*       io.js
*
* Function List
*
*       1. open(aPath, aMode);              // open a file handle for reading, writing or appending
*       2. exists(aPath);                   // check to see if a file exists
*       3. read();                          // returns the contents of a file
*       4. write(aContents, aPermissions);  // permissions are optional 
*       5. append(aDirPath, aFileName);     // append is for abstracting platform specific file paths
*       6. mkdir(aPath, aPermissions);      // permissions are optional
*       7. rmdir(aPath);                    // remove a directory
*       8. rm(aPath);                       // remove a file
*       9. copy(aSource, aDest);            // copy a file from source to destination
*       10.leaf(aPath);                     // leaf is the endmost file string ex: foo.html in /myDir/foo.html
*       11.readDir(aDirPath);               // returns an array listing of a dirs contents 
*       12.permissions(aPath);              // returns the files permissions
*       13.dateModified(aPath);             // returns the last modified date in locale string
*       14.size(aPath);                     // returns the file size
*       15.extension(aPath);                // returns a file extension if there is one
*       16.dirPath(aPath)                   // returns the dir part of a path
*       17.spawn(aPath, aArgs)              // spawns another program 
*       18.nsIFile(aPath)                   // returns an nsIFile obj 
*       19.fileURL(aFileURL)                // converts a `file:///fu/bar' string to filespec obj
*       20.mv(aSource, aDestDir, aNewName)  // moves aSource to aDestDir, aNewName is an optional arg 
*       21.help();                          // currently dumps a list of available functions 
*
*       Instructions:
*
*       First include this js file 
*
*       var file = new File();
*
*       to open a file for reading<"r">, writing<"w"> or appending<"a">, 
*       just call: 
*
*       file.open("/path/i/want/to/open", "w");
*
*       where in this case you will be creating a new file called:
*       '/path/i/want/to/open'
*       with a mode of "w" which means you want to write a new file.
*
*       If you want to read from a file, just call:
*
*       file.open("/path/i/want/to/open", "r");
*       
*       var theFilesContents    = file.read();
*
*       the file contents will be returned to the caller
*       so you can do something usefull with it
*
*       file.close(); 
*
*       destroys any created objects
*
*       if you forget to use file.close() no probs all objects are discarded anyway
*
*       Examples:
*
*         var path='/usr/X11R6/bin/Eterm';
*         file.dirPath(path);  
*         file.spawn(path, ['-e/usr/bin/vi']); 
*         *note* all args passed to spawn must be in the form of an array
*
*       Warning: these API's are not for religious types
*
*************/

/****************** Globals **********************/

const LOCAL_CID               = "@mozilla.org/file/local;1";
const F_CHANNEL_CID           = "@mozilla.org/network/local-file-channel;1";
const I_STREAM_CID            = "@mozilla.org/scriptableinputstream;1";
const NETWORK_STD_CID         = "@mozilla.org/network/standard-url;1";

const I_LOCAL_FILE            = "nsILocalFile";
const I_FILE_CHANNEL          = "nsIFileChannel";
const I_SCRIPTABLE_IN_STREAM  = "nsIScriptableInputStream";
const I_URL                                                                     = "nsIURL";

const INIT_W_PATH             = "initWithPath";

const FilePath                = new Components.Constructor( LOCAL_CID, I_LOCAL_FILE, INIT_W_PATH);
const FileChannel             = new Components.Constructor( F_CHANNEL_CID, I_FILE_CHANNEL );
const InputStream             = new Components.Constructor( I_STREAM_CID, I_SCRIPTABLE_IN_STREAM );

const READ                    = 0x01;     // 1
const WRITE                   = 0x08;     // 8
const APPEND                  = 0x10;     // 16

const FILE                    = 0x00;     // 0
const DIRECTORY               = 0x01;     // 1

const IO_OK                   = true;

/****************** Globals **********************/


/****************** File Object Class *********************/

function File(aPath, aMode) {
    
    if(aPath)
      this.open(aPath, aMode);

} // constructor

File.prototype  = {

mode            : null,

fileInst        : null,

fileChannel     : null,

inputStream     : null,

/********************* EXISTS ***************************/

exists : function (aPath) {

    if(!aPath)
    {
      dump("io.js:exists:ERROR: Missing path argument\n");
      return null;
    }

    var retval;
    try
    { 
      var file                = new FilePath(aPath);
      retval=file.exists();
    }

    catch(e) { 
      dump("io.js:exists:ERROR: "+e.name+"\n"+e.message+'\n'); 
      retval=null;
    }

  return retval;
},

/********************* EXISTS ***************************/


/********************* OPEN *****************************/


open : function(aPath, aSetMode) {

    if(!aSetMode)
      aSetMode="r";

    if(!aPath)
    {
      dump("io.js:open:ERROR: Missing path argument\n");
      return null;
    }

  var retval;

  switch(aSetMode)
  {

  case "w":

    try
    {
      this.close();
      this.fileInst           = new FilePath(aPath);
      var fileExists          = this.exists(aPath);

    if(fileExists)
    {

      if(this.fileInst.isDirectory())
      {
        dump("io.js:open:ERROR: Can't open for writing, file is a directory . . .\n");
        break;
      }

      else
      {
        this.fileInst["delete"](false);
        fileExists=false;
      }
    }

    if (!fileExists)
    {
      this.fileInst.create(FILE, 0644);
      fileExists=true;
    }

    this.mode=WRITE;
    retval=IO_OK;

    }

    catch (e)
    { 
      dump("io.js:open:ERROR: "+e.name+"\n"+e.message+'\n');
      return null;
    }

    break;

  case "a":

    try
    {
      this.close();
      this.fileInst           = new FilePath(aPath);
      fileExists              = this.exists(aPath);

      if(!fileExists)
      {
        this.fileInst.create(FILE, 0644);
        fileExists=true;
      }

      if(this.fileInst.isDirectory())
      {
        dump("io.js:open:ERROR: Can't open file for appending, file is a directory . . .\n");
        retval=null;
        break;
      }

      else
      {
        this.mode=APPEND;
        retval=IO_OK;
      }

    }

    catch (e)
    { 
      dump("io.js:open:ERROR: "+e.name+"\n"+e.message+'\n');
      return null;
    }

    break;


  case "r":

    this.close();

    try 
    {
      this.fileInst                = new FilePath(aPath);

      if(!this.exists(aPath))
      {
        dump("io.js:open:ERROR: Can't open file for reading, file doesn't exist . . .\n");
        retval=null;
        break;
      }

      if(this.fileInst.isDirectory())
      {
        dump("io.js:open:ERROR: Can't open file for reading, file is a directory. "   +
              "Use readDir() instead . . .\n");
        retval=null;
        break;
      }

      else
      {
        this.fileChannel         = new FileChannel();
        this.inputStream         = new InputStream();    
        retval=IO_OK;
      }

    }
        
    catch (e)
    { 
      dump("io.js:open:ERROR: "+e.name+"\n"+e.message+'\n');
      retval=null;
    }

    break;

    default:
    dump("io.js:open:WARNING: \""+setMode+"\" is an Invalid file mode\n");
    return null;

    }

  return retval;

},

/********************* OPEN *****************************/


/********************* WRITE ****************************/

write : function(aBuffer, aPerms) {

    if(!this.fileInst)
    {
      dump("io.js:write:ERROR: Please open a file handle first\n");
      return null;
    }
    
    if(!aBuffer)
      aBuffer                  = " ";

      var buffSize            = aBuffer.length;

    try
    {
      if(!this.fileChannel)
        this.fileChannel      = new FileChannel();

    if(aPerms)
    {
        var checkPerms        = this.validatePermissions(aPerms);

    if(!checkPerms){
      dump("io.js:write:ERROR: Sorry invalid permissions set\n");
      return null;
    }               

    }

    if(!aPerms)
      aPerms=0644;

    this.fileChannel.init(this.fileInst, this.mode, aPerms)
    try
    { 
      this.fileInst.permissions=aPerms;
    }

    catch(e){}

    var fileSize            = parseInt( this.fileInst.fileSize );

    var outStream           = this.fileChannel.openOutputStream();

    if( outStream.write(aBuffer, buffSize) )
      dump("io.js:write: Write to file successful . . . \n");
      outStream.flush();
      outStream.close();
    }

    catch (e)
    { 
      dump("io.js:write:ERROR: "+e.name+"\n"+e.message+'\n');
      return null;
    }

  return IO_OK;

},

/********************* WRITE ****************************/


/********************* READ *****************************/

read : function() {

    var retval;

    try 
    {

    if(!this.fileInst || !this.fileChannel || !this.inputStream)
    {
      dump("io.js:read:ERROR: Please open a valid file handle for reading\n");
      return null;
    }
        
    var fileExists          = this.exists(this.fileInst.path);

    if(!fileExists) 
    {
        dump("io.js:read:ERROR: \""+this.fileInst.path+"\" does not exist\n");
        return null;
    }

    var offset              = this.fileInst.fileSize;
    try
    {
      var perm                = this.fileInst.permissions;     
    }
    
    catch(e) { perm=null; }

    this.fileChannel.init(this.fileInst, READ, perm);

    var inStream = this.fileChannel.openInputStream();

    this.inputStream.init(inStream);

    var fileContents        = this.inputStream.read(this.fileInst.fileSize);

    inStream.close();
        
    retval=fileContents;

    }

    catch (e)
    { 
      dump("io.js:read:ERROR: "+e.name+"\n"+e.message+'\n');
      retval=null;
    }

    return retval;

},

/********************* READ *****************************/


/********************* MKDIR ****************************/

mkdir : function(aPath, aPermissions) {

  this.close();

  if(!aPath)
  {
    dump("io.js:rm:ERROR: Missing path argument\n");
    return null;
  }

  var fileExists    = this.exists(aPath);

  if(aPermissions){
    var checkPerms  = this.validatePermissions(aPermissions);

  if(!checkPerms){
    dump("io.js:mkdir:ERROR: Sorry invalid permissions set\n");
    return null;
  }               

  var baseTen       = permissions.toString(10);

  if(baseTen.substring(0,1) != 0)
    aPermissions = 0+baseTen;
  }

  var retval;
  try
  { 
    this.fileInst  = new FilePath(aPath);

    if (!fileExists)
    {
      if(!aPermissions)
        aPermissions     = 0755;

    this.fileInst.create( DIRECTORY, parseInt(aPermissions) ); 
    retval=IO_OK;

    }

    else 
      retval=false; 
  }

  catch (e)
  { 
    dump("io.js:mkdir:ERROR: "+e.name+"\n"+e.message+'\n');
    return null;
  }

  this.close();
  return retval;

},

/********************* MKDIR ****************************/


/********************* RM *******************************/

rm : function (aPath) {

    this.close();

    if(!aPath)
    {
      dump("io.js:rm:ERROR: Missing path argument\n");
      return null;
    }

    if(!this.exists(aPath))
    {
      dump("io.js:rm:ERROR: Sorry, file "+aPath+" doesn't exist\n");
      return null;
    }

    var retval;
    try
    { 

    this.fileInst            = new FilePath(aPath);

      if(this.fileInst.isDirectory())
      {
        dump("io.js:rm:ERROR:Sorry file is a directory. Try rmdir() instead\n");
        return null;
      }

    this.fileInst['delete'](false);
    this.close();
    retval=IO_OK;

    }

    catch (e)
    { 
      dump("io.js:rm:ERROR: "+e.name+"\n"+e.message+'\n');
      retval=null;
    }

  return retval;

},

/********************* RM *******************************/


/********************* RMDIR ****************************/

rmdir : function (aPath) {

  this.close();

  if(!aPath)
  {
    dump("io.js:rmdir:ERROR: Missing path argument\n");
    return null;
  }

  if(!this.exists(aPath))
  {
    dump("io.js:rmdir:ERROR: Sorry, file "+aPath+" doesn't exist\n");
    return null;
  }

  var retval;
  try
  { 
    this.fileInst            = new FilePath(aPath);
    this.fileInst['delete'](true);
    this.close();
    retval=IO_OK;
  }

  catch (e)
  { 
    dump("io.js:rmdir: ERROR: "+e.name+"\n"+e.message+'\n');
    retval=null;
  }

  return retval;

},

/********************* RMDIR ****************************/

/********************* COPY *****************************/

copy  : function (aSource, aDest) {

  if(!aSource || !aDest)
  {
    dump("io.js:copy:ERROR: Missing path argument\n");
    return null;
  }

  if(!this.exists(aSource))
  {
    dump("io.js:copy:ERROR: Sorry, source file "+source+" doesn't exist\n");
    return null;
  }

  var retval;
  try
  { 
    var fileInst      = new FilePath(aSource);
    var dir           = new FilePath(aDest);

    var copyName      = fileInst.leafName;

    if(fileInst.isDirectory())
    {
      dump("io.js:copy:File is a directory ignoring . . .\n");
      return null;
    }

    if(!this.exists(aDest) || !dir.isDirectory())
    {
      copyName          = dir.leafName;
      dir               = new FilePath(dir.path.replace(copyName,''));

      if(!this.exists(dir.path))
      {
        dump("io.js:copy:ERROR: Sorry, dest directory "+dir.path+" doesn't exist\n");
        return null;
      }

      if(!dir.isDirectory())
      {
        dump("io.js:copy:ERROR: Sorry, destination dir "+dir.path+" is not a valid dir path\n");
        return null;
      }

    }

    if(this.exists(this.append(dir.path, copyName)))
    {
      dump("io.js:copy:ERROR: Sorry destination file "+this.append(dir.path, copyName)+" already exists\n");
      return null;
    }

    fileInst.copyTo(dir, copyName);
    dump('io.js:copy successful!\n');
    this.close();
    retval=IO_OK;

  }

  catch (e)
  { 
    dump("io.js:copy:ERROR: "+e.name+"\n"+e.message+'\n');
    retval=null;
  }

  return retval;

},

/********************* COPY *****************************/


/********************* LEAF *****************************/

leaf  : function (aPath) {

  if(!aPath)
  {
    dump("io.js:leaf:ERROR: Missing path argument\n");
    return null;
  }

  if(!this.exists(aPath))
  {
    dump("io.js:leaf:ERROR: File doesn't exist\n");
    return null;
  }

  var retval;
  try
  {
    var fileInst = new FilePath(aPath);
    retval=fileInst.leafName;
  }

  catch(e)
  { 
    dump("io.js:leaf:ERROR: "+e.name+"\n"+e.message+'\n');
    retval=null;
  }
  
  return retval;

},

/********************* LEAF *****************************/


/********************* READDIR **************************/

readDir : function (aDirPath) {

  if(!aDirPath)
  {
    dump("io.js:readDir:ERROR: Missing path argument\n");
    return null;
  }

  if(!this.exists(aDirPath))
  {
    dump("io.js:readDir:ERROR: Sorry, directory "+dirPath+" doesn't exist\n");
    return null;
  }

  var retval;
  try
  {

    var file      = new FilePath(aDirPath);

    if(!file.isDirectory())
    {
      dump("io.js:readDir:ERROR: Sorry, "+aDirPath+" is not a directory\n");
      return null;
    }

    var files     = file.directoryEntries;
    var listings  = new Array();

    while(files.hasMoreElements())
    listings.push(files.getNext().QueryInterface(Components.interfaces.nsILocalFile).path);

    retval=listings;

  }

  catch(e)
  { 
    dump("io.js:readDir:ERROR: "+e.name+"\n"+e.message+'\n');
    retval=null;
  }
  
  return retval;

},

/********************* READDIR **************************/


/********************* APPEND ***************************/

append : function (aDirPath, aFileName) {

    if(!aDirPath || !aFileName)
    {
      dump("io.js:append:ERROR: Missing path argument\n");
      return null;
    }

    if(!this.exists(aDirPath))
    {
      dump("io.js:append:ERROR: File doesn't exist\n");
      return null;
    }

    var retval;
    try
    { 
      this.fileInst           = new FilePath(aDirPath);
      var fileAppended        = this.fileInst.append(aFileName);
      fileAppended            = this.fileInst.path;
      retval=fileAppended;
    }

    catch(e)
    { 
      dump("io.js:append:ERROR: "+e.name+"\n"+e.message+'\n');
      retval=null;
    }

    return retval;
},

/********************* APPEND ***************************/


/********************* VALIDATE PERMISSIONS *************/

validatePermissions : function(aNum) {

    if ( parseInt(aNum.toString(10).length) < 3 ) 
      return false;

    return true;

},

/********************* VALIDATE PERMISSIONS *************/


/********************* PERMISSIONS **********************/

permissions : function (aPath) {

    if(!aPath)
    {
      dump("io.js:permissions:ERROR: Missing path argument\n");
      return null;
    }
    
    if(!this.exists(aPath))
    {
      dump("io.js:permissions:ERROR: File doesn't exist\n");
      return null;
    }

    var retval;
    try
    { 
      var fileInst  = new FilePath(aPath); 
      retval=fileInst.permissions.toString(8);
    }

    catch(e)
    { 
      dump("io.js:permissions:ERROR: "+e.name+"\n"+e.message+'\n');
    retval=null;
    }

    return retval;

},

/********************* PERMISSIONS **********************/


/********************* MODIFIED *************************/

dateModified  : function (aPath) {

    if(!aPath)
    {
      dump("io.js:dateModified:ERROR: Missing path argument\n");
      return null;
    }
    
    if(!this.exists(aPath))
    {
      dump("io.js:dateModified:ERROR: File doesn't exist\n");
      return null;
    }

    var retval;
    try
    { 
      var fileInst  = new FilePath(aPath); 
      var date = new Date(fileInst.lastModificationDate).toLocaleString();
      retval=date;
    }

    catch(e)
    { 
      dump("io.js:dateModified:ERROR: "+e.name+"\n"+e.message+'\n');
      retval=null;
    }

    return retval;

},

/********************* MODIFIED *************************/


/********************* SIZE *****************************/

size  : function (aPath) {

    if(!aPath)
    {
      dump("io.js:size:ERROR: Missing path argument\n");
      return null;
    }
    
    if(!this.exists(aPath))
    {
      dump("io.js:size:ERROR: File doesn't exist\n");
      return null;
    }

    var retval;
    try
    { 
      var fileInst    = new FilePath(aPath); 
      retval          = fileInst.fileSize;
    }

    catch(e)
    { 
      dump("io.js:size:ERROR: "+e.name+"\n"+e.message+'\n');
      retval=0;
    }

    return retval;

},

/********************* SIZE *****************************/


/********************* EXTENSION ************************/

extension  : function (aPath) {

    if(!aPath)
    {
      dump("io.js:extension:ERROR: Missing path argument\n");
      return null;
    }
    
    if(!this.exists(aPath))
    {
      dump("io.js:extension:ERROR: File doesn't exist\n");
      return null;
    }

    var retval;
    try
    { 
      var fileInst  = new FilePath(aPath); 
      var leafName  = fileInst.leafName; 
      var dotIndex  = leafName.lastIndexOf('.'); 
      retval=(dotIndex >= 0) ? leafName.substring(dotIndex+1) : ""; 
    }

    catch(e)
    { 
      dump("io.js:extension:ERROR: "+e.name+"\n"+e.message+'\n');
      retval=null;
    }

    return retval;

},

/********************* EXTENSION ************************/

/********************* DIRPATH **************************/

dirPath   : function (aPath) {

    if(!aPath)
    {
      dump("io.js:dirPath:ERROR: Missing path argument\n");
      return null;
    }
    
    var retval;
    try
    { 

    var fileInst            = new FilePath(aPath);

      if(fileInst.isFile())
        retval=fileInst.path.replace(this.leaf(aPath), "");

      if(fileInst.isDirectory())
        retval=fileInst.path;

    }

    catch (e)
    { 
      dump("io.js:dirPath:ERROR: "+e.name+"\n"+e.message+'\n');
      retval=null;
    }

  return retval;

},

/********************* DIRPATH **************************/

/********************* SPAWN ****************************/

spawn : function (aPath, aArgs) {

    if(!aPath)
    {
      dump("io.js:spawn:ERROR: Missing path argument\n");
      return null;
    }
    
    if(!this.exists(aPath))
    {
      dump("io.js:spawn:ERROR: File doesn't exist\n");
      return null;
    }

    var len=0;

    if(aArgs)
      len = aArgs.length;

    else
      aArgs=null;
  
    var retval;
    try
    { 

    var fileInst            = new FilePath(aPath);

      if(!fileInst.isExecutable())
      {
        dump("io.js:spawn:ERROR: File is not executable\n");
        retval=null;
      }

      if(fileInst.isDirectory())
      {
        dump("io.js:spawn:ERROR: File is not a program\n");
        retval=null
      }

      else
        return fileInst.spawn(aArgs, len);

    }

    catch (e)
    { 
      dump("io.js:spawn:ERROR: "+e.name+"\n"+e.message+'\n');
      retval=null;
    }

  return retval;

},
/********************* SPAWN ****************************/

/********************* nsIFILE **************************/
nsIFile : function (aPath) {

    if(!aPath)
    {
      dump("io.js:nsIFile:ERROR: Missing path argument\n");
      return null;
    }
    
    var retval;
    try
    {
      retval = new FilePath(aPath);
    }

    catch (e)
    { 
      dump("io.js:nsIFile:ERROR: "+e.name+"\n"+e.message+'\n');
      retval=null;
    }

    return retval;

},
/********************* nsIFILE **************************/

/********************* fileURL **************************/
fileURL : function (aFileURL) {

    if(!aFileURL)
    {
      dump("io.js:nsIFile:ERROR: Missing path argument\n");
      return null;
    }
    
    var retval;
    try
    {
      var FileURL		= new  Components.Constructor(NETWORK_STD_CID, I_URL);
      var fileURL		= new FileURL();
      fileURL.spec	= aFileURL;
      retval 				= fileURL;
    }

    catch (e)
    { 
      dump("io.js:nsIFile:ERROR: "+e.name+"\n"+e.message+'\n');
      retval=null;
    }

    return retval;

},
/********************* fileURL **************************/

/********************* MV *******************************/
mv  : function (aSource, aDestDir, aNewName) {

  if(!aSource || !aDestDir)
  {
    dump("io.js:mv:ERROR: Missing path argument\n");
    return null;
  }

	if(!aNewName)
		aNewName=null;

	else
		if(this.exists(this.append(aDestDir, aNewName)))
  	{
    	dump("io.js:mv:ERROR: Destination already exists\n");
    	return null;
  	}

  if(!(this.exists(aSource) && this.exists(aDestDir)))
  {
    dump("io.js:mv:ERROR: Sorry, source file '"+aSource+"' or '"+aDestDir+"'  doesn't exist\n");
    return null;
  }

  var retval;
  try
  { 
    var fileInst      = new FilePath(aSource);
    var dir           = new FilePath(aDestDir);

		if(!dir.isDirectory())
  	{
    	dump("io.js:mv:ERROR: Sorry, destination directory "+aDestDir+" is not a directory\n");
    	return null;
  	}
			

    fileInst.moveTo(dir, aNewName);
    dump('io.js:move successful!\n');
    this.close();
    retval=IO_OK;

  }

  catch (e)
  { 
    dump("io.js:mv:ERROR: "+e.name+"\n"+e.message+'\n');
    retval=null;
  }

  return retval;

},
/********************* MV *******************************/

/********************* HELP *****************************/

help  : function() {

var help =

"\n\nFunction List:\n"                         +
"\n"                                           +
"*       1. open(aPath, aMode);\n"             +
"*       2. exists(aPath);\n"                  +
"*       3. read();\n"                         +
"*       4. write(aContents, aPermissions);\n" +
"*       5. append(aDirPath, aFileName);\n"    +
"*       6. mkdir(aPath, aPermissions);\n"     +
"*       7. rmdir(aPath);\n"                   +
"*       8. rm(aPath);\n"                      +
"*       9. copy(aSource, aDest);\n"           +
"*       10.leaf(aPath);\n"                    +
"*       11.readDir(aDirPath);\n"              +
"*       12.permissions(aPath);\n"             +
"*       13.dateModified(aPath);\n"            +
"*       14.size(aPath);\n"                    +
"*       15.extension(aPath);\n"               +
"*       16.dirPath(aPath);\n"                 + 
"*       17.spawn(aPath, aArgs);\n"            + 
"*       18.nsIFile(aPath);\n"                 + 
"*       19.fileURL(afileURL);\n"              + 
"*       20.mv(aSource, aDestDir, aNewName);\n"+ 
"*       21.help();\n";

dump(help+'\n');

return "";

},

/********************* HELP *****************************/


/********************* CLOSE ****************************/

close : function() {

    /***************** Destroy Instances *********************/
    if(this.fileInst)       delete this.fileInst;
    if(this.fileChannel)    delete this.fileChannel;
    if(this.inputStream)    delete this.inputStream;
    if(this.mode)           this.mode=null;
    /***************** Destroy Instances *********************/
  
  return IO_OK;

}

/********************* CLOSE ****************************/

};

/****************** End File Object Class *********************/


/************** Create an instance of file object *************/

//file          = new File(); // commenting out for now, the user can initialize the object . . . 

/************** Create an instance of file object *************/

