#!/usr/bin/env python
# Copyright (c) 2000-2001 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

import sys, os, re
import tempfile
from xpcom import components, ServerException, COMException, nsError
from koLogger import logError, logWarning, logDebug
import process
import koprocessutils


#---- components

class KoAppInfo:
    """Virtual application info base class."""
    def get_installed(self):
        """True if any version of the given application is locally installed."""
        return 0

    def _isLicensed(self, featureNames):
        """Use ActiveState's licensing mechanism to determine if something
        is licensed given a list of relevant feature names.
        """
        licSvc = components.classes["@activestate.com/koLicense;8"]\
                 .getService(components.interfaces.koILicense)
        # Arbitrarily pick 1 Jan 2001 for the "build date". The build date
        # here is irrelevant for normal products (as opposed to, say, ASPN
        # Tcl).
        rv, featureName, feature = licSvc.checkEx([featureName], 2001, 1, 1)
        return rv == licSvc.LIC_VALID
                
    def get_haveLicense(self):
        """True if user has license to run the current install."""
        return 0
    def get_version(self):
        """True if any version of the given application is locally installed."""
        return None
    def get_candebug(self):
        return 1
    # Indicate whether any help information is available and launch it.
    def get_localHelpFile(self):
        """Return a path to a launchable local help file, else return None."""
        return None
    def get_webHelpURL(self):
        """Return a web URL for help on this app, else return None."""
        return None


class KoPerlInfo(KoAppInfo):
    """Perl info.
    Not just *Active*Perl although some queries may only return successful
    results if the Perl in question *is* an ActivePerl.
    """
    _com_interfaces_ = [components.interfaces.koIAppInfo]
    _reg_clsid_ = "9b71cc84-0f0d-439a-8949-2a1d8a47be34"
    _reg_contractid_ = "@activestate.com/koPerlInfo;1"
    _reg_desc_ = "Perl Information"

    def __init__(self):
        self._sysUtils = components.classes["@activestate.com/koSysUtils;1"]\
                         .getService(components.interfaces.koISysUtils)

    def get_installed(self):
        """Return true if any Perl is installed.
        To tell if *Active*Perl is installed on Windows:
            buildNum = HKLM/Software/ActiveState/ActivePerl/CurrentVersion
            installDir = HKLM/Software/ActiveState/ActivePerl/$buildNum/(Default)
        however, `which perl` is more what we want.
        """
        return self._sysUtils.FastCheckIfHaveExecutable("perl")

    def get_haveLicense(self):
        return 1

    def get_localHelpFile(self):
        """Return a path to a launchable local help file, else return None.
        If there is an html/index.html in the install directory found
        via `which perl`. An *Active*Perl installation could be found via the
        registry on Windows (see get_installed()) since this is the only type
        of Perl installation likely to have the html/index.html subfile.
        However the current suffices.
        """
        perlExe = self._sysUtils.Which("perl")
        indexHtml = os.path.join(os.path.dirname(perlExe),
                                 "..", "html", "index.html")
        if os.path.isfile(indexHtml):
            return indexHtml
        else:
            return None

    def get_webHelpURL(self):
        """Return a web URL for help on this app, else return None."""
        return "http://aspn.activestate.com/ASPN/Products/ASPNTOC-ACTIVEPERL"

class KoPerlInfoEx:
    _com_interfaces_ = [components.interfaces.koIPerlInfoEx]
    _reg_clsid_ = "adb73505-eed5-46c5-8425-ce0bd8a5ec47"
    _reg_contractid_ = "@activestate.com/koPerlInfoEx;1"
    _reg_desc_ = "Extended Perl Information"
    
    def __init__(self):
        self._sysUtils = components.classes["@activestate.com/koSysUtils;1"]\
                         .getService(components.interfaces.koISysUtils)
        self.installationPath = None

    def _GetPerlExeName(self):
        assert self.installationPath is not None
        if sys.platform.startswith("win"):
            return os.path.join(self.installationPath, "bin", "perl.exe")
        else:
            return os.path.join(self.installationPath, "bin", "perl")

    # koIAppInfoEx routines
    def FindInstallationPaths(self):
        perlExes = self._sysUtils.WhichAll("perl")
        perlInstallationPaths = [self.getInstallationPathFromBinary(p)\
                                 for p in perlExes]
        return perlInstallationPaths

    def getInstallationPathFromBinary(self, binaryPath):
        return os.path.dirname(os.path.dirname(binaryPath))

    def get_executablePath(self):
        return self._GetPerlExeName()

    def get_haveLicense(self):
        return 1

    def get_version(self):
        perlExe = self._GetPerlExeName()
        if not os.path.exists(perlExe):
            raise ServerException(nsError.NS_ERROR_FILE_NOT_FOUND)
        argv = [perlExe, "-v"]
        p = process.ProcessOpen(argv)
        perlVersionDump = p.stdout.read()
        p.close()
        # Old perls look like: This is perl, version 5.005_03 built for MSWin32-x86-object
        # New perls look like: This is perl, v5.6.1 built for MSWin32-x86-multi-thread
        perlVersionMatch = re.search("This is perl, v(?:ersion )?([0-9._]+)", perlVersionDump)
        if perlVersionMatch:
            return perlVersionMatch.group(1)
        return ''

    def get_buildNumber(self):
        argv = [self.get_executablePath(), "-v"]
        p = process.ProcessOpen(argv)
        versionDump = p.stdout.read()
        p.close()
        pattern = re.compile("Binary build (\d+) provided by ActiveState")
        match = pattern.search(versionDump)
        if match:
            return int(match.group(1))
        else:
            # This is likely not an ActivePerl installation.
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)
 
    def get_localHelpFile(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def get_webHelpURL(self):
        return KoPerlInfo().get_webHelpURL()
    
    # koIPerlInfoEx routines
    def haveModules(self, modules):
        perlExe = self.get_executablePath()
        for module in modules:
            argv = [perlExe, "-M"+module, "-e1"]
            p = process.ProcessOpen(argv)
            retval = p.wait()
            p.close()
            if retval: # if returns non-zero, then don't have that module
                return 0
        else:
            return 1


class KoPythonInfo(KoAppInfo):
    """Python, not ActivePython, app info.

    Python, rather than ActivePython, is used because the functionality 
    being used here should be standard between Python and ActivePython
    installs or, at least, is standard for those portions of this interface
    that matter in that regard.
    """
    _com_interfaces_ = [components.interfaces.koIAppInfo]
    _reg_clsid_ = "e870b8e9-0c7b-460e-82e4-742ae4d66b43"
    _reg_contractid_ = "@activestate.com/koPythonInfo;1"
    _reg_desc_ = "Python Information"

    def __init__(self):
        self._sysUtils = components.classes["@activestate.com/koSysUtils;1"]\
                         .getService(components.interfaces.koISysUtils)

    def get_installed(self):
        """Return true if any Python is installed.
        Windows:
            Could use the registry value at:
            'HKLM/Software/Python/PythonCore/<major>.<minor>/InstallPath'
            but will use `which python` instead as this is the effective
            drop test to see if Python works on a machine.
        Linux:
            Installed if `which python` returns a result.
        """
        return self._sysUtils.FastCheckIfHaveExecutable("python")

    def get_haveLicense(self):
        return 1

    def get_localHelpFile(self):
        """Return a path to a launchable local help file, else return None.
        Windows:
            If there is a
            'HKLM/Software/Python/PythonCore/<major>.<minor>/Help/Main Python Documentation'
            and if the identified file exists.
        Linux/Solaris:
            If there is an html/index.html in the install directory found
            via `which python`.
        """
        if sys.platform.startswith("win"):
            import _winreg
            # get the base PythonCore registry key
            try:
                pythonCoreKey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                                                "SOFTWARE\\Python\\PythonCore")
            except EnvironmentError:
                return None
            # get a list of each installed version 
            versions = []
            index = 0
            while 1:
                try:
                    versions.append(_winreg.EnumKey(pythonCoreKey, index))
                except EnvironmentError:
                    break
                index += 1
            # try to find a existing help file (prefering the latest
            # installed version)
            versions.sort()
            versions.reverse()
            for version in versions:
                try:
                    helpFileKey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                        "SOFTWARE\\Python\\PythonCore\\%s\\Help\\Main "\
                        "Python Documentation" % version)
                    helpFile, keyType = _winreg.QueryValueEx(helpFileKey, "")
                    if os.path.isfile(helpFile):
                        return helpFile
                except EnvironmentError:
                    pass
            return None
        else:
            pythonExe = self._sysUtils.Which("python")
            indexHtml = os.path.join(os.path.dirname(pythonExe),
                                     "..", "html", "index.html")
            if os.path.isfile(indexHtml):
                return indexHtml
            else:
                return None

    def get_webHelpURL(self):
        """Return a web URL for help on this app, else return None."""
        return "http://aspn.activestate.com/ASPN/Products/ASPNTOC-ACTIVEPYT_"

class KoPythonInfoEx:
    _com_interfaces_ = [components.interfaces.koIAppInfoEx]
    _reg_clsid_ = "b76bc2ee-261e-4597-b1ef-446e9bb89d7c"
    _reg_contractid_ = "@activestate.com/koPythonInfoEx;1"
    _reg_desc_ = "Extended Python Information"
    
    def __init__(self):
        self._sysUtils = components.classes["@activestate.com/koSysUtils;1"]\
                         .getService(components.interfaces.koISysUtils)
        self.installationPath = None

    def _GetPythonExeName(self):
        assert self.installationPath is not None
        if sys.platform.startswith("win"):
            return os.path.join(self.installationPath, "python.exe")
        else:
           return os.path.join(self.installationPath, "bin", "python")

    # koIAppInfoEx routines
    def FindInstallationPaths(self):
        pythonExes = self._sysUtils.WhichAll("python")
        pythonInstallationPaths = [self.getInstallationPathFromBinary(p)\
                                   for p in pythonExes]
        return pythonInstallationPaths

    def getInstallationPathFromBinary(self, binaryPath):
        if sys.platform.startswith("win"):
            return os.path.dirname(binaryPath)
        else:
            return os.path.dirname(os.path.dirname(binaryPath))

    def get_executablePath(self):
        return self._GetPythonExeName()

    def get_haveLicense(self):
        return 1

    def get_version(self):
        # python 1.52 does not support the -V option
        versionScript = \
        r"import sys;"\
        r"import string;"\
        r"paren = string.find(sys.version, '(');"\
        r"version = string.strip(sys.version[:paren]);"\
        r"sys.stdout.write(version)"

        pythonExe = self._GetPythonExeName()
        pythonDir = os.path.dirname(pythonExe)
        argv = [pythonExe, "-c", versionScript]
        p = process.ProcessOpen(argv, cwd=pythonDir)
        versionDump = p.stdout.read()
        p.close()
        # we are only to rely on the first 2 digits being in the form x.y
        match = re.search("([0-9.]+)", versionDump)
        if match:
            return match.group(1)
        return ''

    def get_localHelpFile(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def get_webHelpURL(self):
        return KoPythonInfo().get_webHelpURL()


class KoTclInfo(KoAppInfo):
    _com_interfaces_ = [components.interfaces.koIAppInfo]
    _reg_clsid_ = "7c50de99-70c9-4da5-aef9-0b479cb21866"
    _reg_contractid_ = "@activestate.com/koTclInfo;1"
    _reg_desc_ = "Tcl Information"

    def __init__(self):
        self._sysUtils = components.classes["@activestate.com/koSysUtils;1"]\
                         .getService(components.interfaces.koISysUtils)

    def get_installed(self):
        """Return true if any Tcl is installed.
        The version of Tcl required by Komodo will have a "tclsh" (in addition
        to the previously lonely tclsh83.exe, or whatever).
        """
        return self._sysUtils.FastCheckIfHaveExecutable("tclsh")

    def get_haveLicense(self):
        return 1

    def get_localHelpFile(self):
        """Return a path to a launchable local help file, else return None.
        Windows:
            If there is a 'HKLM\Software\ActiveState\ActiveTcl\<CurVer>\Help'
            and if the identified file exists.
        Linux/Solaris:
            Nada. Just man files, which I don't consider "launchable" in a
            browser context. XXX Perhaps they *are* in Nautilus? 
        """
        if sys.platform.startswith("win"):
            import _winreg
            # get the base ActiveTcl registry key
            try:
                activeTclKey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                                               "SOFTWARE\\ActiveState\\ActiveTcl")
            except EnvironmentError:
                return None
            # get a list of each installed version 
            versions = []
            index = 0
            while 1:
                try:
                    versions.append(_winreg.EnumKey(activeTclKey, index))
                except EnvironmentError:
                    break
                index += 1
            # try to find a existing help file (prefering the latest
            # installed version)
            versions.sort()
            versions.reverse()
            for version in versions:
                try:
                    helpFileKey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                        "SOFTWARE\\ActiveState\\ActiveTcl\\%s\\Help" % version)
                    helpFile, keyType = _winreg.QueryValueEx(helpFileKey, "")
                    if os.path.isfile(helpFile):
                        return helpFile
                except EnvironmentError:
                    pass
            return None
        else:
            return None

    def get_webHelpURL(self):
        return "http://aspn.activestate.com/ASPN/Products/ASPNTOC-ACTIVETCL_"


class KoTclInfoEx:
    _com_interfaces_ = [components.interfaces.koITclInfoEx]
    _reg_clsid_ = "DF64A66F-FD69-4F5E-92B2-B3C9F8638F66"
    _reg_contractid_ = "@activestate.com/koTclInfoEx;1"
    _reg_desc_ = "Extended Tcl Information"

    def __init__(self):
        self.installationPath = None
    
    def get_executablePath(self):
        #XXX For 'tclsh' or 'wish'. This attribute is ambiguous for Tcl.
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def _getTclshExeName(self):
        if sys.platform.startswith('win'):
            return 'tclsh.exe'
        else:
            return 'tclsh'

    def _getWishExeName(self):
        if sys.platform.startswith('win'):
            return 'wish.exe'
        else:
            return 'wish'

    # koIAppInfoEx routines
    def FindInstallationPaths(self):
        tclshs = self._sysUtils.WhichAll("tclsh")
        installPaths = [self.getInstallationPathFromBinary(tclsh)\
                        for tclsh in tclshs]
        return installPaths

    def get_haveLicense(self):
        return 1

    def get_version(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
 
    def get_buildNumber(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
 
    def get_localHelpFile(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def get_webHelpURL(self):
        return KoTclInfo().get_webHelpURL()

    def getInstallationPathFromBinary(self, binaryPath):
        return os.path.dirname(os.path.dirname(binaryPath))

    # koITclInfoEx routines
    def selectDefault(self):
        paths = self.FindInstallationPaths()
        if paths:
            self.installationPath = paths[0]
            return 1
        else:
            self.installationPath = None
            return 0

    def get_tclsh_path(self):
        return os.path.join(self.installationPath, "bin",
                            self._getTclshExeName())
        
    def get_wish_path(self):
        return os.path.join(self.installationPath, "bin",
                            self._getWishExeName())



class KoTclProInfo(KoTclInfo):
    _com_interfaces_ = [components.interfaces.koIAppInfo]
    _reg_clsid_ = "3ee7b744-2ecf-49c2-b5ee-da871c89dca8"
    _reg_contractid_ = "@activestate.com/koTclProInfo;1"
    _reg_desc_ = "TclPro Information"

    def __init__(self):
        KoTclInfo.__init__(self)

    def _getProchecks(self):
        """Return a list of paths to available procheck executables.

        On Linux this is easy: procheck will be beside tclsh which must
        be on the PATH for normal operation. On Windows tclsh.exe is not
        actually put on the PATH, it is registered via "App Paths" in the
        registry, therefore have to look for tclsh.exe and then look for
        procheck.exe in the same dir.
        """
        if sys.platform.startswith("win"):
            tclshs = self._sysUtils.WhichAll("tclsh")
            prochecks = []
            for tclsh in tclshs:
                procheck = os.path.join(os.path.dirname(tclsh),
                                        "procheck.exe")
                if os.path.exists(procheck):
                    prochecks.append(procheck)
        else:
            prochecks = self._sysUtils.WhichAll("procheck")
        return prochecks

    def get_installed(self):
        # Search for "procheck", which is a part of TclPro but not
        # vanilla Tcl.
        prochecks = self._getProchecks()
        if prochecks:
            return 1
        else:
            return 0

    def get_haveLicense(self):
        """Return true iff TclPro is licensed.
        
        See if any available 'procheck' executables are licensed.
        If licensed, launching 'procheck' will write this on stdout (after
        stdin is closed):
            scanning: stdin
            checking: stdin
        and return 0. If not licensed, launching 'procheck' will return
        non-zero and will include the word 'license' in the first line
        of output on stderr, e.g.:
            application-specific initialization failed: Your license file does not contain the TclPro2 feature.
            Regular TclPro not found
            Please purchase a license from http://www.ActiveState.com/Products/ASPN_Tcl
            <snip>
        This relies on the unlicensed semantics not changing for all versions
        of TclPro/TDK.
        """
        prochecks = self._getProchecks()
        for procheck in prochecks:
            p = process.ProcessOpen(procheck)
            p.stdin.close()
            retval = p.wait()
            error = p.stderr.readlines()
            p.close()
            if retval and error[0].find("license") != -1:
                continue
            else:
                return 1
        else:
            return 0


class KoTclProInfoEx:
    _com_interfaces_ = [components.interfaces.koITclInfoEx]
    _reg_clsid_ = "0a606abf-bd1e-4220-9491-c9c3b95da7ea"
    _reg_contractid_ = "@activestate.com/koTclProInfoEx;1"
    _reg_desc_ = "Extended TclPro Information"
    
    def __init__(self):
        self._sysUtils = components.classes["@activestate.com/koSysUtils;1"]\
                         .getService(components.interfaces.koISysUtils)
        self.installationPath = None
    
    def get_executablePath(self):
        #XXX For 'tclsh' or 'wish'. This attribute is ambiguous for Tcl.
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def _getTclshExeName(self):
        if sys.platform.startswith('win'):
            return 'tclsh.exe'
        else:
            return 'tclsh'

    def _getWishExeName(self):
        if sys.platform.startswith('win'):
            return 'wish.exe'
        else:
            return 'wish'

    def _getProcheckName(self):
        if sys.platform.startswith('win'):
            return 'procheck.exe'
        else:
            return 'procheck'

    def _getProcheckPath(self, installationPath):
        return os.path.join(installationPath, "bin",
                            self._getProcheckName())

    def _getProchecks(self):
        """Return a list of paths to available procheck executables.

        On Linux this is easy: procheck will be beside tclsh which must
        be on the PATH for normal operation. On Windows tclsh.exe is not
        actually put on the PATH, it is registered via "App Paths" in the
        registry, therefore have to look for tclsh.exe and then look for
        procheck.exe in the same dir.
        """
        if sys.platform.startswith("win"):
            tclshs = self._sysUtils.WhichAll("tclsh")
            prochecks = []
            for tclsh in tclshs:
                procheck = os.path.join(os.path.dirname(tclsh),
                                        self._getProcheckName())
                if os.path.exists(procheck) and procheck not in prochecks:
                    prochecks.append(procheck)
        else:
            prochecks = self._sysUtils.WhichAll(self._getProcheckName())
        return prochecks

    def _isInstallationLicensed(self, installationPath):
        """Return true iff the given TclPro installation is licensed.
        
        See if the 'procheck' executable in this installation is licensed.
        If licensed, launching 'procheck' will write this on stdout (after
        stdin is closed):
            scanning: stdin
            checking: stdin
        and return 0. If not licensed, launching 'procheck' will return
        non-zero and will include the word 'license' in the first line
        of output on stderr, e.g.:
            application-specific initialization failed: Your license file does not contain the TclPro2 feature.
            Regular TclPro not found
            Please purchase a license from http://www.ActiveState.com/Products/ASPN_Tcl
            <snip>
        This relies on the unlicensed semantics not changing for all versions
        of TclPro/TDK.
        """
        procheck = self._getProcheckPath(installationPath)
        if not os.path.exists(procheck):
            return 0
        p = process.ProcessOpen(procheck)
        p.stdin.close()
        retval = p.wait()
        error = p.stderr.readlines()
        p.close()
        if retval and error[0].find("license") != -1:
            return 0
        else:
            return 1

    # koIAppInfoEx routines
    def FindInstallationPaths(self):
        return [self.getInstallationPathFromBinary(p)\
                for p in self._getProchecks()]

    def get_haveLicense(self):
        return self._isInstallationLicensed(self.installationPath)

    def get_version(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
 
    def get_localHelpFile(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def get_webHelpURL(self):
        return KoTclInfo().get_webHelpURL()

    def getInstallationPathFromBinary(self, binaryPath):
        return os.path.dirname(os.path.dirname(binaryPath))

    def selectDefault(self):
        paths = self.FindInstallationPaths()
        
        # Preferentially use a licensed TclPro
        for installationPath in paths: 
            if self._isInstallationLicensed(installationPath):
                self.installationPath = installationPath
                return 1
        else:
            # Otherwise use whatever is left
            if paths:
                self.installationPath = paths[0]
                return 1
            else:
                self.installationPath = None
                return 0

    def get_tclsh_path(self):
        return os.path.join(self.installationPath, "bin",
                            self._getTclshExeName())
        
    def get_wish_path(self):
        return os.path.join(self.installationPath, "bin",
                            self._getWishExeName())


class KoPHPInfo(KoAppInfo):
    _com_interfaces_ = [components.interfaces.koIAppInfo]
    _reg_clsid_ = "218ef59d-b84e-43bf-a444-536782686b65"
    _reg_contractid_ = "@activestate.com/koPHPInfo;1"
    _reg_desc_ = "PHP Information"

    def __init__(self):
        self._sysUtils = components.classes["@activestate.com/koSysUtils;1"]\
                         .getService(components.interfaces.koISysUtils)
        self.prefService = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService)

    def get_installed(self):
        """Return true if any PHP is installed."""
        return self._GetPHPExeName() != None

    def get_candebug(self):
        return self._DebugExtensionLoadable();

    def get_haveLicense(self):
        return 1

    def get_localHelpFile(self):
        """Return a path to a launchable local help file, else return None.
        Nada for PHP. There is no *standard* local documentation link or any
        real de facto standard.
        """
        return None

    def get_webHelpURL(self):
        return "http://aspn.ActiveState.com/ASPN/PHP/Reference"

    def _getInterpreterConfig(self):
        if self.prefService.prefs.hasStringPref("phpConfigFile") and\
           self.prefService.prefs.getStringPref("phpConfigFile"):
            iniFile = self.prefService.prefs.getStringPref("phpConfigFile")
        else:        
            iniFile = None
        return iniFile

    def _GetPHPExeName(self):
        if self.prefService.prefs.hasStringPref("phpDefaultInterpreter") and\
           self.prefService.prefs.getStringPref("phpDefaultInterpreter"):
            php = self.prefService.prefs.getStringPref("phpDefaultInterpreter")

            #XXX Is this still necessary now that we are using process.py?
            #    It existed because popen2 cannot handle unicode paths.
            try:
                php = str(php)
            except UnicodeError:
                raise ServerException, (nsError.NS_ERROR_NOT_AVAILABLE,
                        "Path to PHP interpreter cannot contain Unicode characters.") 
        else:
            php = self._sysUtils.Which("php")
        return php or None

    def _GetPHPOutput(self, phpCode):
        php = self._GetPHPExeName()
        if not php:
            return None #XXX Would be better, IMO, to raise an exception here.
        env = koprocessutils.getUserEnv()
        ini = self._getInterpreterConfig()
        if ini:
            env["PHPRC"] = ini
        argv = [php, '-q']
        # PHP sets stdin, stdout to binary on it's end.  If we do not do
        # the same, some wierd things can happen, most notably with shell
        # comments that end with \r\n.
        p = process.ProcessOpen(argv, mode='b', env=env)
        p.stdin.write(phpCode)
        p.stdin.close()
        output = p.stdout.read()
        # For some reason, PHP linter causes an exception on close
        # with errno = 0, will investigate in PHP later.
        try:
            p.close()
        except IOError, e:
            if e.errno == 0:  
                pass
            else:
                raise

        return output.strip()

    def _DebugExtensionLoadable(self):
        return self._GetPHPOutput("<?php echo extension_loaded('Active Debugger')?'Yes':'No'; ?>") == "Yes"


class KoPDKInfo(KoAppInfo):
    """PDK info."""
    _com_interfaces_ = [components.interfaces.koIAppInfo]
    _reg_clsid_ = "96a12738-d862-45cd-af3e-6ca180daca91"
    _reg_contractid_ = "@activestate.com/koPDKInfo;1"
    _reg_desc_ = "PDK Information"

    def __init__(self):
        self._sysUtils = components.classes["@activestate.com/koSysUtils;1"]\
                         .getService(components.interfaces.koISysUtils)

    def get_installed(self):
        """Return true if the PDK is installed."""
        return self._sysUtils.FastCheckIfHaveExecutable("perlapp")

    def get_haveLicense(self):
        """Return true if the PDK is licensed.
        
        Let PDK tell us if it is licensed by running a contrived perlapp
        build and seeing if it works. If the PDK is not licensed it will
        spit out an error on the first line of stderr mentioning license.
        For example:
            You don't seem to have a valid license.
            Please purchase a license from http://www.ActiveState.com/PDK
        
        Note: this can be _slow_.
        """
        try:
            # Setup a contrived perlapp build.
            scriptName = tempfile.mktemp('LC.pl')
            exeName = tempfile.mktemp('LC.exe')
            scriptFile = open(scriptName, 'w+')
            scriptFile.write(' ') # File cannot be completely empty
            scriptFile.close()
    
            # Run 'perlapp'.
            # '-f' is "required". I don't know why. -- TM
            #XXX Will we run into spaces-in-path problems here?
            argv = ['perlapp', '-f', scriptName, '-exe=%s' % exeName]
            try:
                p = process.ProcessOpen(argv)
                error = p.stderr.read()
                p.close()
            except process.ProcessError, ex:
                return 0

        finally:
            # Cleanup.
            try:
                if os.path.exists(scriptName):
                    os.remove(scriptName)
                if os.path.exists(exeName):
                    os.remove(exeName)
            except IOError:
                pass

        if not error:
            return 1
        else:
            firstLine = error.split('\n')[0]
            return firstLine.find('license') == -1
    
    def get_version(self):
        """Return the version of the installed PDK, None if not installed.
        PerlApp/Ctrl/Svc all carry the same version (that of the PDK).
        """
        perlapp = self._sysUtils.Which("perlapp")
        if not perlapp:
            return None
        else:
            argv = [perlapp, '-h']
            p = process.ProcessOpen(argv)
            perlappVersionDump = p.stdout.read()
            p.close()

            perlappVersionMatch = re.search(
                "\(PerlApp version: (?P<version>[0-9.]+)\)",
                perlappVersionDump)
            if perlappVersionMatch:
                return perlappVersionMatch.group("version")
            # PerlApp 4.0.0 build 401
            perlappVersionMatch = re.search(
                "PerlApp (?P<version>[0-9.]+)",
                perlappVersionDump)
            if perlappVersionMatch:
                return perlappVersionMatch.group("version")
            else:
                raise ServerException(nsError.NS_ERROR_UNEXPECTED)        

    def get_localHelpFile(self):
        """Return a path to a launchable local help file, else return None.
        Both a CHM and HTML are installed on Windows. Prefer the CHM.
        """
        perlappExe = self._sysUtils.Which("perlapp")
        if perlappExe == "":
            return None
        elif sys.platform.startswith("win"):
            helpFile = os.path.join(os.path.dirname(perlappExe),
                                    "..", "Docs", "PDK", "PDKUserGuide.chm")
        else:
            helpFile = os.path.join(os.path.dirname(perlappExe),
                                    "..", "Docs", "PDK", "HTML", "index.html")
        if os.path.isfile(helpFile):
            return helpFile
        else:
            return None

    def get_webHelpURL(self):
        """Return a web URL for help on this app, else return None."""
        return "http://aspn.activestate.com/ASPN/Products/ASPNTOC-PERLDEVKI_"


class KoPHPInfoEx:
    _com_interfaces_ = [components.interfaces.koIPHPInfoEx]
    _reg_clsid_ = "ea1519a8-4e4d-4767-aec4-2f0342c33e7a"
    _reg_contractid_ = "@activestate.com/koPHPInfoEx;1"
    _reg_desc_ = "PHP Information"

    def __init__(self):
        self._sysUtils = components.classes["@activestate.com/koSysUtils;1"]\
                         .getService(components.interfaces.koISysUtils)
        self.prefService = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService)
        self.installationPath = None

    def _GetPHPExeName(self):
        assert self.installationPath is not None
        if sys.platform.startswith("win"):
            return os.path.join(self.installationPath, "php.exe")
        else:
            return os.path.join(self.installationPath, "bin", "php")

    def _getInterpreterConfig(self):
        if self.prefService.prefs.hasStringPref("phpConfigFile") and\
           self.prefService.prefs.getStringPref("phpConfigFile"):
            iniFile = self.prefService.prefs.getStringPref("phpConfigFile")
        else:        
            iniFile = None
        return iniFile

    def _GetPHPOutput(self, phpCode):
        php = self._GetPHPExeName()
        if not php:
            return None #XXX Would be better, IMO, to raise an exception here.
        env = koprocessutils.getUserEnv()
        ini = self._getInterpreterConfig()
        if ini:
            env["PHPRC"] = ini
        argv = [php, '-q']
        # PHP sets stdin, stdout to binary on it's end.  If we do not do
        # the same, some wierd things can happen, most notably with shell
        # comments that end with \r\n.
        p = process.ProcessOpen(argv, mode='b', env=env)
        p.stdin.write(phpCode)
        p.stdin.close()
        output = p.stdout.read()
        # For some reason, PHP linter causes an exception on close
        # with errno = 0, will investigate in PHP later.
        try:
            p.close()
        except IOError, e:
            if e.errno == 0:  
                pass
            else:
                raise

        return output.strip()

    def _GetPHPConfigVar(self, varName):
        return self._GetPHPOutput("<?php echo(get_cfg_var('%s')); ?>"\
                                  % varName)
    
    def _GetPHPIniVar(self, varName):
        return self._GetPHPOutput("<?php echo(ini_get('%s')); ?>"\
                                  % varName)

    # koIAppInfoEx routines
    def FindInstallationPaths(self):
        phpExes = self._sysUtils.WhichAll("php")
        phpInstallationPaths = [self.getInstallationPathFromBinary(p)\
                                   for p in phpExes]            
        return phpInstallationPaths

    def getInstallationPathFromBinary(self, binaryPath):
        if sys.platform.startswith("win"):
            return os.path.dirname(binaryPath)
        else:
            return os.path.dirname(os.path.dirname(binaryPath))

    def get_executablePath(self):
        return self._GetPHPExeName()

    def get_haveLicense(self):
        return 1

    def get_version(self):
        return self._GetPHPOutput("<?php echo(phpversion()); ?>")

    def get_localHelpFile(self):
        """Return a path to a launchable local help file, else return None.
        Nada for PHP. There is no *standard* local documentation link or any
        real de facto standard.
        """
        return None

    def get_webHelpURL(self):
        return "http://aspn.ActiveState.com/ASPN/PHP/Reference"

    # additional koIPHPInfoEx routines
    def get_cfg_file_path(self):
        return self._GetPHPConfigVar("cfg_file_path")
    
    def get_include_path(self):
        return self._GetPHPIniVar("include_path")
    
    def GetIncludePathArray(self):
        includePath = self.get_include_path().split(os.pathsep)
        # cull out any empty entries (resulting from, say, include_path="a;;b")
        includePath = [path for path in includePath if path]
        return includePath
    
    def get_extension_dir(self):
        return self._GetPHPIniVar("extension_dir")

    def get_isDebuggerExtensionLoadable(self):
        return self._GetPHPOutput("<?php echo extension_loaded('Active Debugger')?'Yes':'No'; ?>") == "Yes"


class KoCVSInfoEx:
    _com_interfaces_ = [components.interfaces.koIAppInfoEx]
    _reg_clsid_ = "C3A7A887-D0D3-426A-8C67-2CC3E2946636"
    _reg_contractid_ = "@activestate.com/koCVSInfoEx;1"
    _reg_desc_ = "CVS Information"

    def __init__(self):
        self._sysUtilsSvc = components.classes["@activestate.com/koSysUtils;1"]\
                            .getService(components.interfaces.koISysUtils)
        self.installationPath = None

    def _getCVSExeName(self):
        if sys.platform.startswith('win'):
            return 'cvs.exe'
        else:
            return 'cvs'

    # koIAppInfoEx routines
    def FindInstallationPaths(self):
        cvss = self._sysUtils.WhichAll("cvs")
        cvsInstallationPaths = [self.getInstallationPathFromBinary(cvs)\
                                for cvs in cvss]
        return cvsInstallationPaths

    def getInstallationPathFromBinary(self, binaryPath):
        if sys.platform.startswith("win"):
            return os.path.dirname(binaryPath)
        else:
            return os.path.dirname(os.path.dirname(binaryPath))

    def get_executablePath(self):
        if sys.platform.startswith("win"):
            return os.path.join(self.installationPath, self._getCVSExeName())
        else:
            return os.path.join(self.installationPath, "bin",
                                self._getCVSExeName())

    def get_haveLicense(self):
        return 1

    def get_version(self):
        """A CVS version include not only the standard 1.2.3-type numbers
        but also the "build family", of which CVSNT is a different one.
        For example:
            1.11.2 CVS
            1.11.1.3 CVSNT
        Returns None if the version cannot be determined.
        """
        cvsExe = self.get_executablePath()
        p = process.ProcessOpen([cvsExe, '-v'])
        p.stdin.close()
        output = p.stdout.read()
        error = p.stderr.read()
        retval = p.wait()
        p.close()
        
        versionRe = re.compile(r'Concurrent Versions System '\
                                '\((?P<family>.+?)\) '\
                                '(?P<version>[\d\.]+?) ', re.MULTILINE)
        match = versionRe.search(output)
        if match:
            version = "%s %s" % (match.group('version'),
                                 match.group('family'))
            return version
        else:
            return None
 
    def get_buildNumber(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def get_localHelpFile(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)
    def get_webHelpURL(self):
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)



#---- self test code

if __name__ == "__main__":
    def getCOMAttribute(obj, property, default = 'not implemented'):
        try:
            return getattr(obj, property)
        except COMException, e:
            if e.errno == nsError.NS_ERROR_NOT_IMPLEMENTED:
                return default
            else:
                raise
                    
    print "TEST: koIAppInfo implementations"
    for app in ["PDK", "Perl", "Python", "PHP", "Tcl", "TclPro"]:
        try:
            appInfo = components.classes["@activestate.com/ko%sInfo;1"%app]\
                       .createInstance()
            print "%s installed?: %s" % (app, getCOMAttribute(appInfo, 'installed'))
            if appInfo.installed:
                print "%s version: %s" % (app, getCOMAttribute(appInfo, 'version'))
                print "%s licensed?: %s" % (app, getCOMAttribute(appInfo, 'haveLicense'))
            print "%s Local Help File: %s" % (app, getCOMAttribute(appInfo, 'localHelpFile'))
            print "%s Web Help URL: %s" % (app, getCOMAttribute(appInfo, 'webHelpURL'))
            print "%s Can Debug: %s" % (app, getCOMAttribute(appInfo, 'candebug'))
            print
        except Exception, e:
            print 'Exception encountered during %r test - continuing' % (app)
    print "TEST: koIAppInfoEx implementations"

    for app in ["Perl", "Python", "PHP", "TclPro"]:
        appInfoExe = components.classes["@activestate.com/ko%sInfoEx;1"%app]\
                   .createInstance()
        
        installations = appInfoExe.FindInstallationPaths()
        for installation in installations:
            appInfoExe.installationPath = installation
            print "+------ %s installation: %s" % (app, appInfoExe.installationPath)
            print "| haveLicense: %s" % getCOMAttribute(appInfoExe, 'haveLicense')
            print "| executable location: %s" % getCOMAttribute(appInfoExe, 'executablePath')
            print "| version: %s" % getCOMAttribute(appInfoExe, 'version')
            print "| localHelpFile: %s" % getCOMAttribute(appInfoExe, 'localHelpFile')
            print "| webHelpURL: %s" % getCOMAttribute(appInfoExe, 'webHelpURL')
            
            if app == "TclPro":
                print "|\t+------ TclPro extra features ------"
                print "|\t| tclsh_path: %s" % getCOMAttribute(appInfoExe, 'tclsh_path')
                print "|\t| wish_path: %s" % getCOMAttribute(appInfoExe, 'wish_path')
                print "|\t+---------------------------------"
            elif app == "PHP":
                print "|\t+------ PHP extra features ------"
                print "|\t| cfg_file_path: %s" % appInfoExe.cfg_file_path
                print "|\t| include_path: %s %s" % (appInfoExe.include_path,
                                                    appInfoExe.GetIncludePathArray())
                print "|\t| extension_dir: %s" % appInfoExe.extension_dir
                print "|\t+---------------------------------"
            
            print "+---------------------------------"
    


