#!/usr/bin/env python
# Copyright (c) 2000-2002 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

"""
Komodo Commandment system service
"""

import os
import sys
import threading
import re
if sys.platform.startswith("win"):
    import glob
    import win32api
    import win32con
    import win32event
    import win32gui
    from win32com.shell import shellcon, shell
else:
    import fcntl

from xpcom import components, nsError, ServerException, COMException
from koLogger import logError, logWarning, logDebug
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject


#---- globals

_gObserverSvc = None
if sys.platform.startswith("win"):
    # A handle to the Komodo main window for use by some of the commandments.
    _gHWnd = None



#---- support routines

def _handleCommandment(commandment):
    # Parse the commandment string.
    parts = commandment.split('\t')
    name, args = parts[0], parts[1:]

    if name == "open":
        filename = args[0]

        if sys.platform.startswith("win"):
            try:
                win32gui.SetForegroundWindow(_gHWnd)
            except win32api.error, ex:
                # XXX This sporadically fails and I don't know why:
                # api_error: (0, 'SetForegroundWindow', 'No error message is available')
                pass

        # Do file globbing on windows if necessary.
        if sys.platform.startswith("win")\
           and ('*' in filename or '?' in filename):
            filenames = glob.glob(filename)
        else:
            filenames = [filename]

        for f in filenames:
            try:
                _gObserverSvc.Notify(None, "open_file", f)
            except COMException, e:
                logWarning("No one observing 'open_file' notification.")

    else:
        logError("unknown commandment name: '%s'" % name)


#---- main commandment reader thread
if sys.platform.startswith("win"):
    class _CommandmentsReader(threading.Thread):
        """Consume commandments from the commandments.txt file.
        
        The .txt file must by guarded by the appropriate mutex and an
        event is signalled when new commandments are added.
        """
        def __init__(self, commandmentsFileName, commandmentsEventName,
                     commandmentsLockName):
            self._commandmentsFileName = commandmentsFileName
            self._commandmentsEventName = commandmentsEventName
            self._commandmentsLockName = commandmentsLockName
            
            threading.Thread.__init__(self)

            if os.path.exists(self._commandmentsFileName):
                os.unlink(self._commandmentsFileName)

        def run(self):
            lock = win32event.CreateMutex(None, 0, self._commandmentsLockName)
            newCommandments = win32event.CreateEvent(None, 1, 0,
                self._commandmentsEventName)

            while 1:
                # Wait for new commandments.
                rv = win32event.WaitForSingleObject(newCommandments,
                                                    win32event.INFINITE)
                if rv == win32event.WAIT_OBJECT_0:
                    retval = 1
                else:
                    raise "XXX Error waiting for new commandments: %r" % rv
                # Grab the lock.
                win32event.WaitForSingleObject(lock, win32event.INFINITE)
                # Consume the commandments.
                f = open(self._commandmentsFileName, 'r')
                cmds = []
                for line in f.readlines():
                    if line.endswith('\n'):
                        cmds.append( line[:-1] )
                f.close()
                os.unlink(self._commandmentsFileName)
                # Reset the "new commandments" event.
                win32event.ResetEvent(newCommandments)
                # Release the lock.
                win32event.ReleaseMutex(lock)
                # Handle the commandments.
                exit = 0
                for cmd in cmds:
                    if cmd == "__exit__":
                        exit = 1
                        break
                    else:
                        _handleCommandment(cmd)
                if exit:
                    break

            win32api.CloseHandle(newCommandments)
            win32api.CloseHandle(lock)

        def exit(self):
            # Grab the lock.
            lock = win32event.CreateMutex(None, 0, self._commandmentsLockName)
            win32event.WaitForSingleObject(lock, win32event.INFINITE)
            # Send __exit__ commandment.
            f = open(self._commandmentsFileName, 'a')
            f.write("__exit__\n")
            f.close()
            # Signal hat there are new commandments.
            newCommandments = win32event.CreateEvent(None, 1, 0,
                self._commandmentsEventName)
            win32event.SetEvent(newCommandments)
            win32api.CloseHandle(newCommandments)
            # Release the lock.
            win32event.ReleaseMutex(lock)
            win32api.CloseHandle(lock)

            self.join()

else:
    class _CommandmentsReader(threading.Thread):
        def __init__(self, commandmentsFileName):
            self._commandmentsFileName = commandmentsFileName
            
            threading.Thread.__init__(self)

            if os.path.exists(self._commandmentsFileName):
                os.unlink(self._commandmentsFileName)
            os.mkfifo(self._commandmentsFileName)
            self._pipe = os.open(self._commandmentsFileName, os.O_RDWR)
    
        def run(self):
            buf = ""
            while 1:
                text = os.read(self._pipe, 4096)
                buf += text
                lines = buf.splitlines(1)
                # parse out commands
                cmds = []
                for line in buf.splitlines(1):
                    if line.endswith('\n'):
                        cmds.append( line[:-1] )
                    else:
                        buf = line
                        break
                else:
                    buf = ""
                # handle each command
                exit = 0
                for cmd in cmds:
                    if cmd == "__exit__":
                        exit = 1
                        break
                    else:
                        _handleCommandment(cmd)
                if exit:
                    break
    
        def exit(self):
            os.write(self._pipe, "__exit__\n")
            self.join()
            os.close(self._pipe)
            os.unlink(self._commandmentsFileName)


#---- component implementation
class KoCommandmentService:
    _com_interfaces_ = [components.interfaces.koICommandmentService]
    _reg_clsid_ = "{1ADBCE86-1AE4-4F90-AA85-B230B7510D7B}"
    _reg_contractid_ = "@activestate.com/koCommandmentService;1"
    _reg_desc_ = "Komodo commandment system service"

    def __init__(self):
        # Platform-specific handle on object indicating Komodo is running.
        self._running = None
        # Commandment reader thread.
        self._reader = None

    def initialize(self):
        global _gObserverSvc
        proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"]\
            .getService(components.interfaces.nsIProxyObjectManager)
        observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        _gObserverSvc = getProxyForObject(None,
            components.interfaces.nsIObserverService,
            observerSvc, PROXY_ALWAYS | PROXY_SYNC)
        if sys.platform.startswith("win"):
            global _gHWnd
            _gHWnd = win32gui.GetActiveWindow()

        # Determine file/event/lock names.
        infoSvc = components.classes["@activestate.com/koInfoService;1"].getService()
        ver = infoSvc.version          # '1.9.0-devel'
        ver = re.split('\.|-', ver)    # ['1', '9', '0', 'devel']
        ver = "%s.%s" % tuple(ver[:2]) # '1.9'
        if sys.platform.startswith("win"):
            self._statusLockName = "komodo-%s-status-lock" % ver
            self._runningEventName = "komodo-%s-running" % ver
            self._commandmentsLockName = "komodo-%s-commandments-lock" % ver
            self._commandmentsEventName = "komodo-%s-new-commandments" % ver
            self._commandmentsFileName = os.path.join(
                shell.SHGetFolderPath(0, shellcon.CSIDL_APPDATA, 0, 0),
                "ActiveState", "Komodo", ver, "commandments.txt")
        else:
            self._runningLockFileName = os.path.expanduser("~/.komodo/%s/running.lock" % ver)
            self._commandmentsFileName = os.path.expanduser("~/.komodo/%s/commandments.fifo" % ver)

        # Start commandment handler thread.
        if sys.platform.startswith("win"):
            self._reader = _CommandmentsReader(self._commandmentsFileName,
                                               self._commandmentsEventName,
                                               self._commandmentsLockName)
        else:
            self._reader = _CommandmentsReader(self._commandmentsFileName)
        self._reader.start()

        # Tell listening processes that Komodo's commandment system is up
        # and running (as per commandments-spec.txt).
        if sys.platform.startswith("win"):
            lock = win32event.CreateMutex(None, 0, self._statusLockName)
            #XXX Perhaps should have a reasonable timeout?
            win32event.WaitForSingleObject(lock, win32event.INFINITE)

            self._running = win32event.CreateEvent(None, 1, 0,
                                                   self._runningEventName)
            win32event.SetEvent(self._running)

            win32event.ReleaseMutex(lock)
            win32api.CloseHandle(lock)
        else:
            self._running = open(self._runningLockFileName, 'w')
            self._running.write( str(os.getpid()) )
            self._running.flush()
            fcntl.flock(self._running, fcntl.LOCK_EX)

    def finalize(self):
        # Tell listening processes that Komodo's commandment system is
        # no longer running.
        if sys.platform.startswith("win"):
            win32event.ResetEvent(self._running)
            win32api.CloseHandle(self._running)
        else:
            fcntl.flock(self._running, fcntl.LOCK_UN)
            self._running.close()
            os.unlink(self._runningLockFileName)
 
        # Shutdown handler thread.
        self._reader.exit()

