#!python
# Copyright (c) 2000-2001 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

from xpcom import components
import sys, os, re

# A service for providing Komodo-specific directory/filename information.

if sys.platform.startswith("win"):
    from win32com.shell import shellcon, shell
    ##  XXX win32com.shellcon is missing CSIDL_COMMON_APPDATA
    shellcon.CSIDL_COMMON_APPDATA = 0x23


def SmallestVersionFirst(a, b):
    """Compare two version strings and return:
        -1 if a < b
         0 if a == b
        +1 if a > b
    """
    return cmp([int(elem) for elem in a.split('.')],
               [int(elem) for elem in b.split('.')])


class koDirs:
    _com_interfaces_ = [components.interfaces.koIDirs]
    _reg_clsid_ = "{98435d6d-e24d-4057-a72a-dfcd1d282fae}"
    _reg_contractid_ = "@activestate.com/koDirs;1"
    _reg_desc_ = "Komodo Directory Information Service"

    def __init__(self):
        pass

    def GetAllUserDataDirs(self):
        """Return list of the full paths to all existing Komodo
        user data dirs. This list is ordered from lowest Komodo version
        to highest."""
        if sys.platform.startswith("win"):
            base = os.path.join(
                shell.SHGetFolderPath(0, shellcon.CSIDL_APPDATA, 0, 0),
                "ActiveState", "Komodo")
        else:
            base = os.path.expanduser("~/.komodo")
        # If there are any *files* in the base dir then Komodo 1.0 was
        # probably run on this machine.
        try:
            paths = os.listdir(base)
        except OSError:
            # 'base' does not exist
            return []
        userDataDirDict = {}
        versionRe = re.compile("^\d+\.\d+$")
        for path in paths:
            fullPath = os.path.join(base, path)
            if os.path.isfile(fullPath):
                userDataDirDict["1.0"] = os.path.dirname(fullPath)
            elif os.path.isdir(fullPath):
                # Ensure it is a Komodo version dir (i.e. noone came in an
                # added there own dir).
                if versionRe.search(path):
                    userDataDirDict[path] = fullPath
        # put the dirs in the correct order
        versions = userDataDirDict.keys()
        versions.sort(SmallestVersionFirst)
        userDataDirs = [userDataDirDict[version] for version in versions]
        return userDataDirs

    def _GetUserDataDirForVersion(self, major, minor):
        if sys.platform.startswith("win"):
            base = shell.SHGetFolderPath(0, shellcon.CSIDL_APPDATA, 0, 0)
            if (major,minor) == (1,0):
                return os.path.join(base, "ActiveState", "Komodo")
            else:
                return os.path.join(base, "ActiveState", "Komodo",
                                    "%s.%s" % (major, minor))
        else:
            if (major,minor) == (1,0):
                return os.path.expanduser("~/.komodo")
            else:
                return os.path.expanduser("~/.komodo/%s.%s" % (major, minor))

    def _GetCommonDataDirForVersion(self, major, minor):
        if sys.platform.startswith("win"):
            base = shell.SHGetFolderPath(0, shellcon.CSIDL_COMMON_APPDATA, 0, 0)
            if (major,minor) == (1,0):
                return os.path.join(base, "ActiveState", "Komodo")
            else:
                return os.path.join(base, "ActiveState", "Komodo",
                                    "%s.%s" % (major, minor))
        else:
            if (major,minor) == (1,0):
                return "/etc/komodo"
            else:
                return "/etc/komodo/%s.%s" % (major, minor)

    def get_userDataDir(self):
        infoSvc = components.classes["@activestate.com/koInfoService;1"].\
            getService(components.interfaces.koIInfoService)
        major, minor = infoSvc.version.split('.', 2)[:2]
        return self._GetUserDataDirForVersion(major, minor)

    def get_commonDataDir(self):
        infoSvc = components.classes["@activestate.com/koInfoService;1"].\
               getService(components.interfaces.koIInfoService)
        major, minor = infoSvc.version.split('.', 2)[:2]
        return self._GetCommonDataDirForVersion(major, minor)

    def get_installDir(self):
        # KOMODO_INSTALLDIR is (supposed to be) set by komodo.exe
        import os
        return os.environ["KOMODO_INSTALLDIR"]

    def get_mozBinDir(self):
        # An installed Komodo can build this from the "installDir".
        # A development Komodo can just get it from the KOMODO_MOZBINDIR
        # environment variable, which, in general will not be defined for
        # installed Komodo's.
        import os
        mozBinDir = os.environ.get("KOMODO_MOZBINDIR", None)
        if mozBinDir is None:
            mozBinDir = os.path.join(self.get_installDir(), "Mozilla")
        return mozBinDir

    def get_perlUtilsDir(self):
        # Currently this is just the perl site lib. This dir is
        # however dep on the platform and the Perl version (on linux),
        # neither of which should be hardcoded here.
        if sys.platform.startswith("win"):
            return os.path.join(self.get_installDir(), "perl", "site", "lib")
        else:
            siloedPerlVersion = "5.6.0" #XXX should not be hardcoded
            return os.path.join(self.get_installDir(), "perl", "lib",
                                "site_perl", siloedPerlVersion)

if __name__ == "__main__":
    dirs = components.classes['@activestate.com/koDirs;1'].createInstance(components.interfaces.koIDirs)
    print dirs.userDataDir
    print dirs.commonDataDir
    try:
        print dirs.installDir
    except:
        print "couldn't get installDir -- env't variable not set right most likely"

