#!/usr/bin/env python
# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

import sys, os, re
from xpcom import components, ServerException, nsError
from koLogger import logError, logWarning, logDebug


#---- globals

if sys.platform.startswith("win"):
    _gEnvStrRe = re.compile("^(?P<name>[^&=\r\n\t]+)=(?P<value>.*)$")
    _gEnvStrResToSkip = []
else:
    _gEnvStrRe = re.compile("^(?P<name>[\w]+)=(?P<value>.*)$")
    # On linux we want to skip function definitions like this:
    #
    #    mc=()
    #    {
    #        echo hello;
    #        ls --color=auto ~;
    #        echo bye
    #    }
    # 
    # or this:
    #
    #    mc=() {  mkdir -p $HOME/.mc/tmp 2>/dev/null;
    #     chmod 700 $HOME/.mc/tmp;
    #     MC=$HOME/.mc/tmp/mc-$$;
    #     /usr/bin/mc -P "$@" >"$MC";
    #     cd "`cat $MC`";
    #     rm -i -f "$MC";
    #     unset MC
    #    }
    #
    _gEnvStrResToSkip = [re.compile("^[_\w]+=\(\).*$"),  # function defs
                         re.compile("^\s+.*$"), # leading whitespace == in function def
                         re.compile("^{.*$"), # function start
                         re.compile("^}$"), # function end
                        ]


#---- support routines

def _ParseEnvStr(envStr):
    """Return the (env var name, env var value) pair from the given
    'name=value' environment string.
    """
    # skip unwanted lines formats
    for reToSkip in _gEnvStrResToSkip:
        if reToSkip.search(envStr):
            return None

    # parse out name and value
    envStrMatch = _gEnvStrRe.search(envStr)
    if envStrMatch:
        name = envStrMatch.group("name")
        if sys.platform.startswith("win"):
            name = name.upper()  # put everything upper on Windows
        value = envStrMatch.group("value")
        return name, value
    else:
        logWarning("bogus environment string: '%s'" % envStr)


# simple interface to be able to work with the environment
class KoEnviron:
    _com_interfaces_ = [components.interfaces.koIEnviron]
    _reg_clsid_ = "{68E2A458-186F-4e2c-898E-99F3B21DA819}"
    _reg_contractid_ = "@activestate.com/koEnviron;1"
    _reg_desc_ = "System Environment Interaction Service"

    def has(self, key):
        return os.environ.has_key(key)
    def get(self, key):
        return os.environ[key]
    def set(self, key, value):
        os.environ[key] = value


class KoUserEnviron:
    _com_interfaces_ = [components.interfaces.koIUserEnviron]
    _reg_clsid_ = "{B6289D1A-983E-46e0-85E2-75221D44A5FC}"
    _reg_contractid_ = "@activestate.com/koUserEnviron;1"
    _reg_desc_ = "User Environmnet Interaction Service"

    def __init__(self, startupEnvFileName=None):
        # By default the startup environment file is in the Komodo AppData
        # directory. For testing purposes, though, it can be useful to
        # specify a test file.
        self._startupEnv = None
        if startupEnvFileName is None:
            self._startupEnvBaseName = "startup-env.tmp"
            koDirs = components.classes["@activestate.com/koDirs;1"].\
                getService(components.interfaces.koIDirs)
            self.startupEnvFileName = os.path.join(koDirs.userDataDir,
                                                   "startup-env.tmp")
        else:
            self.startupEnvFileName = startupEnvFileName
    
    def _GetStartupEnv(self):
        if not self._startupEnv:
            self._startupEnv = {}
            # read in startup env delta file and fill in the startup env
            fin = open(self.startupEnvFileName, "r")
            content = fin.read()
            fin.close()
            if not sys.platform.startswith("win"):
                # Replace "\"-continued lines in *Linux* environment dumps.
                # Cannot do this on Windows because a "\" at the end of the
                # line does not imply a continuation.
                content = content.replace("\\\n", "")
            for line in content.split("\n"):
                line = line.rstrip()
                if not line:
                    continue  # skip black lines
                item =  _ParseEnvStr(line)
                if item:
                    self._startupEnv[item[0]] = item[1]

    def has(self, key):
        self._GetStartupEnv()
        if sys.platform.startswith("win"):
            key = key.upper()  # put everything upper on Windows
        return self._startupEnv.has_key(key)

    def get(self, key):
        self._GetStartupEnv()
        if sys.platform.startswith("win"):
            key = key.upper()  # put everything upper on Windows
        return self._startupEnv[key]

    def GetEnvironmentStrings(self):
        self._GetStartupEnv()
        return ["%s=%s" % (item[0], item[1])\
            for item in self._startupEnv.items()]

    def GetEncodedEnvironment(self):
        self._GetStartupEnv()
        envStrings = self.GetEnvironmentStrings()
        return '\n'.join(envStrings)


class KoEnvironUtils:
    _com_interfaces_ = [components.interfaces.koIEnvironUtils]
    _reg_clsid_ = "{cb1e55b4-6837-4659-a9e0-9f6ab4acb54d}"
    _reg_contractid_ = "@activestate.com/koEnvironUtils;1"
    _reg_desc_ = "Environment Variable Utilities"

    def MergeEnvironmentStrings(self, baseEnvStrs, diffEnvStrs):
        envDict = {}
        #print "MergeEnvironmentStrings: merge %s into %s" % (diffEnvStrs,
        #                                                     baseEnvStrs)
        for baseEnvStr in baseEnvStrs:
            item = _ParseEnvStr(baseEnvStr)
            if item:
                envDict[item[0]] = item[1]
        for diffEnvStr in diffEnvStrs:
            #print "merge in '%s' ..." % diffEnvStr
            item = _ParseEnvStr(diffEnvStr)
            if item:
                name = item[0]
                value = item[1]
            else:
                continue
            # an empty value means: remove env. var.
            if value == "":
                if envDict.has_key(name):
                    del envDict[name]
                #print "... delete '%s' variable" % name
            # interpolate any other value into the resultant dictionary
            else:
                value = self._Interpolate(value, envDict)
                envDict[name] = value
                #print "... %s=%s" % (name, value)
        return ["%s=%s" % item for item in envDict.items()]

    def _Interpolate(self, envValue, envDict):
        """Interpolate the variable in the given envDict into the envValue.
        Use platform-specific syntax, i.e. %FOO% on Windows and $FOO or $(FOO)
        elsewhere.
        """
        if sys.platform.startswith("win"):
            envRefRe = re.compile("""
                ^(?P<pre>.*)                    # text before env. var ref
                (?P<envRef>%                    # env. var. reference
                    (?P<envName>[^&=\r\n\t%]+)  # env. var. name
                %)
                (?P<post>.*)$                   # test after env. var ref
                """, re.VERBOSE)
        else:
            envRefRe = re.compile(r"""
                ^(?P<pre>.*)                    # text before env. var ref
                (?P<envRef>
                    \$\b(?P<envName>\w+)\b      # simple $FOO like reference
                    |
                    \$\((?P<envName2>\w+)\)     # parenthesized $(FOO) like ref
                )
                (?P<post>.*)$                   # test after env. var ref
                """, re.VERBOSE)
        interpolated = envValue
        while 1:
            envRefMatch = envRefRe.search(interpolated)
            if envRefMatch:
                envName = envRefMatch.group("envName") \
                          or envRefMatch.group("envName2")
                if sys.platform.startswith("win"):
                    envName = envName.upper()
                # Tough call on how interpolation of undefined variables
                # should be handled because:
                #    windows> echo %NOTDEFINED%
                #    %NOTDEFINED%
                #    linux> echo $NOTDEFINED
                #
                # We will expect the Linux behaviour on both Windows and Linux
                # for now.
                interpolated = envRefMatch.group("pre") \
                               + envDict.get(envName, "") \
                               + envRefMatch.group("post")
            else:
                break
        #print "_Interpolate: interpolate '%s' -> '%s'" % (envValue, interpolated)
        return interpolated



#---- self test code
# "bk start src/misc-components/koEnviron.py" to run it. Requires Python
# 2.1 or at least requires PyUnit on the PythonPath.

try:
    import unittest
except ImportError:
    if __name__ == "__main__":
        print "*** No PyUnit on PythonPath with which to run unit tests."
        raise
    else:
        pass
else: 
    class UserEnvironParsingTestCase(unittest.TestCase):
        def _GetUserEnviron(self, content):
            import tempfile
            tmpFileName = tempfile.mktemp()
            tmpFile = open(tmpFileName, "w")
            tmpFile.write(content)
            tmpFile.close()
            return KoUserEnviron(tmpFileName)

        def test_simple(self):
            userEnviron = self._GetUserEnviron("FOO=bar")
            self.assert_(userEnviron.has("FOO"))
            self.assertEqual(userEnviron.get("FOO"), "bar")

        def test_spaces(self):
            # this should only be possible on windows
            if sys.platform.startswith("win"):
                userEnviron = self._GetUserEnviron("FOO =bar")
                self.assert_(userEnviron.has("FOO "))
                self.assertEqual(userEnviron.get("FOO "), "bar")

        def test_leading_semicolon(self):
            # This should only be possible on windows. This sort of this
            # should just be skipped as it is obviously an error in the users
            # environment.
            if sys.platform.startswith("win"):
                userEnviron = self._GetUserEnviron(";Path=blah")
                self.assert_(not userEnviron.has("Path"))
                self.assert_(userEnviron.has(";Path"))

        def test_quotes(self):
            userEnviron = self._GetUserEnviron('FOO=bar;"spam";"eggs and ham"')
            self.assert_(userEnviron.has("FOO"))
            self.assertEqual(userEnviron.get("FOO"),
                             'bar;"spam";"eggs and ham"')

        def test_merge_simple1(self):
            utils = KoEnvironUtils()
            base = ["MOZ_BITS=32"]
            diff = ["MOZ_BITS=52"]
            merged = utils.MergeEnvironmentStrings(base, diff)
            self.assert_("MOZ_BITS=52" in merged)

        def test_merge_simple2(self):
            utils = KoEnvironUtils()
            base = ["MOZ_BITS=32", "FOO=bar"]
            diff = ["MOZ_BITS=52"]
            merged = utils.MergeEnvironmentStrings(base, diff)
            self.assert_("MOZ_BITS=52" in merged)

        def test_merge_interpolate1(self):
            utils = KoEnvironUtils()
            base = ["MOZ_BITS=32", "FOO=bar"]
            if sys.platform.startswith("win"):
                diff = ["MOZ_BITS=MOZ_BITS was %MOZ_BITS%"]
            else:
                diff = ["MOZ_BITS=MOZ_BITS was $MOZ_BITS"]
            merged = utils.MergeEnvironmentStrings(base, diff)
            self.assert_("MOZ_BITS=MOZ_BITS was 32" in merged)

        def test_merge_interpolate2(self):
            utils = KoEnvironUtils()
            base = ["MOZ_BITS=32", "FOO=bar"]
            if sys.platform.startswith("win"):
                diff = ["MOZ_BITS=MOZ_BITS was %MOZ_BITS%"]
            else:
                # different interp form than for test_merge_interpolate1()
                diff = ["MOZ_BITS=MOZ_BITS was $(MOZ_BITS)"]
            merged = utils.MergeEnvironmentStrings(base, diff)
            self.assert_("MOZ_BITS=MOZ_BITS was 32" in merged)

        def test_merge_interpolate3(self):
            # interpolating an undefined variable
            utils = KoEnvironUtils()
            base = ["MOZ_BITS=32", "FOO=bar"]
            if sys.platform.startswith("win"):
                diff = ["MOZ_BITS=MOZ_BITS was not %MOZ_NOTDEFINED%"]
            else:
                diff = ["MOZ_BITS=MOZ_BITS was not $(MOZ_NOTDEFINED)"]
            merged = utils.MergeEnvironmentStrings(base, diff)
            self.assert_("MOZ_BITS=MOZ_BITS was not " in merged)

        def test_environment_function1(self):
            if sys.platform.startswith("win"): return
            # Data is from bug
            # http://bugs.activestate.com/Komodo/show_bug.cgi?id=18855
            envStrings = """
LC_COLLATE=en_US
CURRENT_PROJECT=ADB-QA
mc=() {  mkdir -p $HOME/.mc/tmp 2>/dev/null;
 chmod 700 $HOME/.mc/tmp;
 MC=$HOME/.mc/tmp/mc-$$;
 /usr/bin/mc -P "$@" >"$MC";
 cd "`cat $MC`";
 rm -i -f "$MC";
 unset MC
}
_=./komodo
"""
            userEnviron = self._GetUserEnviron(envStrings)
            self.assert_(userEnviron.has("LC_COLLATE"))
            self.assertEqual(userEnviron.get("LC_COLLATE"), 'en_US')
            self.assert_(userEnviron.has("_"))
            self.assertEqual(userEnviron.get("_"), './komodo')

        def test_environment_function2(self):
            if sys.platform.startswith("win"): return
            # This is the way *my* bash shows defined functions.
            envStrings = """
LC_COLLATE=en_US
CURRENT_PROJECT=ADB-QA
mc=()
{
    echo hello;
    ls --color=auto ~;
    echo bye
}
_=./komodo
"""
            userEnviron = self._GetUserEnviron(envStrings)
            self.assert_(userEnviron.has("LC_COLLATE"))
            self.assertEqual(userEnviron.get("LC_COLLATE"), 'en_US')
            self.assert_(userEnviron.has("_"))
            self.assertEqual(userEnviron.get("_"), './komodo')

        def test_environment_function3(self):
            # See http://bugs.activestate.com/show_bug.cgi?id=20685
            # Basically just want to make sure that this environment can
            # be parsed -- properly skipping the first (bogus?) arg.
            if not sys.platform.startswith("win"): return
            # This is the way *my* bash shows defined functions.
            envStrings = """
=\\Umax\umax c=\\Umax\umax c\WINDOWS\FONTS\
ALLUSERSPROFILE=C:\Documents and Settings\All Users
APPDATA=C:\Documents and Settings\Larry\Application Data
CLIENTNAME=Console
CommonProgramFiles=C:\Program Files\Common Files
COMPUTERNAME=LARRY_DELL
ComSpec=C:\WINDOWS\system32\cmd.exe
HOMEDRIVE=C:
HOMEPATH=\Documents and Settings\Larry
LOGONSERVER=\\LARRY_DELL
NUMBER_OF_PROCESSORS=1
OS=Windows_NT
Path=C:\Program Files\Python\.;C:\Perl\bin\;C:\WINDOWS\system32;C:\WINDOWS;C:\WINDOWS\System32\Wbem
PATHEXT=.COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.py;.pyc;.pyo;.pyw;.pys
PROCESSOR_ARCHITECTURE=x86
PROCESSOR_IDENTIFIER=x86 Family 15 Model 2 Stepping 4, GenuineIntel
PROCESSOR_LEVEL=15
PROCESSOR_REVISION=0204
ProgramFiles=C:\Program Files
SESSIONNAME=Console
SystemDrive=C:
SystemRoot=C:\WINDOWS
TEMP=C:\DOCUME~1\Larry\LOCALS~1\Temp
TMP=C:\DOCUME~1\Larry\LOCALS~1\Temp
USERDOMAIN=LARRY_DELL
USERNAME=Larry
USERPROFILE=C:\Documents and Settings\Larry
windir=C:\WINDOWS
"""
            userEnviron = self._GetUserEnviron(envStrings)
            self.assert_(userEnviron.has("CLIENTNAME"))
            self.assertEqual(userEnviron.get("CLIENTNAME"), 'Console')


    if __name__ == '__main__':
        unittest.main()

