#!/usr/bin/env python
# Copyright (c) 2000-2002 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

import os
import sys
import types
import threading
import Queue
from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC
from koLogger import logError, logWarning, logDebug


#---- internal logging facility

class Logger:
    DEBUG, INFO, WARN, ERROR, FATAL = range(5)
    def __init__(self, name, level=None, streamOrFileName=sys.stderr):
        self.name = name
        if level is None:
            self.level = self.WARN
        else:
            self.level = level
        if type(streamOrFileName) == types.StringType:
            self.stream = open(streamOrFileName, 'w')
            self._opennedStream = 1
        else:
            self.stream = streamOrFileName
            self._opennedStream = 0
    def __del__(self):
        if self._opennedStream:
            self.stream.close()
    def _getLevelName(self, level):
        levelNameMap = {
            self.DEBUG: "DEBUG",
            self.INFO: "INFO",
            self.WARN: "WARN",
            self.ERROR: "ERROR",
            self.FATAL: "FATAL",
        }
        return levelNameMap[level]
    def log(self, level, msg, *args):
        if level < self.level:
            return
        message = "%s: %s: " % (self.name, self._getLevelName(level).lower())
        message = message + (msg % args) + "\n"
        self.stream.write(message)
        self.stream.flush()
    def debug(self, msg, *args):
        self.log(self.DEBUG, msg, *args)
    def info(self, msg, *args):
        self.log(self.INFO, msg, *args)
    def warn(self, msg, *args):
        self.log(self.WARN, msg, *args)
    def error(self, msg, *args):
        self.log(self.ERROR, msg, *args)
    def fatal(self, msg, *args):
        self.log(self.FATAL, msg, *args)

if 1:   # normal/production usage
    log = Logger("fss", Logger.WARN)
else:   # development/debugging usage
    log = Logger("fss", Logger.DEBUG, sys.stdout)



#---- component implementation

class KoFeatureStatus:
    _com_interfaces_ = [components.interfaces.koIFeatureStatus]
    _reg_clsid_ = "{62DA2673-1653-4CC6-9A06-8FDBFE39F025}"
    _reg_contractid_ = "@activestate.com/koFeatureStatus;1"
    _reg_desc_ = "The status of a particular Komodo feature."

    def __init__(self, featureName, status, reason=None):
        self.featureName = featureName
        self.status = status
        self.reason = reason


class KoFeatureStatusService:
    """Listen for a "feature_status_request" notifications to initiate
    status retrieval in a separate thread and then return the result in
    a "feature_status_result".
    
    Implementation Notes:
        This service currently will startup a background thread only when
        necessary to avoid having a thread always sitting there to
        gather data that is very infrequently gathered. This thread is
        smart enough to keep going if requests for status on multiple
        features come in in quick succession.
        
        This setup makes the code a little bit more complex that it would
        be with just a single long running thread.
    """
    _com_interfaces_ = [components.interfaces.koIFeatureStatusService,
                        components.interfaces.nsIObserver]
    _reg_clsid_ = "{46A0BD16-D67C-4DCD-AE19-5C9CECB6A669}"
    _reg_contractid_ = "@activestate.com/koFeatureStatusService;1"
    _reg_desc_ = "Give functional status info for features in Komodo."

    def __init__(self):
        self._fsLock = threading.Lock()
        # A queue of feature names for which the status has been requested.
        self._fsQueue = Queue.Queue()
        # A thread to gather the status of the features named in the Queue.
        self._fsThread = None
        self._fsThreadExiting = threading.Event()

    def initialize(self):
        log.info("initialize")
        # Get a schwack of services and components and proxy them all.
        self._proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"].\
            getService(components.interfaces.nsIProxyObjectManager)
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._observerProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.nsIObserverService, self._observerSvc,
            PROXY_ALWAYS | PROXY_ASYNC)

        self._lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"].\
            getService(components.interfaces.koILastErrorService)
        self._lastErrorProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.koILastErrorService, self._lastErrorSvc,
            PROXY_ALWAYS | PROXY_SYNC)
        self._prefSvc = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService)
        self._prefProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.koIPrefService, self._prefSvc,
            PROXY_ALWAYS | PROXY_SYNC)
        self._sysUtilsSvc = components.classes["@activestate.com/koSysUtils;1"].\
            getService(components.interfaces.koISysUtils)
        self._sysUtilsProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.koISysUtils, self._sysUtilsSvc,
            PROXY_ALWAYS | PROXY_SYNC)
        self._cvsSvc = components.classes["@activestate.com/koCVS;1"].\
            getService(components.interfaces.koICVS)
        self._cvsProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.koICVS, self._cvsSvc,
            PROXY_ALWAYS | PROXY_SYNC)
        self._p4Svc = components.classes["@activestate.com/koP4;1"].\
            getService(components.interfaces.koIP4)
        self._p4Proxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.koIP4, self._p4Svc,
            PROXY_ALWAYS | PROXY_SYNC)
        self._vpmSvc = components.classes["@activestate.com/koVPMService;1"].\
            getService(components.interfaces.koIVPMService)
        self._vpmProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.koIVPMService, self._vpmSvc,
            PROXY_ALWAYS | PROXY_SYNC)

        self._perlInfoEx = components.classes["@activestate.com/koPerlInfoEx;1"].\
            createInstance(components.interfaces.koIAppInfoEx)
        self._phpInfoEx = components.classes["@activestate.com/koPHPInfoEx;1"].\
            createInstance(components.interfaces.koIAppInfoEx)
        self._pythonInfoEx = components.classes["@activestate.com/koPythonInfoEx;1"].\
            createInstance(components.interfaces.koIAppInfoEx)
        self._tdkInfoEx = components.classes["@activestate.com/koTclProInfoEx;1"].\
            createInstance(components.interfaces.koITclInfoEx)
        self._tclInfoEx = components.classes["@activestate.com/koTclInfoEx;1"].\
            createInstance(components.interfaces.koITclInfoEx)

        self._ignoreExceptions = 0
        self._observerSvc.AddObserver(self, "feature_status_request")

    def finalize(self):
        log.info("finalize")
        self._observerSvc.RemoveObserver(self, "feature_status_request")
        # Would like to .join() on the acquirer thread but that results
        # in deadlock because while the main thread is waiting on the
        # join the proxied services being used in the acquirer thread
        # are trying to use the main thread. Sol'n: just let the
        # acquirer thread run its course but tell it to ignore
        # exceptions.
        self._ignoreExceptions = 1
        # XXX Would like to make self._fsQueue a priority Queue and
        #     insert a high priority __exit__ directive on finalize.
        #     No need to process remaining requests.

    def Observe(self, dummy, topic, featureName):
        if topic == "feature_status_request":
            log.info("request status of feature '%s'", featureName)
            # Queue up that feature name.
            self._fsQueue.put(featureName)
            # Start up the background thread if there is not already one
            # running.
            self._fsLock.acquire()
            if self._fsThread is None:
                self._fsThread = threading.Thread(target=self._fsAcquirer)
                self._fsThreadExiting.clear()
                self._fsThread.start()
            elif self._fsThreadExiting.isSet():
                self._fsThread.join()
                self._fsThread = threading.Thread(target=self._fsAcquirer)
                self._fsThreadExiting.clear()
                self._fsThread.start()
            self._fsLock.release()
        else:
            errmsg = "Unexpected nsIObserver topic: '%s'" % topic
            log.error(errmsg)
            logError(errmsg)
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)

    def _fsAcquirer(self):
        """Acquire status for all the features named in self._fsQueue and
        send notifications with the results until the queue is empty.
        """
        log.info("start feature status acquirer thread")
        while 1:
            self._fsLock.acquire()
            try:
                try:
                    featureName = self._fsQueue.get_nowait()
                except Queue.Empty:
                    log.info("status acquirer thread exiting")
                    self._fsThreadExiting.set()
                    break
            finally:
                self._fsLock.release()
                
            try:
                status = "Status Unknown"
                reason = None
                if featureName == "Perl Debugging":
                    if self._haveSufficientPerl(minVersion="5.6.0"):
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                elif featureName == "Perl Syntax Checking":
                    if self._haveSufficientPerl():
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                elif featureName == "PHP Debugging":
                    if self._haveSufficientPHP(feature="Debugging",
                                                minVersion="4.0.5"):
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                elif featureName == "PHP Syntax Checking":
                    if self._haveSufficientPHP(feature="Syntax Checking",
                                                minVersion="4.0.5"):
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                elif featureName == "Python Debugging":
                    if self._haveSufficientPython(minVersion="1.5.2"):
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                elif featureName == "Python Syntax Checking":
                    if self._haveSufficientPython():
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                elif featureName == "Tcl Debugging":
                    if self._haveSufficientTDK(feature="Debugging"):
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                elif featureName == "Tcl Syntax Checking":
                    if self._haveSufficientTDK(feature="Syntax Checking"):
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                elif featureName == "CVS":
                    if self._prefProxy.prefs.getBooleanPref("cvsEnabled"):
                        if self._cvsProxy.isFunctional:
                            status = "Ready"
                        else:
                            status = "Not Functional"
                            reason = self._cvsProxy.reasonNotFunctional
                    else:
                        status = "Disabled"
                elif featureName == "Perforce":
                    if self._prefProxy.prefs.getBooleanPref("p4Enabled"):
                        if self._p4Proxy.isFunctional:
                            status = "Ready"
                        else:
                            status = "Not Functional"
                            reason = self._p4Proxy.reasonNotFunctional
                    else:
                        status = "Disabled"
                elif featureName == "VPM":
                    if self._vpmProxy.isFunctional:
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                elif featureName == "Perl Dialog":
                    if self._haveSufficientPerl(haveModules=['Tk']):
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                elif featureName == "Python Dialog":
                    if self._haveSufficientPython(minVersion="1.5.2"):
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                elif featureName == "Tcl Dialog":
                    # XXX Would like to require >= version 8.3 but koITclInfoEx
                    #     does not support determining the version of an
                    #     installed Tcl.
                    if self._haveSufficientTcl():
                        status = "Ready"
                    else:
                        status = "Not Functional"
                        reason = self._lastErrorProxy.getLastErrorMessage()
                else:
                    errmsg = "Unexpected feature name: %s" % featureName
                    log.error(errmsg)
                    logError(errmsg)
                    raise ServerException(nsError.NS_ERROR_UNEXPECTED)
                
                log.info("status of %r feature: %s (%s)", featureName, status,
                         reason)
                st = KoFeatureStatus(featureName, status, reason)
                try:
                    self._observerProxy.Notify(st, "feature_status_ready",
                                               featureName)
                except COMException, e:
                    # do nothing: Notify sometimes raises an exception if
                    # receivers are not registered?
                    pass
            except:
                if not self._ignoreExceptions:
                    raise

    def _isSufficientPerl(self, perlInfoEx, minVersion=None,
                          isActivePerl=None, minActivePerlBuild=None,
                          haveModules=[]):
        """Return true iff the given perl installation meets the given
        criteria.
            "perlInfoEx" is a koPerlInfoEx instance loaded with the perl
                installation to check.

        If false is returned, then the last error (see koILastErrorService)
        is set with a reason why.
        """
        installDir = perlInfoEx.installationPath
        exePath = perlInfoEx.executablePath
        log.info("%s: is sufficient perl?", installDir)

        if not os.path.exists(exePath):
            log.info("%s: does not exist", exePath)
            self._lastErrorProxy.setLastError(0,
                "Perl installation does not exist: \"%s\"" % exePath)
            return 0
        else:
            log.info("%s: exists", exePath)

        if minVersion is not None:
            try:
                version = perlInfoEx.version
            except COMException:
                log.info("%s: couldn't get version", installDir)
                self._lastErrorProxy.setLastError(0,
                    "Could not determine Perl version: %s. " % installDir)
                return 0
            else:
                if version < minVersion:
                    log.info("%s: %s < %s", installDir, version, minVersion)
                    self._lastErrorProxy.setLastError(0,
                        "Insufficient Perl version (\"%s\" is version %s). "\
                        "Require at least version %s."\
                        % (installDir, version, minVersion))
                    return 0
                else:
                    log.info("%s: %s >= %s", installDir, version, minVersion)
        
        buildNumber = None
        if isActivePerl is not None or minActivePerlBuild is not None:
            try:
                buildNumber = perlInfoEx.buildNumber
            except COMException, ex:
                log.info("%s: not ActivePerl", installDir)
                self._lastErrorProxy.setLastError(0,
                    "Perl installation is not ActivePerl: %s" % installDir)
                return 0
            else:
                log.info("%s: is ActivePerl", installDir)

        if minActivePerlBuild is not None:
            if buildNumber is None:
                log.info("%s: couldn't determine ActivePerl build number",
                         installDir)
                self._lastErrorProxy.setLastError(0,
                    "Could not determine ActivePerl build number: %s"\
                    % installDir)
                return 0
            elif buildNumber < minActivePerlBuild:
                log.info("%s: build %s < build %s", installDir, buildNumber,
                         minActivePerlBuild)
                self._lastErrorProxy.setLastError(0,
                    "Insufficient ActivePerl build (\"%s\" is build %s). "\
                    "Require at least version %s."\
                    % (installDir, buildNumber, minActivePerlBuild))
                return 0
            else:
                log.info("%s: build %s >= build %s", installDir, buildNumber,
                         minActivePerlBuild)

        if haveModules:
            if not perlInfoEx.haveModules(haveModules):
                log.info("%s: does not have all of the following modules: %r",
                         installDir, haveModules)
                self._lastErrorProxy.setLastError(0,
                    "Perl installation at '%s' does not have all of the "\
                    "following modules: %s"\
                    % (installDir, ','.join(haveModules)))
                return 0
            else:
                log.info("%s: has required modules: %r", installDir,
                         haveModules)

        return 1

    def _haveSufficientPerl(self, stopOnFirst=1, minVersion=None,
                            isActivePerl=None, minActivePerlBuild=None,
                            haveModules=[]):
        """Return true iff a Perl installation meeting the above conditions
        can be found.
            "stopOnFirst" is a boolean indicating if the search should stop
                on the "first" found interpreter. This is because things like
                debugging and syntax checking are just going to use this
                one anyway. Default is true.

        If false is returned, then the last error (see koILastErrorService)
        is set with a reason why.
        """
        # Try the user's selected perl interpreter.
        perlDefaultInterp = self._prefProxy.prefs.getStringPref("perlDefaultInterpreter")
        if perlDefaultInterp:
            self._perlInfoEx.installationPath =\
                self._perlInfoEx.getInstallationPathFromBinary(perlDefaultInterp)
            if self._isSufficientPerl(self._perlInfoEx, minVersion,
                                      isActivePerl, minActivePerlBuild,
                                      haveModules):
                return 1
            elif stopOnFirst:
                return 0
            
        # Look on PATH.
        perls = self._sysUtilsProxy.WhichAll('perl')
        for perl in perls:
            self._perlInfoEx.installationPath =\
                self._perlInfoEx.getInstallationPathFromBinary(perl)
            if self._isSufficientPerl(self._perlInfoEx, minVersion,
                                      isActivePerl, minActivePerlBuild,
                                      haveModules):
                return 1
            elif stopOnFirst:
                return 0
        
        errmsg = "Could not find a suitable Perl installation."
        self._lastErrorProxy.setLastError(0, errmsg)
        return 0

    def _isSufficientPHP(self, phpInfoEx, feature="Debugging",
                         minVersion=None):
        """Return true iff the given php installation meets the given
        criteria.
            "phpInfoEx" is a koPHPInfoEx instance loaded with the
                php installation to check.

        If false is returned, then the last error (see koILastErrorService)
        is set with a reason why.
        """
        installDir = phpInfoEx.installationPath
        exePath = phpInfoEx.executablePath
        log.info("%s: is sufficient php?", installDir)
        
        if not os.path.exists(exePath):
            log.info("%s: does not exist", exePath)
            self._lastErrorProxy.setLastError(0,
                "PHP installation does not exist: \"%s\"" % exePath)
            return 0
        else:
            log.info("%s: exists", exePath)

        if minVersion is not None:
            version = phpInfoEx.version
            if version < minVersion:
                log.info("%s: %s < %s", installDir, version, minVersion)
                self._lastErrorProxy.setLastError(0,
                    "Insufficient PHP version (\"%s\" is version %s). "\
                    "Require at least version %s."\
                    % (installDir, version, minVersion))
                return 0
            else:
                log.info("%s: %s >= %s", installDir, version, minVersion)
        
        if feature == "Debugging":
            if not phpInfoEx.isDebuggerExtensionLoadable:
                log.info("%s: debugger extension is NOT loadable", installDir)
                self._lastErrorProxy.setLastError(0,
                    "ActiveState's PHP debugger extension is not loadable.")
                return 0
            else:
                log.info("%s: debugger extension IS loadable", installDir)

        return 1

    def _haveSufficientPHP(self, stopOnFirst=1, feature="Debugging",
                           minVersion=None):
        """Return true iff a PHP installation meeting the above conditions
        can be found.
            "stopOnFirst" is a boolean indicating if the search should stop
                on the "first" found interpreter. This is because things like
                debugging and syntax checking are just going to use this
                one anyway. Default is true.
            "feature" is either "Debugging" (default) or "Syntax Checking".
                This is necessary because syntax checking has slightly
                different requirements.

        If false is returned, then the last error (see koILastErrorService)
        is set with a reason why.
        """
        # Try the user's selected php interpreter.
        phpDefaultInterp = self._prefProxy.prefs.getStringPref("phpDefaultInterpreter")
        if phpDefaultInterp:
            self._phpInfoEx.installationPath =\
                self._phpInfoEx.getInstallationPathFromBinary(phpDefaultInterp)
            if self._isSufficientPHP(self._phpInfoEx, feature,
                                     minVersion):
                return 1
            elif stopOnFirst:
                return 0
            
        # Look on PATH.
        phps = self._sysUtilsProxy.WhichAll('php')
        for php in phps:
            self._phpInfoEx.installationPath =\
                self._phpInfoEx.getInstallationPathFromBinary(php)
            if self._isSufficientPHP(self._phpInfoEx, feature,
                                     minVersion):
                return 1
            elif stopOnFirst:
                return 0
        
        errmsg = "Could not find a suitable PHP installation."
        self._lastErrorProxy.setLastError(0, errmsg)
        return 0

    def _isSufficientPython(self, pythonInfoEx, minVersion=None):
        """Return true iff the given python installation meets the given
        criteria.
            "pythonInfoEx" is a koPythonInfoEx instance loaded with the
                python installation to check.

        If false is returned, then the last error (see koILastErrorService)
        is set with a reason why.
        """
        installDir = pythonInfoEx.installationPath
        exePath = pythonInfoEx.executablePath
        log.info("%s: is sufficient python?", installDir)
        
        if not os.path.exists(exePath):
            log.info("%s: does not exist", exePath)
            self._lastErrorProxy.setLastError(0,
                "Python installation does not exist: \"%s\"" % exePath)
            return 0
        else:
            log.info("%s: exists", exePath)

        if minVersion is not None:
            version = pythonInfoEx.version
            if version < minVersion:
                log.info("%s: %s < %s", installDir, version, minVersion)
                self._lastErrorProxy.setLastError(0,
                    "Insufficient Python version (\"%s\" is version %s). "\
                    "Require at least version %s."\
                    % (installDir, version, minVersion))
                return 0
            else:
                log.info("%s: %s >= %s", installDir, version, minVersion)
        
        return 1

    def _haveSufficientPython(self, stopOnFirst=1, minVersion=None):
        """Return true iff a Python installation meeting the above conditions
        can be found.
            "stopOnFirst" is a boolean indicating if the search should stop
                on the "first" found interpreter. This is because things like
                debugging and syntax checking are just going to use this
                one anyway. Default is true.

        If false is returned, then the last error (see koILastErrorService)
        is set with a reason why.
        """
        # Try the user's selected python interpreter.
        pythonDefaultInterp = self._prefProxy.prefs.getStringPref("pythonDefaultInterpreter")
        if pythonDefaultInterp:
            self._pythonInfoEx.installationPath =\
                self._pythonInfoEx.getInstallationPathFromBinary(pythonDefaultInterp)
            if self._isSufficientPython(self._pythonInfoEx, minVersion):
                return 1
            elif stopOnFirst:
                return 0
            
        # Look on PATH.
        pythons = self._sysUtilsProxy.WhichAll('python')
        for python in pythons:
            self._pythonInfoEx.installationPath =\
                self._pythonInfoEx.getInstallationPathFromBinary(python)
            if self._isSufficientPython(self._pythonInfoEx, minVersion):
                return 1
            elif stopOnFirst:
                return 0
        
        errmsg = "Could not find a suitable Python installation."
        self._lastErrorProxy.setLastError(0, errmsg)
        return 0

    def _isSufficientTDK(self, tdkInfoEx, feature):
        """Return true iff the given TDK installation meets the given
        criteria.
            "tdkInfoEx" is a koTclProInfoEx instance loaded with the
                Tcl installation to check.

        If false is returned, then the last error (see koILastErrorService)
        is set with a reason why.
        """
        installDir = tdkInfoEx.installationPath
        exePath = tdkInfoEx.tclsh_path
        log.info("%s: is sufficient tdk?", installDir)
        
        if not os.path.exists(exePath):
            log.info("%s: does not exist", exePath)
            self._lastErrorProxy.setLastError(0,
                "TDK installation does not exist: \"%s\"" % exePath)
            return 0
        else:
            log.info("%s: exists", exePath)

        if not tdkInfoEx.haveLicense:
            log.info("%s: not licensed", installDir)
            self._lastErrorProxy.setLastError(0,
                "TDK installation is not licensed: \"%s\"" % installDir)
            return 0
        else:
            log.info("%s: licensed", installDir)
            
        return 1

    def _haveSufficientTDK(self, stopOnFirst=0, feature="Debugging"):
        """Return true iff a Tcl Dev Kit installation meeting the above
        conditions can be found.
            "stopOnFirst" is a boolean indicating if the search should stop
                on the "first" found interpreter. This is useful if the actual
                systems (i.e. debugging and syntax checking) will just pick the
                first one anyway without checking suitability. Apparently they
                currently *do* check for suitability so this defaults to false.
            "feature" is either "Debugging" (default) or "Syntax Checking".
                This is necessary because syntax checking has slightly
                different requirements.

        If false is returned, then the last error (see koILastErrorService)
        is set with a reason why.
        """
        # Syntax checking just requires a sufficient 'tclsh' setup.
        # Debugging requires a sufficient 'tclsh' *and* 'wish' setup.
        sufficient = []
        
        # Try the user's selected tclsh interpreter.
        tclshDefaultInterp = self._prefProxy.prefs.getStringPref("tclshDefaultInterpreter")
        if tclshDefaultInterp:
            self._tdkInfoEx.installationPath =\
                self._tdkInfoEx.getInstallationPathFromBinary(tclshDefaultInterp)
            if self._isSufficientTDK(self._tdkInfoEx, feature):
                sufficient.append("tclsh")
            elif stopOnFirst:
                return 0

        # Look on PATH unless already found a sufficient one.
        if "tclsh" not in sufficient:
            tclshs = self._sysUtilsProxy.WhichAll('tclsh')
            for tclsh in tclshs:
                installdir = os.path.dirname(os.path.dirname(tclsh))
                self._tdkInfoEx.installationPath = installdir
                if self._isSufficientTDK(self._tdkInfoEx, feature):
                    sufficient.append("tclsh")
                elif stopOnFirst:
                    return 0
            
        if feature == "Debugging":
            # Try the user's selected tclsh interpreter.
            wishDefaultInterp = self._prefProxy.prefs.getStringPref("wishDefaultInterpreter")
            if wishDefaultInterp:
                self._tdkInfoEx.installationPath =\
                    self._tdkInfoEx.getInstallationPathFromBinary(wishDefaultInterp)
                if self._isSufficientTDK(self._tdkInfoEx, feature):
                    sufficient.append("wish")
                elif stopOnFirst:
                    return 0
    
            # Look on PATH unless already found a sufficient one.
            if "wish" not in sufficient:
                wishs = self._sysUtilsProxy.WhichAll('wish')
                # Must also consider 'wish' executables in the same dir as
                # 'tclsh' executables registered in some way other than the PATH
                # (because that is just the way the Tcl installer is).
                for tclsh in tclshs:
                    base, ext = os.path.splitext(tclsh)
                    wish = os.path.join(os.path.dirname(base), 'wish') + ext
                    if os.path.exists(wish) and wish not in wishs:
                        wishs.insert(0, wish)
                for wish in wishs:
                    installdir = os.path.dirname(os.path.dirname(wish))
                    self._tdkInfoEx.installationPath = installdir
                    if self._isSufficientTDK(self._tdkInfoEx, feature):
                        sufficient.append("wish")
                    elif stopOnFirst:
                        return 0

        if feature == "Debugging" and "tclsh" in sufficient and\
           "wish" in sufficient:
            return 1
        elif feature == "Syntax Checking" and "tclsh" in sufficient:
            return 1

        errmsg = "Could not find a suitable TDK installation."
        self._lastErrorProxy.setLastError(0, errmsg)
        return 0

    def _isSufficientTcl(self, tclInfoEx):
        """Return true iff the given Tcl installation meets the given
        criteria.
            "tclInfoEx" is a koTclInfoEx instance loaded with the
                Tcl installation to check.

        If false is returned, then the last error (see koILastErrorService)
        is set with a reason why.
        """
        installDir = tclInfoEx.installationPath
        exePath = tclInfoEx.tclsh_path
        log.info("%s: is sufficient tcl?", installDir)
        
        if not os.path.exists(exePath):
            log.info("%s: does not exist", exePath)
            self._lastErrorProxy.setLastError(0,
                "Tcl installation does not exist: \"%s\"" % exePath)
            return 0
        else:
            log.info("%s: exists", exePath)

        return 1

    def _haveSufficientTcl(self, stopOnFirst=0):
        """Return true iff a Tcl installation meeting the above
        conditions can be found.
            "stopOnFirst" is a boolean indicating if the search should stop
                on the "first" found interpreter. This is useful if the actual
                systems (i.e. debugging and syntax checking) will just pick the
                first one anyway without checking suitability. Apparently they
                currently *do* check for suitability so this defaults to false.

        If false is returned, then the last error (see koILastErrorService)
        is set with a reason why.
        """
        sufficient = []
        
        # Try the user's selected tclsh interpreter.
        tclshDefaultInterp = self._prefProxy.prefs.getStringPref("tclshDefaultInterpreter")
        if tclshDefaultInterp:
            self._tclInfoEx.installationPath =\
                self._tclInfoEx.getInstallationPathFromBinary(tclshDefaultInterp)
            if self._isSufficientTcl(self._tclInfoEx):
                sufficient.append("tclsh")
            elif stopOnFirst:
                return 0

        # Look on PATH unless already found a sufficient one.
        if "tclsh" not in sufficient:
            tclshs = self._sysUtilsProxy.WhichAll('tclsh')
            for tclsh in tclshs:
                installdir = os.path.dirname(os.path.dirname(tclsh))
                self._tclInfoEx.installationPath = installdir
                if self._isSufficientTcl(self._tclInfoEx):
                    sufficient.append("tclsh")
                elif stopOnFirst:
                    return 0
            
        # Try the user's selected tclsh interpreter.
        wishDefaultInterp = self._prefProxy.prefs.getStringPref("wishDefaultInterpreter")
        if wishDefaultInterp:
            self._tclInfoEx.installationPath =\
                self._tclInfoEx.getInstallationPathFromBinary(wishDefaultInterp)
            if self._isSufficientTcl(self._tclInfoEx):
                sufficient.append("wish")
            elif stopOnFirst:
                return 0

        # Look on PATH unless already found a sufficient one.
        if "wish" not in sufficient:
            wishs = self._sysUtilsProxy.WhichAll('wish')
            # Must also consider 'wish' executables in the same dir as
            # 'tclsh' executables registered in some way other than the PATH
            # (because that is just the way the Tcl installer is).
            for tclsh in tclshs:
                base, ext = os.path.splitext(tclsh)
                wish = os.path.join(os.path.dirname(base), 'wish') + ext
                if os.path.exists(wish) and wish not in wishs:
                    wishs.insert(0, wish)
            for wish in wishs:
                installdir = os.path.dirname(os.path.dirname(wish))
                self._tclInfoEx.installationPath = installdir
                if self._isSufficientTcl(self._tclInfoEx):
                    sufficient.append("wish")
                elif stopOnFirst:
                    return 0

        if "tclsh" in sufficient and "wish" in sufficient:
            return 1
        else:
            errmsg = "Could not find a suitable Tcl installation."
            self._lastErrorProxy.setLastError(0, errmsg)
            return 0

