import os.path
import urlparse

from xpcom import components, nsError, ServerException
from koLogger import logError, logWarning, logDebug
import testable
import sys

class koFileSpec:
    _com_interfaces_ = [components.interfaces.koIFileSpec]
    _reg_clsid_ = "{61D52288-1E47-4b39-B7B7-ABFBFA44F03E}"
    _reg_contractid_ = "@activestate.com/koFileSpec;1"
    _reg_desc_ = "Komodo File Specification"

    def __init__(self):
        self.set_url(None)

    def get_url(self):
        return self.__url

    def set_url(self, url):
        #XXX Hack: Mozilla expects URLs of UNC paths to have five slashes
        #XXX but builds them with four.  We fix them up here.
        if url and (url.startswith("file:////") and not url.startswith("file://///")):
            url = url.replace("file:////", "file://///", 1)
            
        self.__url = url

    def get_path(self):
        fileURL = components.classes["@mozilla.org/network/standard-url;1"] \
                  .createInstance(components.interfaces.nsIFileURL)
        fileURL.spec = self.get_url()
        try: path = fileURL.file.path
        except: path = None
        return path
     
    def set_path(self, path):
        localFile = components.classes["@mozilla.org/file/local;1"] \
                .createInstance(components.interfaces.nsILocalFile)
        localFile.initWithPath(path);
        fileURL = components.classes["@mozilla.org/network/standard-url;1"] \
                .createInstance(components.interfaces.nsIFileURL)
        fileURL.file = localFile
        self.set_url(fileURL.spec)

    def get_shortName(self):
        localPath = self.get_path()
        if localPath:
            shortName = os.path.basename(localPath)
        else:
            urlPathParts = urlparse.urlparse(self.get_url())[2].split('/')
            shortName = urlPathParts[-1]
        return shortName

    def get_longName(self):
        longName = self.get_path()
        if not longName:
            longName = self.get_url()
        return longName


class TestKoFileSpec(testable.MultiTestable):
    _com_interfaces_ = [components.interfaces.koITestable,
                        components.interfaces.koIMultiTestable]
    _reg_clsid_ = "{21E8E05B-6E42-479f-B3E6-CFB90D7249EA}"
    unitName = "koFileSpec"
    _reg_contractid_ = "@activestate.com/koTestable-" + unitName + ";1"

    def __init__(self):
        self.__fileSpec = \
                        components.classes["@activestate.com/koFileSpec;1"] \
                        .createInstance(components.interfaces.koIFileSpec)
        # we need to know the platform to determine which tests to run
        self.__infoService = components.classes["@activestate.com/koInfoService;1"] \
                       .getService(components.interfaces.koIInfoService)


    def test_assignAndRetrieveURL(self):
        url = "http://www.activestate.com"
        self.__fileSpec.url = url
        if self.__fileSpec.url == url:
            return "PASS"
        else:
            return "FAIL: assignAndRetrieveURL expected: %s, got: %s" \
                   % (url, self.__fileSpec.url)
                   
 
    def test_assignAndRetrievePath(self):
        if self.__infoService.platform == "win32":
            path = r"C:\Documents and Settings\test\My Documents\somefile.txt"
        else:
            path = "/home/test/somefile.txt"

        self.__fileSpec.path = path
        if self.__fileSpec.path == path:
            return "PASS"
        else:
            return "FAIL: assignAndRetrievePath expected: %s, got: %s" \
                   % (path, self.__fileSpec.path)


    def test_determinePathFromFileURL(self):
        if self.__infoService.platform == "win32":
            url = "file:///C:/Documents%20and%20Settings/test/My%20Documents/somefile.txt"
            path = r"C:\Documents and Settings\test\My Documents\somefile.txt"
        else:
            url = "file:///home/test/somefile.txt"
            path = "/home/test/somefile.txt"

        self.__fileSpec.url = url
        if self.__fileSpec.path == path:
            return "PASS"
        else:
            return "FAIL: determinePathFromFileURL expected: %s, got: %s" \
                   % (path, self.__fileSpec.path)


    def test_determineFileURLFromPath(self):
        if self.__infoService.platform == "win32":
            url = "file:///C:/Documents%20and%20Settings/test/My%20Documents/somefile.txt"
            path = r"C:\Documents and Settings\test\My Documents\somefile.txt"
        else:
            url = "file:///home/test/somefile.txt"
            path = "/home/test/somefile.txt"

        self.__fileSpec.path = path
        if self.__fileSpec.url == url:
            return "PASS"
        else:
            return "FAIL: determineFileURLFromPath expected: %s, got: %s" \
                   % (url, self.__fileSpec.url)


    def test_assignAndRetrieveUNCPath(self):
        if self.__infoService.platform == "win32":
            path = r"\\americano\home\Apps\Komodo\somefile.txt"
            self.__fileSpec.path = path
            if self.__fileSpec.path == path:
                return "PASS"
            else:
                return "FAIL: assignAndRetrieveUNCPath expected: %s, got: %s" % (path, self.__fileSpec.path)
        else:
            return "SKIP"


    def test_determineUNCPathFromFileURL(self):
        if self.__infoService.platform == "win32":
            path = r"\\americano\home\Apps\Komodo\somefile.txt"
            url = "file://///americano/home/Apps/Komodo/somefile.txt"
            self.__fileSpec.url = url
            if self.__fileSpec.path == path:
                return "PASS"
            else:
                return "FAIL: determineUNCPathFromFileURL expected: %s, got: %s" % (path, self.__fileSpec.path)
        else:
            return "SKIP"

    def test_determineFileURLFromUNCPath(self):
        if self.__infoService.platform == "win32":
            path = r"\\americano\home\Apps\Komodo\somefile.txt"
            url = "file://///americano/home/Apps/Komodo/somefile.txt"
            self.__fileSpec.path = path
            if self.__fileSpec.url == url:
                return "PASS"
            else:
                return "FAIL: determineFileURLFromUNCPath expected: %s, got: %s" % (url, self.__fileSpec.url)
        else:
            return "SKIP"
            




