#!python
# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.
#
# Shane Caraveo
#
# Provide file status information
# runs background thread to periodicaly update status on a file
# has plugin capability to add new types of status retreivers (CVS, Perforce)
#
# observer listens for:
#   view_closed (url)
#   view_opened (url)
#   file_changed (url)
#   file_added (url)
#   file_removed (url)
#   file_project (path)
#
# observer notifications:
#   file_status (url)
#   file_status_updated (None)
#
# when a file status changes, the service sends a notification file_status_update.
# any observer that listens for this will receive the url of the file that has
# been updated.  It must then call back to the service to get the status
# array.
#

from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC
from koLogger import logError, logWarning, logDebug
import convertURL
import sys, os, re, strptime, time, stat, string
import threading


class fileStatus:
    def __init__(self, url):
        self.url = url
        self.path = convertURL.LocalPathFromURL(url)
        self.basename = ''
        self.dirname = ''
        self.isdir = 0
        if self.path:
            self.path = os.path.normpath(self.path)
            self.isdir = os.path.isdir(self.path)
            if self.isdir:
                self.dirname = self.path
            else:
                self.dirname = os.path.dirname(self.path)
        self.update = 0
        self.open = 0
        self.persist = 0
        self.remove = 0
        self.dofileupdate = 0
        self.status = {}
    
class KoFileStatusService:
    _com_interfaces_ = [components.interfaces.koIFileStatusService, components.interfaces.nsIObserver]
    _reg_clsid_ = "{20732408-43DA-4ca2-BC9F-B82437A3CB2B}"
    _reg_contractid_ = "@activestate.com/koFileStatusService;1"
    _reg_desc_ = "Komodo File Status Service"

    def __init__(self):
        self._proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"].\
            getService(components.interfaces.nsIProxyObjectManager)
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._observerProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.nsIObserverService, self._observerSvc,
            PROXY_ALWAYS | PROXY_ASYNC)
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
        
        # set up data cache
        self._urllist = {}
        self._statusCheckers = []
        self.shutdown = 0
        self._continue = 0
        self._background = 0
        self._diskChecker = None
        
        # set up the background thread
        self._thread = None
        self._tlock = threading.Lock()
        self._cv = threading.Condition()
        self._shutdown = threading.Condition()

    def init(self):
        self._observerSvc.AddObserver(self, "view_closed")
        self._observerSvc.AddObserver(self, "view_opened")
        self._observerSvc.AddObserver(self, "file_changed")
        self._observerSvc.AddObserver(self, "file_status_now")
        self._observerSvc.AddObserver(self, "file_added")
        self._observerSvc.AddObserver(self, "file_removed")
        self._observerSvc.AddObserver(self, "file_project")
        self._observerSvc.AddObserver(self, "file_update_now")
        
        self._diskChecker = KoDiskFileChecker()
        self._statusCheckers.append(self._diskChecker)
        self._statusCheckers.append(KoCVSFileChecker())
        self._statusCheckers.append(KoP4FileChecker())
        
        self._globalPrefs.AddObserver(self)

        self._tlock.acquire()
        self._thread = threading.Thread(target = KoFileStatusService._process,
                                     args = (self,))
        self._thread.start()
        self._tlock.release()

    def unload(self):
        self.shutdown = 1;
        if self._thread.isAlive():
            self._cv.acquire()
            self._cv.notify()
            self._cv.release()
            self._shutdown.acquire()
            self._shutdown.wait(5)
            self._shutdown.release()
        self._observerSvc.RemoveObserver(self, "view_closed")
        self._observerSvc.RemoveObserver(self, "view_opened")
        self._observerSvc.RemoveObserver(self, "file_changed")
        self._observerSvc.RemoveObserver(self, "file_status_now")
        self._observerSvc.RemoveObserver(self, "file_added")
        self._observerSvc.RemoveObserver(self, "file_removed")
        self._observerSvc.RemoveObserver(self, "file_project")
        self._observerSvc.RemoveObserver(self, "file_update_now")
    
    def notifyError(self, error, detail):
        # XXX 'scc_status_message' is evil (see bug 20568). In addition, it
        #     is poorly named because the code is structured such that the
        #     non-scc checkers here could set an error.
        if error:
            #print "attempting to send notification"
            try:
                self._observerProxy.Notify(error, 'scc_status_message', detail)
            except COMException, e:
                pass
        
    def _initUrl(self, url):
        if not self._urllist.has_key(url):
            self._urllist[url] = fileStatus(url)
    
    def _getFileStatusNow(self, url):
        lasterror = None
        self._initUrl(url)
        urllist = [self._urllist[_url] for _url in self._urllist.keys() if _url.find(url)==0]
        # force an update to the status
        for checker in self._statusCheckers:
            if not checker.isActive():
                continue

            if checker.scc and (('scc' in self._urllist[url].status and \
                self._urllist[url].status['scc'] and \
                self._urllist[url].status['scc'] != checker.scc) or \
                ('scc_can' in self._urllist[url].status and \
                self._urllist[url].status['scc_can'] and \
                self._urllist[url].status['scc_can'] != checker.scc)):
                continue

            # invalidate all the url's
            checker.invalidateUrl(self._urllist[url],self._urllist[url].isdir)
            for _url in urllist:
                checker.invalidateUrl(_url)
                
            # only check for the primary url
            checker.getFileStatus(self._urllist[url],not self._urllist[url].isdir)
            if checker.error and (not lasterror or checker.error.msg != lasterror.msg):
                lasterror = checker.error
                self.notifyError(checker.error,checker.error_detail)
                checker.error = None
            
            # reset the status indicators
            for _url in urllist:
                checker.getFileStatus(_url)
        
        # send notifications of status updates
        for _url in urllist:
            try:
                self._observerSvc.Notify(self, 'file_status', _url.url)
            except COMException, e:
                if e.errno == nsError.NS_ERROR_FAILURE:
                    # if noone is listening, we get an exception
                    pass
                else:
                    #print "caught unknown exception in notify"
                    raise            
        try:
            self._observerSvc.Notify(self, 'file_status_updated',None)
        except COMException, e:
            if e.errno == nsError.NS_ERROR_FAILURE:
                # if noone is listening, we get an exception
                pass
            else:
                #print "caught unknown exception in notify"
                raise

    def hasFileChanged(self,url):
        changed = 0
        if self._urllist.has_key(url) and \
           self._urllist[url].update:
            changed = self._diskChecker.hasChanged(self._urllist[url])
        return changed

    def getFileStatus(self, url):
        if self._urllist.has_key(url) and self._urllist[url].status:
            status = self._urllist[url].status
            keys = map(str, status.keys())
            values = map(str, status.values())
            return keys,values
        return [],[]

    def isRepository(self, url):
        if url in self._urllist:
            if self._urllist[url].status.get('scc',''):
                return self._urllist[url].status['scc']
            if self._urllist[url].status.get('scc_can',''):
                return self._urllist[url].status['scc_can']
        path = convertURL.LocalPathFromURL(url)
        if not path:
            return ''
        path = os.path.normpath(path)
        if not os.path.exists(path):
            return ''
        
        for checker in self._statusCheckers:
            if not checker.isActive() or not checker.scc:
                continue
            if checker.isRepository(path):
                return checker.scc

        return ''
    
    # nsIObserver interface
    def Observe(self, subject, topic, url):
        if topic == self._globalPrefs.id:
            # url is actually the pref name that was changed
            for checker in self._statusCheckers:
                if url == checker.enabledPrefName or \
                   url == checker.backgroundPrefName:
                    enabled = self._globalPrefs.getBooleanPref(checker.enabledPrefName)
                    background = self._globalPrefs.getBooleanPref(checker.backgroundPrefName)
                    if enabled != checker.enabled or background != checker.background:
                        self._cv.acquire()
                        self._cv.notify()
                        self._cv.release()
                        break
            # we don't continue in this observer for pref changes
            return
        
        #print "status observer received ",topic,url
        _url = None
        if url:
            self._initUrl(url)
            _url = self._urllist[url]
            
        if topic == 'view_closed':
            _url.open = 0
            if _url.persist <= 0:
                # remove from our list
                _url.remove = 1
        elif topic == "view_opened":
            _url.open = 1
            _url.remove = 0
        elif topic == "file_project":
            _url.persist += 1
            _url.update = 0
            _url.remove = 0
        elif topic == "file_status_now":
            #_url.update = 0
            self._getFileStatusNow(url)
            return
        elif topic == "file_changed":
            _url.update = 0
            _url.dofileupdate = 1
        elif topic == "file_added":
            _url.persist += 1
            _url.update = 0
            _url.remove = 0
        elif topic == "file_removed":
            _url.persist -= 1
            if _url.persist <= 0 and not _url.open:
                # remove from our list
                _url.remove = 1

        self._cv.acquire()
        # notify is an immediate release, but we want the thread
        # to continue until there is no more activity, so we flag
        # it to do so
        self._continue = 1
        self._cv.notify()
        self._cv.release()

    def _process(self):
        #print "starting file status background thread"
        urllist = []
        lasterror = None
        while 1:
            try:
                updated = []
                #print "file status waiting to do update..."
                
                # if we have more url's than we checked last
                # time, lets do them now so the user doesn't
                # have to wait 30 seconds
                if not self._continue:
                    self._cv.acquire()
                    self._cv.wait(30)
                    # if self._continue is set, this is not a timeout
                    self._background = not self._continue
                    self._cv.release()
                # reset flag, see Observer for more info
                self._continue = 0
                
                if self.shutdown:
                    self._shutdown.acquire()
                    self._shutdown.notify()
                    self._shutdown.release()
                    #print "status thread shutting down"
                    return
                
                
                #print "doing file status update"
                urllist = self._urllist.keys()
                
                start = time.time()
                
                # invalidate old status information
                # we do this seperately so that the scc modules can
                # catch full directories of urls
                for checker in self._statusCheckers:
                    urls = []
                    
                    if not checker.isActive():
                        continue
                    checker.background = self._globalPrefs.getBooleanPref(checker.backgroundPrefName)
                    checker.recursive = checker.recursivePrefName and self._globalPrefs.getBooleanPref(checker.recursivePrefName)
                    checker.duration = self._globalPrefs.getLongPref(checker.durationPrefName)
                    if self._background and not checker.background: continue
                    
                    # invalidate paths first so they can be recached
                    # build a list of url's that we can actually update
                    for url in urllist:
                        if self._urllist[url].update == 0 or \
                           (checker.background and checker.invalidatePath(self._urllist[url])):
                            # if this file is marked to have scc, and it's scc does
                            # not match the checker, then skip passing the file to the checker
                            if not (checker.scc and (('scc' in self._urllist[url].status and \
                               self._urllist[url].status['scc'] and \
                               self._urllist[url].status['scc'] != checker.scc) or \
                                ('scc_can' in self._urllist[url].status and \
                               self._urllist[url].status['scc_can'] and \
                               self._urllist[url].status['scc_can'] != checker.scc))):
                                urls.append(self._urllist[url])

                    for url in urls:
                        # if we're in process, but a shutdown is requested, do it
                        if self.shutdown:
                            self._shutdown.acquire()
                            self._shutdown.notify()
                            self._shutdown.release()
                            print "status thread shutting down"
                            return
                        
                        # check the status of the url
                        done = 0
                        #print "url time is %d, now is %d" %(self._urllist[url].update, time.time())
                        
                        #print "%s checking %s %s" %(checker.scc,url.status['scc'],url.url)
                        if url.dofileupdate:
                            checker.invalidateUrl(url)
                            
                        if not checker.getFileStatus(url,url.dofileupdate):
                            if checker.error and (not lasterror or checker.error.msg != lasterror.msg):
                                lasterror = checker.error
                                self.notifyError(checker.error,checker.error_detail)
                                checker.error = None
                            continue
                        
                        updated.append(url)

                        # prevent the missing status from being enabled if the file
                        # is in scc, let the scc properties be important
                        if not checker.scc and 'missing' in url.status \
                           and 'scc' in url.status and \
                           url.status['scc']:
                            url.status['missing'] = '0'
                            
                        try:
                            self._observerProxy.Notify(self, 'file_status',url.url)
                        except COMException, e:
                            if e.errno == nsError.NS_ERROR_FAILURE:
                                # if noone is listening, we get an exception
                                pass
                        
                    # on every checker we notify, since a particular checker
                    # can take some time
                    if updated:
                        try:
                            # this call is proxied to the UI thread.  The outliners are invalidated
                            # in the observer for this call
                            self._observerProxy.Notify(self, 'file_status_updated',None)
                        except COMException, e:
                            if e.errno == nsError.NS_ERROR_FAILURE:
                                # if noone is listening, we get an exception
                                pass
                
                lasterror = None
                
                for url in urls:
                    url.dofileupdate = 0
                        
                end =  time.time()
                for url in updated:
                    url.update = end

                # remove url's that have been closed
                for url in self._urllist.keys():
                    if self._urllist[url].remove:
                        del self._urllist[url]
                        
            except:
                # we catch any exception so we can clear our thread variable
                # this allows us to restart the thread again later
                # we re raise the exception
                import traceback
                errmsg = traceback.format_exception(*sys.exc_info())
                logDebug('KoFileStatusService thread exception %s' % errmsg)

        print "Status Thread ending"

class KoFileCheckerBase:
    def __init__(self):
        self.type = None
        self.scc = None
        self.duration = 15
        self.durationPrefName = None
        self.enabledPrefName = None
        self.enabled = 0
        self.backgroundPrefName = None
        self.background = 0
        self.recursivePrefName = None
        self.recursive = 0
        self.error = None
        self.error_detail = None
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
    
    def setError(self, msg='', detail=''):
        self.error = None
        if not msg: return
        
        self.error = components.classes["@activestate.com/koStatusMessage;1"]\
                .createInstance(components.interfaces.koIStatusMessage)
        self.error.category = "scc_error_message"
        self.error.msg = msg
        self.error.timeout = 5000
        self.error.highlight = 1
        self.error_detail = str(detail)
    
    def isRepository(self, path):
        return 0
    
    def isActive(self):
        self.enabled = self._globalPrefs.getBooleanPref(self.enabledPrefName)
        return self.enabled

    def hasChanged(self, url):
        return 0

    def invalidatePath(self, url,force=0):
        return 0

    def invalidateUrl(self, url,force=0):
        return 0

    def getFileStatus(self, url, dofile=1):
        return None
    
    
class KoDiskFileChecker(KoFileCheckerBase):
    def __init__(self):
        KoFileCheckerBase.__init__(self)
        self.type = 'disk'
        self.durationPrefName = 'diskBackgroundMinutes'
        self.enabledPrefName = 'diskStatusEnabled'
        self.backgroundPrefName = 'diskBackgroundCheck'

    def hasChanged(self, url):
        if not url.path: return 0
        _url = fileStatus(url.url)
        self.getFileStatus(_url)
        if _url.status['missing']: return -1
        for item in _url.status.keys():
            if not item in ['mtime','size','fileread','filewrite',\
                     'readonly','readwrite','missing']: continue
            if not item in url.status or \
               str(url.status[item]) != str(_url.status[item]):
                return 1
        return 0

    def invalidatePath(self, url,force=0):
        return force or url.update < time.time() - (self.duration * 60)

    def invalidateUrl(self, url,force=0):
        return force or url.update < time.time() - (self.duration * 60)

    def getFileStatus(self, url, dofile=1):
        self.setError()
        # convert to a local file
        items = {'mode':'','ino':'','dev':'','nlink':'',\
                 'uid':'','gid':'','size':'','atime':'',\
                 'mtime':'','ctime':'',"fileread":1,
                 "filewrite":1,"fileexec":0,"readonly":0,
                 "readwrite":1,'missing':0}
        path = url.path
        if not path:
            if not url.url.startswith('koremote://'):
                items['readonly']=1
            url.status.update(items)
            return None
        #print "getting STAT for ",repr(path)
        try:
            stats = os.stat(path)
            #mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
            items['mode'] = stats[stat.ST_MODE]
            items['ino'] = stats[stat.ST_INO]
            items['dev'] = stats[stat.ST_DEV]
            items['nlink'] = stats[stat.ST_NLINK]
            items['uid'] = stats[stat.ST_UID]
            items['gid'] = stats[stat.ST_GID]
            items['size'] = stats[stat.ST_SIZE]
            items['atime'] = stats[stat.ST_ATIME]
            items['mtime'] = stats[stat.ST_MTIME]
            items['ctime'] = stats[stat.ST_CTIME]

            items["fileread"] = os.access(path, os.R_OK)
            items["filewrite"] = os.access(path, os.W_OK)
            items["fileexec"] = os.access(path, os.X_OK)
            
            items["readonly"] = items["fileread"] and not items["filewrite"]
            items["readwrite"] = items["fileread"] and items["filewrite"]

            items['missing'] = 0
                
        except OSError,e:
            if e.errno != 2:
                logDebug("File %s exception errno %d" %(path,e.errno))
                
            # xxx allow this to go through as a missing file
            items['missing'] = 1
            
        url.status.update(items)
        return 1


import p4lib
class KoP4FileChecker(KoFileCheckerBase):
    def __init__(self):
        KoFileCheckerBase.__init__(self)
        self._have = {}
        self._open = {}
        self._files = {}
        self._not = {}
        #XXX rename _paths to _directories
        self._paths = {}
        self.p4 = p4lib.P4()
        self.type = 'scc'
        self.scc = 'p4'
        self.durationPrefName = 'p4BackgroundMinutes'
        self.enabledPrefName = 'p4Enabled'
        self.backgroundPrefName = 'p4BackgroundCheck'
        self.recursivePrefName = 'p4Recursive'
        self._p4Svc = components.classes["@activestate.com/koP4;1"].\
            getService(components.interfaces.koIP4)

    def notifyError(self, e):
        if str(e).find("is not under client's root") != -1:
            return
        logDebug(e)
        self.setError('P4 Status Error, see SCC Output pane for details',e)

    # we don't cache the where information into _path here, since that
    # would prevent getFileStatus from running correctly (ie retrieving
    # files/have/open data)
    def isRepository(self, path):
        npath = os.path.normcase(path)
        if npath in self._not: return 0
        if npath in self._paths: return 1
        try:
            basename = path
            if not os.path.isdir(path):
                basename = os.path.dirname(path)
            if self.p4.where(path,cwd=basename): return 1
        except p4lib.P4LibError, e:
            self.notifyError(e)
            # XXX what is right? should we _not it?
            return 0
        self._not[npath] = time.time()
        return 0
    
    def isActive(self):
        self.enabled = self._globalPrefs.getBooleanPref(self.enabledPrefName)
        return self._p4Svc.isFunctional and self.enabled

    # returns true if the path is invalidated
    def invalidatePath(self, url, force=0):
        if not url.path:
            return 0
        if url.isdir:
            return self._invalidate(url.path, force)
        else:
            return self._invalidate(url.dirname, force)
    
    # returns true if the path is invalidated    
    def invalidateUrl(self, url, force=0):
        if not url.path:
            return 0
        return self._invalidate(url.path, force)
        
    def _invalidate(self, path, force):
        npath = os.path.normcase(path)
        if npath in self._paths:
            # this will force a refetch of the entire path
            # if the last stat on the path is older than
            # the duration
            lastcheck = self._paths[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._paths[npath] = 0
            else:
                return 0
        elif npath in self._not:
            lastcheck = self._not[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._not[npath] = 0
            else:
                return 0
        elif npath in self._have:
            # this will force a refetch of just the one file
            depotFile = self._have[npath]['depotFile']
            if self._files.has_key(depotFile):
                del self._files[depotFile]
            if self._open.has_key(depotFile):
                del self._open[depotFile]
            del self._have[npath]
        return 1
    
    def getFileStatus(self, url, dofile=0):
        # initialize all items.  This helps to clear the item properties
        # when it is removed from scc
        items = {'scc':'','scc_can':'','scc_exclude':'','haveOnDisk':'','myRevision':'','revdate':'',\
                 'depotRevision':'','needSync':'0','sync':'0','ok':'0',\
                 'conflict':'0','action':'','status':'','change':''}
        
        # we use nbasepath and npath for keys, we use
        # path and basepath for sending to P4 commands
        # this is because P4 is case sensitive
        basepath = path = url.path
        if not path:
            url.status.update(items)
            return None
        nbasepath = npath = os.path.normcase(path)
        #print "CHECKING p4 for ",basepath
        if not url.isdir:
            basepath = url.dirname
            nbasepath = os.path.normcase(basepath)

        try:
            if nbasepath in self._not:
                url.status.update(items)
                return None
                
            if (nbasepath not in self._paths or \
                self._paths[nbasepath] == 0) and \
                os.path.isdir(basepath):
                # execute p4 on the path
                if self.recursive:
                    subpath = "/..."
                else:
                    subpath = "/*"
                    
                where = self.p4.where(path,cwd=basepath)
                if where:
                    # store the basepath to signal that this
                    # is a perforce directory
                    items['scc_can']=items['scc']=self.scc
                else:
                    self._not[nbasepath]=time.time()
                    url.status.update(items)
                    return None
                    
                #print "p4 checking ",basepath + subpath
                p4have = self.p4.have(basepath + subpath,
                                      cwd=basepath)
                for file in p4have:
                    filekey = os.path.normcase(file['localFile'])
                    path = os.path.normcase(os.path.dirname(filekey))
                    self._paths[path]=time.time()
                    file['ondisk'] = os.path.exists(filekey);
                    self._have[filekey] = file
                
                p4files = self.p4.files(basepath + subpath,cwd=basepath)
                for file in p4files:
                    self._files[file['depotFile']] = file

                p4files = self.p4.opened(basepath + subpath,cwd=basepath)
                for file in p4files:
                    self._open[file['depotFile']] = file
                    if not file['depotFile'] in self._files or \
                        self._files[file['depotFile']]['action'] == 'delete':
                        where = self.p4.where(file['depotFile'],cwd=basepath)
                        if len(where):
                            self._files[file['depotFile']] = file
                            filekey = os.path.normcase(where[0]['localFile'])
                            file['localFile'] = where[0]['localFile']
                            file['ondisk'] = os.path.exists(filekey);
                            self._have[filekey] = file
                
                self._paths[nbasepath]=time.time()

            if dofile and npath not in self._have and not os.path.isdir(npath):
                #print "checking p4 for ",path
                # file not in cache, get it now and stuff it into the cache
                p4have = self.p4.have(path,cwd=basepath)
                for file in p4have:
                    filekey = os.path.normcase(file['localFile'])
                    file['ondisk'] = os.path.exists(filekey);
                    self._have[filekey] = file
                
                p4files = self.p4.files(path,cwd=basepath)
                for file in p4files:
                    self._files[file['depotFile']] = file

                p4files = self.p4.opened(path,cwd=basepath)
                for file in p4files:
                    self._open[file['depotFile']] = file
                    if not file['depotFile'] in self._files or \
                        self._files[file['depotFile']]['action'] == 'delete':
                        where = self.p4.where(file['depotFile'],cwd=basepath)
                        if len(where):
                            self._files[file['depotFile']] = file
                            filekey = os.path.normcase(where[0]['localFile'])
                            file['localFile'] = where[0]['localFile']
                            file['ondisk'] = os.path.exists(filekey);
                            self._have[filekey] = file
                        
        
        except p4lib.P4LibError, e:
            self.notifyError(e)
            if nbasepath in self._paths:
                # this ensures that a path that IS in a repository, but contains
                # no files will be marked correctly
                items['scc_can']=self.scc
            url.status.update(items)
            return None

        if not self._have.has_key(npath):
            # still don't have it, not a perforce file
            if nbasepath in self._paths:
                # this ensures that a path that IS in a repository, but contains
                # no files will be marked correctly
                items['scc_can']=self.scc
            url.status.update(items)
            return None
        
        # get p4 info for the path, not the file
        
        have = self._have[npath]
        #print "getting status for ",repr(have['depotFile'])
        
        items['scc'] =  'p4'
        items['haveOnDisk']= have['ondisk']
        items['myRevision'] = have['rev']
        if have['depotFile'] in self._files and 'rev' in self._files[have['depotFile']]:
            items['depotRevision'] = self._files[have['depotFile']]['rev']
        items['needSync'] = items['myRevision'] != items['depotRevision']
        items['sync'] = '0'
        items['ok'] = '0'
        items['conflict'] = '0'
        
        if items['needSync']:
            items['status'] = 'out of date'
            items['sync'] = '1'
        else:
            items['status'] = 'up to date'
            items['ok'] = '1'
        
        if self._open.has_key(have['depotFile']):
            open = self._open[have['depotFile']]
            items['action'] = open['action']
            items['change'] = open['change']
        else:
            items['action'] = ''
            items['change'] = ''
            
        url.status.update(items)
        return 1


import cvslib
class KoCVSFileChecker(KoFileCheckerBase):
    def __init__(self):
        KoFileCheckerBase.__init__(self)
        self._files = {}
        # XXX change _paths to _directories
        self._paths = {}
        self._not = {}
        self.cvs = cvslib.CVS()
        self.type = 'scc'
        self.scc = 'cvs'
        self.duration = 15
        self.durationPrefName = 'cvsBackgroundMinutes'
        self.enabledPrefName = 'cvsEnabled'
        self.backgroundPrefName = 'cvsBackgroundCheck'
        self.recursivePrefName = 'cvsRecursive'
        self._cvsSvc = components.classes["@activestate.com/koCVS;1"].\
            getService(components.interfaces.koICVS)

    # we don't cache the where information into _path here, since that
    # would prevent getFileStatus from running correctly (ie retrieving
    # files/have/open data)
    def isRepository(self, path):
        basepath = os.path.dirname(path)
        npath = os.path.normcase(basepath)
        if npath in self._not: return 0
        if npath in self._paths: return 1
        cvsRepositoryFile = os.path.join(basepath,'CVS','Repository')
        if os.path.exists(cvsRepositoryFile): return 1
        self._not[npath]=time.time()
        return 0
        
    def isActive(self):
        self.enabled = self._globalPrefs.getBooleanPref(self.enabledPrefName)
        return self._cvsSvc.isFunctional and self.enabled

    def invalidatePath(self, url, force=0):
        if not url.path:
            return 0
        if url.isdir:
            return self._invalidate(url.path, force)
        else:
            return self._invalidate(url.dirname, force)
        
    def invalidateUrl(self, url, force=0):
        if not url.path:
            return 0
        return self._invalidate(url.path, force)
        
    def _invalidate(self, path, force):
        #print "invalidating ",path
        npath = os.path.normcase(path)
        if npath in self._paths:
            # this will force a refetch of the entire path
            # if the last stat on the path is older than
            # the duration
            lastcheck = self._paths[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._paths[npath] = 0
            else:
                return 0
        elif npath in self._not:
            lastcheck = self._not[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._not[npath] = 0
            else:
                return 0
        elif npath in self._files:
            del self._files[npath]
        return 1

    def _raiseCVSError(self, msg):
        # since we have no way to know if this is
        # informational or not, we have to parse for
        # what we want to know
        if msg.find('authorization failed') > -1:
            raise cvslib.CVSLibError(msg)
    
    def getFileStatus(self, url, dofile=0):
        # initialize all items.  This helps to clear the item properties
        # when it is removed from scc
        items = {'scc':'','scc_can':'','scc_exclude':'','haveOnDisk':'','myRevision':'','revdate':'',\
                 'depotRevision':'','needSync':'0','sync':'0','ok':'0',\
                 'action':'','status':'','conflict':'0'}

        #print "getFileStatus CVS %d %s"%(dofile,url.url)
        basepath = path = url.path
        if not path:
            url.status.update(items)
            return None
        nbasepath = npath = os.path.normcase(path)
        
        if not url.isdir:
            basepath = url.dirname
            nbasepath = os.path.normcase(basepath)
        
        try:
            if nbasepath in self._not:
                url.status.update(items)
                return None

            # cvs is recursive for status, so lets make sure that both the file
            # and the path do not exist before we stat a path
            if (nbasepath not in self._paths or \
                self._paths[nbasepath] == 0) and \
                os.path.isdir(basepath):
                
                # cvs status does not return status on paths, so we
                # have to build a status for the path :(
                cvsRepositoryFile = os.path.join(basepath,'CVS','Repository')
                if not os.path.exists(cvsRepositoryFile):
                    self._not[nbasepath]=time.time()
                    url.status.update(items)
                    return None
                
                items['scc_can']=self.scc
                
                self._files[nbasepath] = {
                    'filename':'',
                    'status':'Up-to-date',
                    'rev':'1',
                    'revdate':'',
                    'rrev':'1',
                    'rpath':'',
                    'path':basepath,
                    'stickytag':'',
                    'stickydate':'',
                    'stickyoptions':'',
                    'ondisk':1
                    }
                
                #print "checking CVS for path ",basepath
                # execute cvs on the path
                # XXX figure out how to parse showTags later
                result = self.cvs.status(cwd=basepath,recursive=self.recursive)
                if 'stderr' in result:
                    self._raiseCVSError(result['stderr'])
                        
                #print repr(result)
                #print "got %d files" % (len(result['files']))
                for file in result['files']:
                    filekey = os.path.normcase(file['path'])
                    path = os.path.dirname(filekey)
                    self._paths[path]=time.time()
                    #print "have ",filekey
                    file['ondisk'] = os.path.exists(filekey);
                    self._files[filekey] = file
                
                self._paths[nbasepath]=time.time()

            if dofile and not self._files.has_key(npath) and \
               not os.path.isdir(path):
                #print "checking cvs for file ",path
                # file not in cache, get it now and stuff it into the cache
                filename = os.path.basename(path)
                result = self.cvs.status(files=filename,cwd=basepath)
                if 'stderr' in result:
                    self._raiseCVSError(result['stderr'])
                    
                #print repr(result)
                for file in result['files']:
                    filekey = os.path.normcase(file['path'])
                    #print "have ",filekey
                    file['ondisk'] = os.path.exists(filekey);
                    self._files[filekey] = file
                
        except cvslib.CVSLibError, e:
            #we get an exception here if the cvs library call errors out
            logDebug(e)
            self.setError('CVS Status Error, see SCC Output pane for details',e)
            url.status.update(items)
            return None
        
        if not self._files.has_key(npath):
            # still don't have it, not a cvs file
            #print "cvs dont have ",npath
            url.status.update(items)
            return None
        
        # get cvs info for the path, not the file
        #print "setting info for ",npath
        file = self._files[npath]
        
        items['scc'] =  'cvs'
        items['haveOnDisk']= file['ondisk']
        items['myRevision'] = file['rev']
        items['revdate'] = file['revdate']
        items['depotRevision'] = file['rrev']
        items['needSync'] = items['myRevision'] != items['depotRevision']
        items['sync'] = '0'
        items['ok'] = '0'
        items['conflict'] = '0'
        items['action'] = ''
        items['status'] = file['status']

        if items['needSync']:
            items['sync'] = '1'
        else:
            items['ok'] = '1'

        # this is independent of sync status
        if items['status'].find('conflict') >= 0:
            items['conflict'] = '1'
            items['ok'] = '0'
            items['sync'] = '0'

        if (items['myRevision'] and \
           (items['myRevision'] == '0' or items['myRevision'][0] == '-')) or \
           items['status'].find('Removed') >= 0:
            items['action'] = 'delete'
        elif (items['revdate'] and items['revdate'] == 'New File!') or \
            items['status'].find('Added') >= 0:
            items['action'] = 'add'
        elif items['ok'] and items['status'] != 'Up-to-date':
            items['action'] = 'edit'

        url.status.update(items)
        return 1
