#!python
# Copyright (c) 2000-2001 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

# Implementation of the outliner view for displaying find results in the find
# results tab in the output pane of Komodo.

from xpcom import components
from koLogger import logError, logWarning, logDebug
import sys, os, re, types

class OutlinerView:
    """A base implementation of the nsIOutlinerView. This
    should be subclassed to get any useful behaviour."""
    _com_interfaces_ = [components.interfaces.nsIOutlinerView]
        
    def __init__(self, debug=0):
        self._debug = debug
        self.selection = None   # nsIOutlinerSelection

    def isSeparator(self, index):
        return 0
    
    def get_rowCount( self ):
        # Result: int32
        if self._debug:
            logDebug("get_rowCount()")
        return 0
    def getRowProperties( self, index, properties ):
        # Result: void - None
        # In: param0: int32
        # In: param1: nsISupportsArray
        if self._debug:
            logDebug("getRowProperties(%s, %s)" % (index, properties))
        pass
    def getCellProperties(self, row, colID, properties):
        # Result: void - None
        # In: param0: int32
        # In: param1: wstring
        # In: param2: nsISupportsArray
        if self._debug:
            logDebug("getCellProperties()")
        pass
    def getColumnProperties(self, colID, colElt, properties):
        # Result: void - None
        # In: param0: wstring
        # In: param1: nsIDOMElement
        # In: param2: nsISupportsArray
        if self._debug:
            logDebug("getColumnProperties(colID=%s, colElt=%s, props=%s)"\
                     % (colID, colElt, properties))
        pass
    def isContainer(self, index):
        # Result: boolean
        # In: param0: int32
        if self._debug:
            logDebug("isContainer()")
        return 0
    def isContainerOpen(self, index):
        # Result: boolean
        # In: param0: int32
        if self._debug:
            logDebug("isContainerOpen()")
        return 0
    def isContainerEmpty( self, param0 ):
        # Result: boolean
        # In: param0: int32
        if self._debug:
            logDebug("isContainerEmpty()")
        pass
    def isSorted( self ):
        # Result: boolean
        if self._debug:
            logDebug("isSorted()")
        return 0
    def canDropOn( self, param0 ):
        # Result: boolean
        # In: param0: int32
        if self._debug:
            logDebug("canDropOn()")
        return 0
    def canDropBeforeAfter( self, param0, param1 ):
        # Result: boolean
        # In: param0: int32
        # In: param1: boolean
        if self._debug:
            logDebug("canDropBeforeAfter()")
        return 0
    def drop( self, param0, param1 ):
        # Result: void - None
        # In: param0: int32
        # In: param1: int32
        if self._debug:
            logDebug("drop()")
        pass
    def getParentIndex( self, param0 ):
        # Result: int32
        # In: param0: int32
        if self._debug:
            logDebug("getParentIndex()")
        #XXX huh?
        pass
    def hasNextSibling( self, param0, param1 ):
        # Result: boolean
        # In: param0: int32
        # In: param1: int32
        if self._debug:
            logDebug("hasNextSibling()")
        return 0
    def getLevel(self, index):
        # Result: int32
        # In: param0: int32
        if self._debug:
            logDebug("getLevel()")
        return 0
    def getCellText(self, row, colID):
        # Result: wstring
        # In: param0: int32
        # In: param1: wstring
        if self._debug:
            logDebug("getCellText()")
            return "row: %s, col: %s" % (row, colID)
        else:
            return ""
    def setOutliner(self, outliner):
        # Result: void - None
        # In: param0: nsIOutlinerBoxObject
        #XXX may move to be an attribute.
        if self._debug:
            logDebug("setOutliner()")
        self._outliner = outliner
    def toggleOpenState(self, index):
        # Result: void - None
        # In: param0: int32
        if self._debug:
            logDebug("toggleOpenState()")
        pass
    def cycleHeader(self, colID, elt):
        # Result: void - None
        # In: param0: wstring
        # In: param1: nsIDOMElement
        if self._debug:
            logDebug("cycleHeader()")
        pass
    def selectionChanged(self):
        # Result: void - None
        if self._debug:
            logDebug("selectionChanged()")
        pass
    def cycleCell(self, row, colID):
        # Result: void - None
        # In: param0: int32
        # In: param1: wstring
        if self._debug:
            logDebug("cycleCell()")
        pass
    def isEditable(self, row, colID):
        # Result: boolean
        # In: param0: int32
        # In: param1: wstring
        if self._debug:
            logDebug("isEditable()")
        return 0
    def setCellText( self, param0, param1, param2 ):
        # Result: void - None
        # In: param0: int32
        # In: param1: wstring
        # In: param2: wstring
        if self._debug:
            logDebug("setCellText()")
        pass
    def performAction(self, action):
        # Result: void - None
        # In: param0: wstring
        if self._debug:
            logDebug("performAction(%s)" % action)
        pass
    def performActionOnRow(self, action, row):
        # Result: void - None
        # In: param0: wstring
        # In: param1: int32
        if self._debug:
            logDebug("performActionOnRow(%s, %s)" % (action, row))
        pass
    def performActionOnCell(self, action, row, colID):
        # Result: void - None
        # In: param0: wstring
        # In: param1: int32
        # In: param2: wstring
        if self._debug:
            logDebug("performActionOnCell(%s, %s, %s)" % (action, row, colID))
        pass


class KoFindResultsView(OutlinerView):
    _com_interfaces_ = [components.interfaces.koIFindResultsView,
                        components.interfaces.nsIOutlinerView]
    _reg_clsid_ = "{154CF0D6-D556-4be5-9F55-65AE5DB46312}"
    _reg_contractid_ = "@activestate.com/koFindResultsView;1"
    _reg_desc_ = "Komodo Find Results Outliner Table View"

    def __init__(self):
        OutlinerView.__init__(self, debug=0)
        self._data = []
        self._sortedBy = None

    def get_rowCount(self):
        return len(self._data)

    def getCellText(self, row, col):
        try:
            datum = self._data[row][col]
        except IndexError:
            logError("no %sth find result" % row)
            return ""
        except KeyError:
            logError("unknown find result column id: '%s'" % col)
            return ""
        if type(datum) not in (types.StringType, types.UnicodeType):
            datum = str(datum)
        return datum

    def Clear(self):
        length = len(self._data)
        self._data = []
        self._sortedBy = None
        self._outliner.rowCountChanged(0, length)
        self._outliner.invalidate()

    def AddFindResult(self, url, startIndex, endIndex, value, fileName,
                      lineNum, columnNum, context):
        datum = {"url": url,
                 "startIndex": startIndex,
                 "endIndex": endIndex,
                 "value": value,
                 "lineNum": lineNum,
                 "columnNum": columnNum,
                 "findresults-filename": fileName,
                 "findresults-linenum": lineNum,
                 "findresults-context": context}
        self._data.append(datum)
        self._sortedBy = None
        self._outliner.rowCountChanged(len(self._data)-2, 1)
        self._outliner.invalidate()  #XXX invalidating too much here?

    def AddReplaceResult(self, url, startIndex, endIndex, value, replacement,
                         fileName, lineNum, columnNum, context):
        datum = {"url": url,
                 "startIndex": startIndex,
                 "endIndex": endIndex,
                 "value": value,
                 "replacement": replacement,
                 "lineNum": lineNum,
                 "columnNum": columnNum,
                 "findresults-filename": fileName,
                 "findresults-linenum": lineNum,
                 "findresults-context": context}
        self._data.append(datum)
        self._sortedBy = None
        self._outliner.rowCountChanged(len(self._data)-2, 1)
        self._outliner.invalidate()  #XXX invalidating too much here?

    def GetUrl(self, index):
        return self._data[index]["url"]
    def GetStartIndex(self, index):
        return self._data[index]["startIndex"]
    def GetEndIndex(self, index):
        return self._data[index]["endIndex"]
    def GetValue(self, index):
        return self._data[index]["value"]
    def GetReplacement(self, index):
        try:
            return self._data[index]["replacement"]
        except KeyError:
            return ""
    def GetLineNum(self, index):
        return self._data[index]["lineNum"]
    def GetColumnNum(self, index):
        return self._data[index]["columnNum"]

    def Sort(self, sortBy):
        """Sort the current data by the given key. If already sorted by this
        key then reverse the sorting order."""
        if self._sortedBy == sortBy:
            self._data.reverse()
        else:
            try:
                if sortBy == 'findresults-context':  # strip leading whitespace for context sort order
                    self._data.sort(lambda dict1,dict2,sortBy=sortBy:
                                        cmp(dict1[sortBy].lstrip(), dict2[sortBy].lstrip())
                                   )
                else:
                    self._data.sort(lambda dict1,dict2,sortBy=sortBy:
                                        cmp(dict1[sortBy], dict2[sortBy])
                                   )
            except KeyError:
                logError("Cannot sort find results by: '%s'" % sortBy)
                raise
        self._sortedBy = sortBy
        self._outliner.invalidate()

    def GetNumUrls(self):
        urlDict = {}
        for datum in self._data:
            urlDict[datum["url"]] = 1
        return len(urlDict.keys())

