#!/usr/bin/env python
# Copyright (c) 2000-2001 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

import sys, os, time, threading
from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject

from koLogger import logError, logWarning, logDebug
import process, koprocessutils
import convertURL


# XXX TODOs for the GUI builder
# implement "test"
# Provide a prefs dialog to let people specify where it is if they have multiples
# provide a prefs & prefs UI to specify what default language to use
# Lookup the pref-set GUIbuilder.exe location in this file
# 

class Listener:
    def __init__(self, mgr, fileurl, pipe):
        self.mgr = mgr
        self.fileurl = fileurl
        self.pipe = pipe
        proxyManager = components.classes["@mozilla.org/xpcomproxy;1"].\
                            getService(components.interfaces.nsIProxyObjectManager)
        observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self.observerSvc = getProxyForObject(None, components.interfaces.nsIObserverService,
                                             observerSvc, PROXY_ALWAYS | PROXY_ASYNC)
        
    def __call__(self):
        laststuff = ''
        if self.observerSvc is None:
            self._getObserver()
        while 1:
            line = self.pipe.readline()
            line = line.strip()
            if not line: break
            try:
                command, line = line.split(' ', 1)
            except ValueError:
                #print "Unknown GUI command", line
                break
            if command == 'exit':
                break
            elif command == 'ping':
                # XXX: This should not be necessary, but on Win2K only the
                # guibuilder can hang when not getting data down the stdin
                # pipe - so this exists to ping/pong
                self.mgr.notifyPong(self.fileurl)
                continue
            elif command == 'help':
                if sys.platform.startswith('win'):
                    line = line.replace('/', '\\')
                # print "help: line = " + line
                url = convertURL.LocalFileToURL(line);
                data = '%s\t%s' % (self.fileurl, url)
                self.observerSvc.Notify(None, 'guiBuilder-help', data)
                continue
            try:
                project, line = line.split(' ', 1)
            except:
                # print "Unknown GUI command 2", line 
                continue # unknown command
            if command == 'save':
                if sys.platform.startswith('win'):
                    line = line.replace('/', '\\')
                #print "save: line = " + line
                url = convertURL.LocalFileToURL(line);
                data = '%s\t%s' % (self.fileurl, url)
                #print line +'\n' + url +'\n'+ data
                self.observerSvc.Notify(None, 'guiBuilder-save', data)
            elif command == 'startup':
                if sys.platform.startswith('win'):
                    line = line.replace('/', '\\')
                #print "startup: line = " + line
                url = convertURL.LocalFileToURL(line);
                data = '%s\t%s' % (self.fileurl, url)
                #print line +'\n' + url +'\n'+ data
                self.observerSvc.Notify(None, 'guiBuilder-startup', data)
                continue
            elif command == 'edit':
                if sys.platform.startswith('win'):
                    line = line.replace('/', '\\')
                #print "edit: line = " + line
                url = convertURL.LocalFileToURL(line);
                data = '%s\t%s' % (self.fileurl, url)
                self.observerSvc.Notify(None, 'guiBuilder-edit', data)
                continue
            elif command == 'dirty':
                dirtyflag = line.strip()
                if dirtyflag == '1':
                    isdirty = 1
                else:
                    isdirty = 0
                self.mgr.notifyDirty(self.fileurl, isdirty)
                continue
            else:
                #print "Unknown GUI command", line
                pass
        # if we get here it's because the pipe was closed.  Need to tell the original thread...
        self.mgr.forget(self.fileurl)

    def close(self):
        logDebug("Gui builder closed pipe");

lastErrorSvc = None

class KoGuiBuilderService:
    _com_interfaces_ = [components.interfaces.koIGuiBuilderService]
    _reg_clsid_ = "8EE42AA0-AD35-406d-8F57-9EB71622356A"
    _reg_contractid_ = "@activestate.com/koGuiBuilderService;1"
    _reg_desc_ = "Service for interacting with the GUI Builder"

    def __init__(self):
        self._processMap = {}
        self._dirtyURLs = {}
        self._forgettingLock = threading.Lock()
        self._languages = None
        global lastErrorSvc
        lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                       .getService(components.interfaces.koILastErrorService)

    def getLanguages(self):
        if self._languages is not None:
            return self._languages
        guibuilder = self._getGuibuilderPath()
        argv = [guibuilder, "--interactive", "--languages"]
        env = koprocessutils.getUserEnv()
        p = process.ProcessOpen(argv, env=env)
        s = p.stdout.read()
        p.close()
        s = s.strip()
        languages = s.split('\t')
        ids = []
        labels = []
        for language in languages:
            id, label = language.split(':')
            ids.append(id)
            labels.append(label)
        self._languages = ids, labels
        return ids, labels

    def getDirtyURLs(self):
        return self._dirtyURLs.keys()
    
    def notifyDirty(self, url, dirtyflag):
        # this is called from other threads, so needs guarding.
        self._forgettingLock.acquire()
        if dirtyflag:
            self._dirtyURLs[url] = dirtyflag
        else:
            if url in self._dirtyURLs:
                del self._dirtyURLs[url]
        self._forgettingLock.release()

    def notifyPong(self, url):
        # XXX: This should not be necessary, but on Win2K only the
        # guibuilder can hang when not getting data down the stdin
        # pipe - so this exists to ping/pong
        self._processMap[url].write('pong\n')

    def forget(self, fileurl):
        # this is called from other threads, so needs guarding.
        self._forgettingLock.acquire()
        del self._processMap[fileurl]
        if fileurl in self._dirtyURLs:
            del self._dirtyURLs[fileurl]
        self._forgettingLock.release()

    def _getGuibuilderPath(self):
        """Return the full path the GUI Builder executable.
        
        If the GUI Builder exe could not be found a suitable execption
        is raised.
        """
        dirSvc = components.classes["@activestate.com/koDirs;1"]\
                 .getService(components.interfaces.koIDirs)
        dname = os.path.join(dirSvc.installDir, "guibuilder")
        if sys.platform.startswith("win"):
            fname = os.path.join(dname, "guibuilder.exe")
        else:
            fname = os.path.join(dname, "guibuilder")
        if os.path.exists(fname):
            return fname
        else:
            ex = "Could not find the GUI Builder executable."
            lastErrorSvc.setLastError(0, ex)
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, ex)

    def _connect(self, fileurl, language=None):
        guibuilder = self._getGuibuilderPath()
        filepath = convertURL.LocalPathFromURL(fileurl)
        dname, fname = os.path.split(filepath)
        dname = dname.replace('\\', '/')
        argv = [guibuilder, "--interactive", "--directory", dname]
        if language is not None:
            argv.extend(['--language', language])
        env = koprocessutils.getUserEnv()
        p = process.ProcessOpen(argv, env=env)
        threading.Thread(target=Listener(self, fileurl, p.stdout)).start()
        self._processMap[fileurl] = p.stdin
        #import sys
        #sys.guistdin = p.stdin
        #sys.guistdout = p.stdout
        #sys.guistderr = p.stderr
        return p, fname

    def Edit(self, fileurl):
        """Start a connection with a new (or existing) GUI builder viz. a particular project"""
        if fileurl in self._processMap:
            i = self._processMap[fileurl]
            try:
                i.write('wm deiconify .\n')
                return 1 # we're done.
            except OSError:
                # if this happens it means that the gui builder quit
                # without telling us about it.  We'll clean up our
                # data structures ourselves and
                # restart it.
                self.forget(fileurl) # fall through to the non-existent connection case.
        p, fname = self._connect(fileurl)
        try:
            p.stdin.write('load '+fname+'\n')
        except OSError:
            self.forget(fileurl)
            ex = "Couldn't tell the GUI builder to load " + fname
            lastErrorSvc.setLastError(0, ex)
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, ex)
        return 1

    def Create(self, fileurl, language):
        """Start a connection with a new (or existing) GUI builder viz. a particular project"""
        p, fname = self._connect(fileurl, language)
        try:
            p.stdin.write('new '+fname+'\n')
        except OSError:
            self.forget(fileurl)
            ex = "Couldn't tell the GUI builder to create " + fname
            lastErrorSvc.setLastError(0, ex)
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, ex)
        return 1

    # NOT TESTED XXX
    def Save(self, filepath):
        assert filepath in self._processMap
        try:
            self._processMap[filepath].write('save\n');
        except OSError:
            self.forget(fileurl)
            ex = "Couldn't tell the GUI builder to save"
            lastErrorSvc.setLastError(0, ex)
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, ex)
        return 1
        
    # NOT TESTED XXX
    def Close(self, filepath):
        assert filepath in self._processMap
        try:
            self._processMap[filepath].write('close_project\n');
        except OSError:
            self.forget(fileurl)
            ex = "Couldn't tell the GUI builder to close"
            lastErrorSvc.setLastError(0, ex)
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, ex)
        return 1

    def Quit(self, filepath):
        assert filepath in self._processMap
        try:
            self._processMap[filepath].write('dirty 0\n');
            self._processMap[filepath].write('exit\n');
        except OSError:
            pass
        return 1

    def QuitAll(self):
        filepaths = self._processMap.keys()
        for filepath in filepaths:
            if not self.Quit(filepath):
                return 0
        return 1
