#!python
# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.


import os, sys
import threading
import time
from xpcom import components, nsError, ServerException, COMException
from xpcom.server.enumerator import *
from koLogger import logWarning, logError
from koEncodingLinter import KoEncodingLinter


class KoLintManager:
    _com_interfaces_ = [components.interfaces.koILintManager]
    _reg_desc_ = "Komodo Lint Manager"
    _reg_clsid_ = "{D1153223-ACAD-43a9-B6DF-4E0E1F3AAE07}"
    _reg_contractid_ = "@activestate.com/koLintManager;1"


    def __init__(self):
        #XXX it would be better to use a generalized priority queue, but
        #XXX a pair of queues, one low priority the other high, will work for now

        self.PRIORITY_HIGH = components.interfaces.koILintManager.PRIORITY_HIGH
        self.PRIORITY_LOW = components.interfaces.koILintManager.PRIORITY_LOW

        self.linters = {}
    
        self.observers = {}
        self.observers[""] = []

        self.activeRequest = None
        self.abortActiveRequest = 0

        self.queue = [[],[]]

        self.die = 0
        self.cv = threading.Condition()
        self.thread = threading.Thread(target=self.run)
        self.thread.start()

    def terminate(self):
        self.cv.acquire()
        self.die = 1
        self.cv.notify()
        self.cv.release()
        self.thread.join()     # wait for child thread to terminate

    def broadcastResultsCleared(self, docname, linterType):
        # no need for the lock here.
        observers = self.observers.get(docname, []) + self.observers[""]
        # notify each one
        for observer in observers:
            observer.notifyClear(docname, linterType)

    def addRequest(self, buffer, encoding, priority, docname, linterType, cwd):
        # NOTE: buffer must be a proxy
        if not linterType:
            self.linters[linterType] = KoEncodingLinter()
        elif not self.linters.has_key(linterType):
            try:
                comClass = components.classes[linterType]
                linter = comClass.createInstance(components.interfaces.koILinter)
            except: # catch all (guard against buggy linters)
                logError("addRequest() failed to create instance of type " + linterType)
            else:
                self.linters[linterType] = KoEncodingLinter(linter)

        if self.linters.has_key(linterType):
            self.cv.acquire()
            self.queue[priority].append({         'buffer':buffer,
                                                  'encoding':encoding,
                                                  'docname':docname,
                                                  'linterType':linterType,
                                                  'cwd':cwd})
            self.cv.notify()
            self.cv.release()

    def cancelRequests(self, docname):
        self.cv.acquire()
        if self.activeRequest and (self.activeRequest['docname'] == docname):
            self.abortActiveRequest = 1         
        for priority in [self.PRIORITY_HIGH, self.PRIORITY_LOW]:
            for item in self.queue[priority]:
                if item['docname'] == docname:
                    del self.queue[priority][self.queue[priority].index(item)]
        self.cv.release()

    def addObserver(self, observer, docname):
        # Note: <observer> must be a proxy
        self.cv.acquire()
        if not docname: # add observer to all documents
            self.observers[""].append(observer)
        elif self.observers.has_key(docname):
            if observer not in self.observers[docname]:
                self.observers[docname].append(observer)
        else:
            self.observers[docname] = [observer]
        self.cv.release()

    def removeObserver(self, observer):
        self.cv.acquire()
        for doc in self.observers.keys():
            if observer in self.observers[doc]:
                del self.observers[doc][self.observers[doc].index(observer)]
                if (not self.observers[doc]) and (doc <> ""):
                    del self.observers[doc]
        self.cv.release()

    def _notifyObservers(self, activeRequest, results, errorString=""):
        """notify observers if the request has not been cancelled by now"""
        self.cv.acquire()
        if not self.abortActiveRequest:
            # get the list of observers (those who care only about this
            # 'docname' + those that care about all docnames)
            observers = self.observers.get(activeRequest['docname'], []) +\
                self.observers[""]
            # notify each one
            for observer in observers:
                # release the lock to avoid deadlock with observer
                self.cv.release()
                observer.notify(activeRequest['docname'],
                                activeRequest['linterType'],
                                activeRequest['cwd'],
                                results,
                                errorString)
                self.cv.acquire()
        self.cv.release()

    def run(self):                # lint manager control thread
        while not self.die:
            # get next request, if available
            self.cv.acquire()
            if (not self.die) and (not self.queue[self.PRIORITY_HIGH]) \
                 and (not self.queue[self.PRIORITY_LOW]):
                self.cv.wait()
            if self.die:
                self.cv.release()
                return
            if self.queue[self.PRIORITY_HIGH]:
                self.activeRequest = self.queue[self.PRIORITY_HIGH][0]
                del self.queue[self.PRIORITY_HIGH][0]
            elif self.queue[self.PRIORITY_LOW]:
                self.activeRequest = self.queue[self.PRIORITY_LOW][0]
                del self.queue[self.PRIORITY_LOW][0]
            self.cv.release()

            if (self.activeRequest):
                try:
                    #XXX should add args, but linters don't support this yet
                    #XXX Copy of entire buffer. Don't do it.
                     results = self.linters[self.activeRequest['linterType']]\
                                  .lint(self.activeRequest['buffer'].text, self.activeRequest['encoding'], self.activeRequest['cwd'])
                except COMException, data:
                    # When XPCOM gets rich exceptions then data.message
                    # will contain the passed message. Until then must
                    # construct stock one from the nsError value.
                    msg = "Background Syntax Checking error: "
                    if data.errno == nsError.NS_ERROR_NOT_AVAILABLE:
                        # used to indicate that a linter engine (i.e.
                        # wPerl.exe for Perl linting) could not be found
                        if data.message:
                            msg += data.message
                        else:
                            # special-case the snot out of this
                            if self.activeRequest['linterType'] ==\
                              "@activestate.com/koPerlCompileLinter;1":
                                msg += "could not find perl.exe for syntax "\
                                       "checking." 
                            elif self.activeRequest['linterType'] ==\
                              "@activestate.com/koPHPLinter;1":
                                msg += "could not find php.exe for syntax "\
                                       "checking. You must have PHP (>= "\
                                       "version 4.0.5) on your Path."
                            else:
                                msg += "could not find a suitable syntax "\
                                       "checking engine."
                    elif data.errno == nsError.NS_ERROR_ILLEGAL_VALUE:
                        if self.activeRequest['linterType'] ==\
                           "@activestate.com/koPHPLinter;1":
                            msg += "Need PHP version 4.0.5 or better."
                        elif self.activeRequest['linterType'] ==\
                             "@activestate.com/koTclLinter;1":
                            msg += "could not find a ActiveState TclPro license (see Tcl preferences)."
                        else:
                            msg += "versioning problem."
                    elif data.errno == nsError.NS_ERROR_FILE_INVALID_PATH:
                        if self.activeRequest['linterType'] ==\
                             "@activestate.com/koPerlCompileLinter;1":
                            msg += "can't do Perl syntax checking of files with UNC paths on Windows 95/98/ME."
                        elif self.activeRequest['linterType'] ==\
                             "@activestate.com/koXPCShellLinter;1":
                            msg += "can't do JavaScript syntax checking of files with UNC paths on Windows 95/98/ME."
                        elif self.activeRequest['linterType'] ==\
                             "@activestate.com/koTclLinter;1":
                            msg += "can't do Tcl syntax checking of files with UNC paths on Windows 95/98/ME."
                        elif self.activeRequest['linterType'] ==\
                           "@activestate.com/koHTMLLinter;1":
                            msg += "can't do HTML syntax checking of files with UNC paths on Windows 95/98/ME."
                        elif self.activeRequest['linterType'] ==\
                           "@activestate.com/koPHPLinter;1":
                            msg += "can't do PHP syntax checking of files with UNC paths on Windows 95/98/ME."
                        else:
                            msg += "invalid current working directory."
                    else:
                        msg += "unexpected error %x." % data.errno
                    self._notifyObservers(self.activeRequest, None, msg)
                    # Disable this for now becase it is common for PHP:
                    # logError(msg)
                except: # catch all (guard against buggy linters)
                    info = sys.exc_info()
                    msg = "Unexpected linting error in %s: %s %s" %\
                      (self.activeRequest['linterType'], info[0], info[1].args)
                    self._notifyObservers(self.activeRequest, None, msg)
                    logError(msg)
                else:
                    self._notifyObservers(self.activeRequest, results)

                self.cv.acquire()
                self.abortActiveRequest = 0
                self.activeRequest = None
                self.cv.release()

