#!python
# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

import os
import sys

import p4lib
import convertURL

from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject


#---- support routines

def _unique(s):
    """Return a list of the elements in s, in arbitrary order, but without
    duplicates. (_Part_ of the Python Cookbook recipe.)
    """
    # get the special case of an empty s out of the way, very rapidly
    n = len(s)
    if n == 0:
        return []

    # Try using a dict first, as that's the fastest and will usually work.
    u = {}
    try:
        for x in s:
            u[x] = 1
    except TypeError:
        del u  # move on to the next method
    else:
        return u.keys()

def _splitall(path):
    """Split the given path into all its directory parts and return the list
    of those parts (see Python Cookbook recipe for test suite.
    """
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts

def _commonprefix(paths):
    """An os.path.commonprefix() more suited to actual paths.
    
    It returns the common path prefix of the given paths or None, if there
    is no such common prefix.
    """
    if not paths:
        return None
    splitpaths = [_splitall(path) for path in paths]
    commonprefix = []
    for set in zip(*splitpaths):
        # This path element is part of the common prefix if it is the same
        # for every give path.
        elem = set[0]
        if sys.platform.startswith("win"):
            # Case-insensitive comparison on Windows.
            set = [p.lower() for p in set]
        if len(_unique(set)) == 1:
            commonprefix.append(elem)
        else:
            break
    if commonprefix:
        retval = os.path.join(*commonprefix)
    else:
        retval = None
    #print "_commonprefix(%r) == %r" % (paths, retval)
    return retval



#---- component implementation

class KoP4:
    _com_interfaces_ = [components.interfaces.koIP4]
    _reg_desc_ = "Komodo P4 Support"
    _reg_contractid_ = "@activestate.com/koP4;1"
    _reg_clsid_ = "{7BD69117-A206-410b-A2C1-FAC34976734A}"
    
    def __init__(self):
        # generic cvs instance to do work with
        self.p4 = p4lib.P4()
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                       .getService(components.interfaces.koILastErrorService)
        self.tmpFileSvc = components.classes["@activestate.com/koTempFileFactory;1"].\
                getService(components.interfaces.koITempFileFactory);
        self.fileStatusSvc = components.classes["@activestate.com/koFileStatusService;1"].\
                getService(components.interfaces.koIFileStatusService);

        self.reasonNotFunctional = None
        self.isFunctional = None
        self.redetermineIfFunctional()
        
    def redetermineIfFunctional(self):
        import which
        p4s = list(which.which('p4'))
        if not p4s:
            self.isFunctional = 0
            self.reasonNotFunctional = "no 'p4' executable could be found"
        else:
            self.isFunctional = 1
            self.reasonNotFunctional = None
        try:
            self._observerSvc.Notify(self, 'subsystem_changed', 'p4')
        except COMException:
            pass # Ignore error if there is no observer.

    def _splitFile(self, file):
        # figure out what repositories files are in
        # and split them up into seperate lists
        # we also convert to local paths here
        try:
            path = convertURL.LocalPathFromURL(file);
            if path:
                if os.path.isdir(path):
                    return path,'./...'
                path = os.path.normpath(path)
                basedir = os.path.dirname(path)
                filename = os.path.basename(path)
                
                # make sure the path is in cvs
                #root = self.p4.have(path)
                #if root:
                #    return basedir, filename
                return basedir, filename
        except p4lib.P4LibError, e:
            # we'll ignore files that are not in perforce
            self.lastErrorSvc.setLastError(0, str(e))
            pass
        return None,None
    
    def _splitFiles(self, files):
        # figure out what repositories files are in
        # and split them up into seperate lists
        # we also convert to local paths here
        _files = {}
        for file in files:
            basedir,filename = self._splitFile(file)
            if basedir:
                if not basedir in _files:
                    _files[basedir] = []
                _files[basedir].append(filename)
        return _files
    
    def _convertFiles(self, files):
        _files = []
        for file in files:
            path = convertURL.LocalPathFromURL(file)
            path = os.path.normpath(path)
            _files.append(path)
        return _files
    
#    diff -- Display diff of client file with depot file
#
#    p4 diff [ -d<flag> -f -sa -sd -se -sr -t ] [ file[rev] ... ]
#
#        Run diff (on the client) of a client file against the corresponding
#        revision in the depot. The file is only compared if the file is
#        opened for edit or the revision provided with the file argument is
#        not the same as the revision had by the client.  See 'p4 help
#        revisions' for help specifying revisions.
#
#        If no file argument is given, diff all open files.
#        This can be used to view pending changelists.
#
#        The -d<flag> passes a flag to the built-in diff routine to modify
#        the output: -dn (RCS), -dc (context), -ds (summary), -du (unified).
#
#        The -f flag forces a diff for every file, regardless of whether
#        they are opened or if the client has the named revision.
#        This can be used to verify the client contents.
#
#        The -s flag reduces the output of diff to the names of files
#        satisfying the following criteria:
#
#                -sa     Opened files that are different than the revision
#                        in the depot, or missing.
#
#                -sd     Unopened files that are missing on the client.
#
#                -se     Unopened files that are different than the revision
#                        in the depot.
#
#                -sr     Opened files that are the same as the revision in the
#                        depot.
#
#        The -t flag forces 'p4 diff' to diff even files with non-text
#        (binary) types.
#
#        If the environment variable $P4DIFF is set then the named program is
#        used rather than the implementation of diff included in the client.
#        The -d<flag>command can be used to pass arguments to the
#        external program.  The -s flag is only implemented internally.
    def diff(self, files, options):
        #print "p4 diff ",repr(files)
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        diff = ''
        for basedir in _files.keys():
            # ,cwd=basedir
            try:
                result, raw = self.p4.diff(files=_files[basedir],
                                           diffFormat=options,
                                           _raw=1,
                                           cwd=basedir)
            except p4lib.P4LibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stdout' in raw:
                diff += raw['stdout']
            elif 'stderr' in raw:
                self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                diff += raw['stderr']
                
        return diff

#    add -- Open a new file to add it to the depot
#
#    p4 add [ -c changelist# ] [ -t filetype ] file ...
#
#        Open a new file for adding to the depot.  If the file exists
#        on the client it is read to determine if it is text or binary.
#        If it does not exist it is assumed to be text.  The file must
#        either not exist in the depot, or it must be deleted at the
#        current head revision.  Files may be deleted and re-added.
#
#        If the -c flag is given the open files are associated with the
#        specified pending changelist number; otherwise the open files are
#        associated with the default changelist.
#
#        If file is already open it is moved into the specified pending
#        changelist.  It is not permissible to reopen a file for add unless
#        it was already open for add.
#
#        If -t filetype is given the file is explicitly opened as that
#        filetype.  Otherwise, the filetype is determined by the file
#        name-to-type mapping table managed by 'p4 typemap'.  If the file
#        name is not mapped in that table, 'p4 add' senses the filetype
#        by examining the file's contents and execution permission bits.
#        See 'p4 help filetypes' for a complete list.
    def add(self, files, mode, message):
        #add(self, files, mode=None, msg=None, cwd=None)
        # first we figure out what repositories files are in
        # and split them up into seperate lists
        # we init the cvs object with the root we retreived
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            # , filetype=mode, msg=message, cwd=basedir
            try:
                result, raw = self.p4.add(files=_files[basedir],
                                          _raw=1, cwd=basedir)
            except p4lib.P4LibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in raw:
                self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
        # do notification that status may have changed
        # we have to use the original urls passed in since
        # they are modified to local paths in _splitFiles
        for file in files:
            if file: self._observerSvc.Notify(self, 'file_status_now', file)
        return output
    
#    submit -- Submit open files to the depot
#
#    p4 submit [ -s ]
#    p4 submit [ -s ] files
#    p4 submit -c changelist#
#    p4 submit -i [ -s ]
#
#        'p4 submit' commits a pending changelist and its files to the depot.
#
#        With no argument 'p4 submit' attempts to submit all files in the
#        'default' changelist.  Submit provides the user with a dialog
#        similar to 'p4 change' so the user can compose a changelist
#        description.  In this dialog the user is presented with the list
#        of files open in changelist 'default'.  Files may be deleted from
#        this list but they cannot be added.  (Use an open command (edit,
#        add, delete) to add additional files to a changelist.)
#
#        If a (single) file pattern is given, only those files in
#        the 'default' changelist that match the pattern will be submitted.
#
#        The -c flag submits the numbered pending changelist that has been
#        previously created with 'p4 change' or a failed 'p4 submit'.
#
#        The -i flag causes a changelist specification (including files to be
#        submitted) to be read from the standard input.  The user's editor
#        is not invoked.
#
#        The -s flag extends the list of jobs to include the fix status
#        for each job, which becomes the job's status when the changelist
#        is committed.  See 'p4 help change' for more notes on this option.
#
#        Before committing a changelist submit locks all associated files not
#        already locked.  If any file cannot be locked, or if the submit
#        fails for any other reason the files are left open in a newly
#        created pending changelist.
#
#        Submit is guaranteed to be atomic.  Either all files will be
#        updated in the depot as a unit or none will be.
    def commit(self, files, message, options):
        #print "cvs commit ",repr(files)
        # p4lib saves message to a message for file
        self.lastErrorSvc.setLastError(0, '')
        _split = self._splitFiles(files)
        _files = self._convertFiles(files)
        basedir = _commonprefix(_split.keys())
        output = ''
        try:
            result, raw = self.p4.submit(_files,
                                         description=message,
                                         _raw=1,
                                         cwd=basedir)
        except p4lib.P4LibError, e:
            self.lastErrorSvc.setLastError(0, str(e))
            return ''

        if 'stderr' in raw:
            self.lastErrorSvc.setLastError(0, str(raw['stderr']))
            output += raw['stderr']
        if 'stdout' in raw:
            output += raw['stdout']
        
        for file in files:
            self._observerSvc.Notify(self, 'file_status_now', file)
        return output
    
#    delete -- Open an existing file to delete it from the depot
#
#    p4 delete [ -c changelist# ] file ...
#
#        Opens a file that currently exists in the depot for deletion.
#        If the file is present on the client it is removed.  If a pending
#        changelist number is given with the -c flag the opened file is
#        associated with that changelist, otherwise it is associated with
#        the 'default' pending changelist.
#
#        Files that are deleted generally do not appear on the have list.
    def remove(self, files, force, recursive):
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            # , force, recursive,cwd=basedir
            try:
                result, raw = self.p4.delete(_files[basedir],
                                             _raw=1, cwd=basedir)
            except p4lib.P4LibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in raw:
                self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
        for file in files:
            self._observerSvc.Notify(self, 'file_status_now', file)
        return output
    
#    sync -- Synchronize the client with its view of the depot
#
#    p4 sync [ -f -n ] [ file[revRange] ... ]
#
#        Sync updates the client workspace to reflect its current view (if
#        it has changed) and the current contents of the depot (if it has
#        changed).  The client view is used to map client file names to
#        depot file names and vice versa.
#
#        Sync adds files that are in the client view but which have not been
#        retrieved before.  Sync deletes previously retrieved files which
#        are no longer in the client view or have been deleted from the
#        depot.  Sync updates files which are still in the client view and
#        which have been updated in the depot.
#
#        Normally, sync affects all files in the client workspace.  If file
#        arguments are given, sync limits its operation to those files.
#        The file arguments may contain wildcards.
#
#        If the file argument includes a revision specifier, then the given
#        revision is retrieved.  Normally, the head revision is retrieved.
#        See 'p4 help revisions' for help specifying revisions.
#
#        If the file argument includes a revision range specification, then
#        only files selected by the revision range are updated, and the
#        highest revision in the range is used.
#
#        Normally, sync will not clobber files in the client workspace that
#        the user has made writable.  Setting the 'clobber' option in the
#        client spec disables this safety check.
#
#        The -f flag forces resynchronization even if the client already
#        has the file, and clobbers writable files.  This flag doesn't affect
#        open files.
    def update(self, files, options):
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            # ,cwd=basedir
            try:
                result, raw = self.p4.sync(_files[basedir],
                                           _raw=1,
                                           cwd=basedir)
            except p4lib.P4LibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in raw:
                self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
            #XXX need to deal with this
            ## build a better status output for komodo
            #if 'files' in result:
            #    for file in result['files']:
            #        localFileName = os.path.normpath(os.path.join(basedir,file['file']))
            #        url = convertURL.LocalFileToURL(localFileName)
            #        #print "%s %s %s" %(file['status'],file['file'],url)
            #        output += "%s %s %s\n" % (file['status'],file['file'],url)
            #output = ""
                    
        for file in files:
            self._observerSvc.Notify(self, 'file_status_now', file)
        
        return output

#    revert -- Discard changes from an opened file
#
#    p4 revert [ -a -c changelist# ] file ...
#
#        Revert an open file back to the revision previously synced from
#        the depot, discarding any pending changelists or integrations that
#        have been made.  This command requires naming files explicitly.
#        After running revert the named files will no longer be locked
#        or open.
#
#        The -a flag tells 'p4 revert' to revert only those files which
#        are opened for edit or integrate and are unchanged or missing.
#        Files with pending integration records are left open.  With the
#        -a flag, the file arguments are optional.
#
#        The -c flag limits 'p4 revert' to files opened under the given,
#        pending changelist.
    def revert(self, files, options):
        # XXX unfortunately this function requires knowledge of
        # item properties
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            try:
                result, raw = self.p4.revert(_files[basedir], _raw=1,
                                             cwd=basedir)
            except p4lib.P4LibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in raw:
                self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
        for file in files:
            self._observerSvc.Notify(self, 'file_status_now', file)
        return output
    
    def edit(self, files):
        # XXX unfortunately this function requires knowledge of
        # item properties
        self.lastErrorSvc.setLastError(0, '')
        _files = self._splitFiles(files)
        output = ''
        for basedir in _files.keys():
            # ,cwd=basedir
            try:
                result, raw = self.p4.edit(_files[basedir],
                                           _raw=1, cwd=basedir)
            except p4lib.P4LibError, e:
                self.lastErrorSvc.setLastError(0, str(e))
                return ''
            if 'stderr' in raw:
                self.lastErrorSvc.setLastError(0, str(raw['stderr']))
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
        for file in files:
            self._observerSvc.Notify(self, 'file_status_now', file)
        
        return output
   
