#!python
# Copyright (c) 2000-2002 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

"""Language package for PHP"""

from xpcom import components, ServerException
import os, time
from koLanguageServiceBase import *
import autoinsert
import process
import koprocessutils

class koPHPLanguage(KoLanguageBase):
    name = "PHP"
    _reg_desc_ = "%s Language" % name
    _reg_contractid_ = "@activestate.com/koLanguage?language=%s;1" \
                       % (name)
    _reg_clsid_ = "{F472CC39-0902-4d92-8A5B-7DD0F612F119}"

    namedBlockRE = r'^[ \t]*(?:function\s+(?:[&]*?)\s*([\w_]*?)\s*(\(.*?\)))|(^[ \t]*(?<=\s)class\s+([\w_]*)\s*(?:(?:extends\s+([\w_]*)\s*)|\s*))'
    namedBlockDescription = 'PHP functions and classes'
    defaultExtension = ".php"
    styleBits = 7
    indicatorBits = 1
    cookbookName = "PHP"
    cookbookUrl = "http://aspn.activestate.com/ASPN/Cookbook/PHP"
    variableIndicators = '$'
    # XXX read from config somewhere, download from ActiveState?
    downloadURL = 'http://aspn.ActiveState.com/ASPN/PHP/'
    remoteDebuggerSupport = 1
    commentDelimiterInfo = {
        "line": [ "#", "//" ],
        "block": [ ("/*", "*/") ],
    }

    def get_debugger(self):
        if self._debugger is None:
            koDirs = components.classes["@activestate.com/koDirs;1"].\
              getService(components.interfaces.koIDirs)
            dbgpy = os.path.join(koDirs.mozBinDir, "components" ,"koPHPDebugger.py")
            if os.path.exists(dbgpy):
                self._debugger = KoDebuggerLanguageService(self.name)
            else:
                self._debugger = None
        return self._debugger

    def get_lexer(self):
        if self._lexer is None:
            self._lexer = KoLexerLanguageService()
            self._lexer.setLexer(components.interfaces.ISciMoz.SCLEX_HTML)
            # XXX config options needed?
            self._lexer.setProperty('fold.html','1')
            #self._lexer.setProperty('fold.compact','1')
            # keyword sets
            # 0 = html # 1 = JavaScript # 2 = VBScript # 3 = PythonScript # 4 = PHP
            self._lexer.setKeywords(1, self._js_keywords)
            self._lexer.setKeywords(2, self._vb_keywords)
            self._lexer.setKeywords(4, self._php_keywords)
            self._lexer.supportsFolding = 1
        return self._lexer

    def get_linter(self):
        if self._linter is None:
            self._linter = KoLinterLanguageService(
                "@activestate.com/koPHPLinter;1", None)
        return self._linter

    def get_methodcompletion(self):
        if self._methodcompletion is None:
            self._methodcompletion = components.classes["@activestate.com/koPHPCompletionLanguageService;1"].getService(components.interfaces.koICompletionLanguageService)
        return self._methodcompletion

    def get_commenter(self):
        #XXX Could override default KoCommenterLangSvc to use styles to tell
        #    if should use HTML or PHP style comment delimiters. 
        if self._commenter is None:
            self._commenter = KoCommenterLanguageService(
                self.commentDelimiterInfo)
        return self._commenter

    def get_inserter(self):
        if self._inserter is None:
            self._inserter = KoWSInsertLanguageService(
                autoinsert.PHPAutoInsert(),
                "$%s = WebService::ServiceProxy(\"%s\", \"%s\");",
                "webservice.php", 1)
        return self._inserter

    def get_interpreter(self):
        if self._interpreter is None:
            self._interpreter = KoPHPInterpreter()
        return self._interpreter

    def get_cookbook(self):
        if self._cookbook is None:
            self._cookbook = KoCookbookLanguageService(
                self.cookbookName, self.cookbookUrl)
        return self._cookbook

    # XXX later these need to be centralized.
    _php_keywords = ["for", "break", "continue", "if", "else",
                 "elseif", "endif", "do", "while", "return",
                 "switch", "case", "function", "class",
                 "new", "static", "define", "__LINE__",
                 "__FILE__", "PHP_VERSION", "PHP_OS", "TRUE",
                 "FALSE", "E_ERROR", "E_WARNING", "E_PARSE",
                 "E_NOTICE", "int", "integer", "real", "double",
                 "float", "string", "array", "object", "var",
                 "extends", "cfunction", "and", "default",
                 "empty", "endfor", "endswitch", "endwhile",
                 "E_ALL", "foreach", "include", "include_once",
                 "require", "require_once", "global", "list",
                 "or", "this", "xor", "virtual", "echo"]

    _js_keywords = ["abstract", "boolean", "break", "byte", "case", "catch",
                 "char", "class", "const", "continue", "debugger", "default",
                 "delete", "do", "double", "else", "enum", "export",
                 "extends", "final", "finally", "float", "for", "function",
                 "goto", "if", "implements", "import", "in", "instanceof",
                 "int", "interface", "long", "native", "new", "package",
                 "private", "protected", "public", "return", "short",
                 "static", "super", "switch", "synchronized", "this", "throw",
                 "throws", "transient", "try", "typeof", "var", "void",
                 "volatile", "while", "with"]

    _vb_keywords = ["and", "as", "attribute", "begin", 
                    "byref", "byval", "case", "call", "const",
                    "continue", "dim", "do", "each", "else",
                    "elseif", "end", "error", "exit", "false",
                    "for", "function", "global", "goto", "if",
                    "in", "loop", "me", "new", "next", "not",
                    "nothing", "on", "optional", "or", "private",
                    "public", "redim", "rem", "resume", "select",
                    "set", "sub", "then", "to", "true", "type",
                    "while", "with", "boolean", "byte", "currency",
                    "date", "double", "integer", "long", "object",
                    "option", "single", "string", "type", "variant"]



# get php configuration information needed for lexing
# this class is almost a duplicate of the php lint class
# TODO:
#      needs to be put somewhere central to be used wherever we need this info
#      have base class, or combine somehow with php linter

PROXY_SYNC    = 0x0001  # acts just like a function call.
PROXY_ASYNC   = 0x0002  # fire and forget.  This will return immediately and you will lose all return information.
PROXY_ALWAYS  = 0x0004  # ignore check to see if the eventQ is on the same thread as the caller, and alway return a proxied object.


class KoPHPInterpreter(KoInterpreterLanguageServiceBase):
    # a com interface to allow us to be an observer
    _com_interfaces_ = [components.interfaces.koIInterpreterLanguageService, components.interfaces.nsIObserver]
    _reg_contractid_ = "@activestate.com/koPHPConfigInfo;1"
    _reg_clsid_ = "{C7C6FE5D-A84D-4ebf-B1C8-7DE9DAA53CE1}"
    _reg_desc_ = "Information about the default PHP interpreter"
    
    configInfo = None
    _haveObserver = 0
    configScript = "<?php\n"+ \
                        'echo("SAPI=".php_sapi_name()."\n");\n'+ \
                        'echo("BUILD_OS=".php_uname()."\n");\n'+ \
                        'echo("INIFILE=".get_cfg_var("cfg_file_path")."\n");\n'+ \
                        'echo("EXTENSION_DIR=".ini_get("extension_dir")."\n");\n'+ \
                        'echo("INCLUDE_PATH=".ini_get("include_path")."\n");\n'+ \
                        'echo("DGB_LOADED=".(extension_loaded("Active Debugger")?"Yes":"No")."\n");\n'+ \
                        'echo("DGB_ENABLED=".ini_get("activedebug.enabled")."\n");\n'+ \
                        'echo("DGB_HOST=".ini_get("activedebug.host")."\n");\n'+ \
                        'echo("DGB_PORT=".ini_get("activedebug.port")."\n");\n'+ \
                        'echo("DGB_REMOTE_PORT=".ini_get("activedebug.remoteport")."\n");\n'+ \
                        'echo("PHPVERSION=".phpversion()."\n");\n'+ \
                        '?>'

    def __init__(self):
        proxyManager = components.classes["@mozilla.org/xpcomproxy;1"].\
                            getService(components.interfaces.nsIProxyObjectManager)

        sysUtils = components.classes["@activestate.com/koSysUtils;1"].\
            getService(components.interfaces.koISysUtils)
        self.sysUtils = proxyManager.getProxyForObject(None,
                                                       components.interfaces.koISysUtils,
                                                       sysUtils,
                                                       PROXY_SYNC | PROXY_ALWAYS)
        prefService = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService)
        self.prefService = proxyManager.getProxyForObject(None,
                                                          components.interfaces.koIPrefService,
                                                          prefService,
                                                          PROXY_SYNC | PROXY_ALWAYS)
        self._globalPrefs = self.prefService.prefs
        self._lastModified = self._lastAccess = 0

    def init(self):
        if not self._haveObserver:
            self._globalPrefs.AddObserver(self)
            self._haveObserver = 1
        # check if php.ini has changed
        # XXX we don't want to always be doing this check, but
        # how often???
        if self._lastAccess < time.time() - 5:
            ini = self._getInterpreterConfig()
            if ini:
                ini = "%s/php.ini" % ini
                if os.path.exists(ini):
                    mt = os.path.getmtime(ini)
                    if mt != self._lastModified:
                        self.configInfo = None
                        self._lastModified = mt
            else:
                self.configInfo = None
        if not self.configInfo:
            self._lastAccess = time.time()
            self._getConfig()
        
    def Observe(self, prefSet, topic, prefName):
        if topic == self._globalPrefs.id and \
           (prefName == "phpDefaultInterpreter" or prefName == "phpConfigFile"):
            # empty the configInfo, next request for info will re-get it
            # executing getConfig directly here would result in php being executed twice
            # every time we edit global prefs
            self.configInfo = None

    def get_version(self):
        self.init()
        if self.configInfo and self.configInfo.has_key("PHPVERSION"):
            return self.configInfo["PHPVERSION"]
        return ""

    def get_interpreterPath(self):
        self.init()
        return self._getInterpreter()

    def get_includePath(self):
        self.init()
        if self.configInfo and self.configInfo.has_key("INCLUDE_PATH"):
            return self.configInfo["INCLUDE_PATH"]
        return ""
    
    def getConfiguration(self):
        self.init()
        return self.configInfo

    def _getInterpreter(self):
        if self.prefService.prefs.hasStringPref("phpDefaultInterpreter") and\
           self.prefService.prefs.getStringPref("phpDefaultInterpreter"):
            php = self.prefService.prefs.getStringPref("phpDefaultInterpreter")

            #XXX Is this still necessary now that we are using process.py?
            #    It existed because popen2 cannot handle unicode paths.
            try:
                php = str(php)
            except UnicodeError:
                raise ServerException, (nsError.NS_ERROR_NOT_AVAILABLE,
                        "Path to PHP interpreter cannot contain Unicode characters.") 
        else:
            php = self.sysUtils.Which("php")
        return php or None

    def _getInterpreterConfig(self):
        if self.prefService.prefs.hasStringPref("phpConfigFile") and\
           self.prefService.prefs.getStringPref("phpConfigFile"):
            iniFile = self.prefService.prefs.getStringPref("phpConfigFile")
        else:        
            iniFile = None
        return iniFile

    def _getConfig(self):
        php = self._getInterpreter()
        if php is None:
            # Exceptions are currently ignored by the LintManager but this
            # could be used later to give the user a useful error message.
            #raise ServerException, (nsError.NS_ERROR_NOT_AVAILABLE,
            #    "Could not find a suitable PHP interpreter.")
            return

        env = koprocessutils.getUserEnv()
        ini = self._getInterpreterConfig()
        if ini:
            env["PHPRC"] = ini
        argv = [php, '-q']
        # PHP sets stdin, stdout to binary on it's end.  If we do not do
        # the same, some wierd things can happen, most notably with shell
        # comments that end with \r\n.
        p = process.ProcessOpen(argv, mode='b', env=env)
        p.stdin.write(self.configScript)
        p.stdin.close()
        lines = p.stdout.readlines()
        # For some reason, PHP linter causes an exception on close
        # with errno = 0, will investigate in PHP later.
        try:
            p.close()
        except IOError, e:
            if e.errno == 0:  
                pass
            else:
                raise

        #print "Received: ", lines
        config = {}
        try:
            for line in lines:
                #print line
                line = line.strip()
                key, value = line.split("=")
                config[key]=value
        except ValueError, e:
            logError("Unable to retreive PHP configuration information.")
        self.configInfo = config
        #print repr(self.configInfo)
