"""
Interface to the Komodo Project Files and their constituent parts.

-- a koIPart is any chunk of data serializable as part of a project file
-- a koIProject is the toplevel koIPart.
-- koIParts can have a prefset attribute which is serialized as a child but available
   as an XPCOM property
"""

from __future__ import generators
import traceback, os, sys, urlparse, string, fnmatch, weakref
from koLogger import logWarning, logError, logDebug
from xpcom import components, ServerException, COMException, nsError
import convertURL
from xml.sax import SAXParseException
from xml.dom import minidom
from koXMLPrefs import NodeToPrefset
from xpcom.server import WrapObject, UnwrapObject
from eol import newl
import shutil, tempfile

KPF_VERSION_STRING = "2"
koLanguageRegistryService = None
gLastProjNum = 1


#---- support routines

def _makeNewProjectName():
    global gLastProjNum
    newname = "Unnamed Project %d" % gLastProjNum
    gLastProjNum += 1
    return newname

def _write_data(writer, data):
    "Writes datachars to writer."
    data = string.replace(data, "&", "&amp;")
    data = string.replace(data, "<", "&lt;")
    data = string.replace(data, "\"", "&quot;")
    data = string.replace(data, ">", "&gt;")
    writer.write(data)
    
def _url2basename(url):
    localfile = convertURL.LocalPathFromURL(url)
    return os.path.splitext(os.path.basename(localfile))[0]


#---- PyXPCOM implementations

class koPart:
    _com_interfaces_ = [components.interfaces.koIPart]
    _reg_desc_ = "Project Part"
    _reg_contractid_ = "@activestate.com/koPart;1"
    _reg_clsid_ = "{06EB864D-C692-4768-AE45-7900EC5B6E43}"
    shortcut_command = ''
    
    def __init__(self):
        self.prefset = None
        self._attributes = {}
        # 'type' must be be something, otherwise the generated XML for
        # "plain parts" is invalid.
        self.type = 'part'
        self.value = ''
        self._project = None
        self.parent = None
        self.flavors = ['text/unicode','application/x-komodo-part',\
                        # XXX for a later release, scintilla needs work in this area
                        'TEXT',#,'COMPOUND_TEXT','STRING','UTF-8' \
                        ]
        self.primaryInterface = 'koIPart'
        
    def __str__(self):
        if self.type in ('file','folder'):
            return 'Part (%s): url=%s, name=%s\nChildren = %s'\
                   % (self.type, getattr(self, 'url', 'NOURL'),
                      getattr(self, 'name', 'NONAME'),
                      getattr(self, 'children', 'No Kids'))
        return 'Part (%s)' % self.type
    
    def get_project(self):
        return self._project
    
    def set_project(self, project):
        self._project = project

    def dump(self, indent):
        print " "*indent + "Part of type '" + self.type +"':"
        for k, v in self._attributes.items():
            print ' '*(indent+1) + "%s: %s" % (k, v)
        print ' '*(indent+1) + 'CDATA: %s' % self.value

    def serialize(self, writer):
        self._serializeHeader(writer)
        self._serializeTail(writer)

    def _serializeHeader(self, writer):
        writer.write("<" + self.type)
        a_names = self._attributes.keys()
        a_names.sort()
        for a_name in a_names:
            if self.type == 'project' and a_name == 'url':
                continue
            writer.write(" %s=\"" % a_name)
            data = self.getStringAttribute(a_name)
            if '\n' in data:
                errmsg = "Illegal newline character in '%s' attribute of "\
                         "%s project item" % (a_name, self.type)
                lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                    .getService(components.interfaces.koILastErrorService)
                lastErrorSvc.setLastError(0, errmsg)
                raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE, errmsg)
            # XXX need to decentralize this processing
            if a_name == 'url' or (self.type == 'dialog' and a_name == 'startup'):
                data = convertURL.RelativizeURL(self._project.url[:self._project.url.rindex('/')+1], data)
            _write_data(writer, data.encode('utf-8'))
            writer.write("\"")
        writer.write(">%s"%(newl))
        if self.prefset:
            self.prefset.serialize(writer)
        # write out the CDATA, i.e. self.value
        _write_data(writer, self.value)

    def _serializeTail(self, writer):
        writer.write("</%s>%s" % (self.type,newl))

    def hasAttribute(self, name):
        return self._attributes.has_key(name)

    def getAttribute(self, name):
        return self._attributes[name]
    
    def setAttribute(self, name, value):
        self._attributes[name] = value
        if self._project:
            self._project.isDirty = 1

    def getStringAttribute(self, name):
        return unicode(self._attributes[name])
    
    def setStringAttribute(self, name, value):
        self.setAttribute(name, unicode(value))

    def getLongAttribute(self, name):
        return int(self._attributes[name])
    
    def setLongAttribute(self, name, value):
        self.setAttribute(name, int(value))

    def getBooleanAttribute(self, name):
        return self.getLongAttribute(name) and 1 or 0

    def setBooleanAttribute(self, name, value):
        self.setAttribute(name, value and 1 or 0)

    def getDragData(self):
        #print "getDragData ",repr(self.value)
        return self.value
    
    def getDragFlavors(self):
        return self.flavors        
    
    def clone(self):
        # This clone function handles only the basic information that is
        # in every koIPart.  Custom parts that have additional information
        # which needs to be cloned will need to implement their own clone
        # function.  One small example is koIContainer
        kopart = components.classes[self._reg_contractid_] \
              .createInstance(self._com_interfaces_[0])
        part = UnwrapObject(kopart)
        part.type = self.type
        part.value = self.value
        
        if hasattr(self, '_url'):
            part._url = self._url
        if hasattr(self, '_name'):
            part._name = self._name
        
        if self.prefset:
            part.prefset = self.prefset.clone()
        else:
            part.prefset = None

        a_names = self._attributes.keys()
        a_names.sort()
        for a_name in a_names:
            data = self.getStringAttribute(a_name)
            part.setStringAttribute(a_name,data)
            
        return WrapObject(part, self._com_interfaces_[0])
    

class koContainerBase(koPart):
    def __init__(self):
        koPart.__init__(self)
        self.childrenwrapped = 1
        self._name = ''
        self._url = ''
        self.numChildren = 0
        self.children = []
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"]\
                    .getService(components.interfaces.nsIObserverService)

    def getChildren(self):
        # Not sure if this is the right time to do XPCOM wrapping.
        if not self.childrenwrapped:
            self._wrapChildren()
        return self.children
    
    def _wrapChildren(self):
        self.childrenwrapped = 1
        self.children = [WrapObject(child, components.interfaces.koIPart)\
                         for child in self.children]

    def hasChild(self, child):
        return child in self.children

    def getChildWithTypeAndStringAttribute(self, type, attrname, attrvalue):
        # This optimization assumes that all koPart implementations are
        # Python objects.  Otherwise we'd need to use getAttribute,
        # which would be much slower.
        if self.childrenwrapped:
            children = [UnwrapObject(child) for child in self.children]
        else:
            children = self.children
        for child in children:
            if child.type == type\
               and child._attributes.get(attrname,None) == attrvalue:
                return child
        return None

    def addChild(self, child):
        if not self.childrenwrapped:
            self._wrapChildren()
        if not self.hasChild(child):
            self.children.append(child)
        if self._project:
            self._project.isDirty = 1
            self._project.registerChildByURL(child)
            child.project = self._project # this calls the setter
        child.parent= self # cycle - need to investigate WeakReference
        
    def get_project(self):
        return self._project
    
    def set_project(self, project):
        self._project = project
        # Need to recursively go through the child's children to set the
        # project property.
        for child in self.children:
            child.project = project# This is recursive 
        
    def removeChild(self, child):
        if self.childrenwrapped:
            children = [UnwrapObject(c) for c in self.children]
        child = UnwrapObject(child)
        children.remove(child)
        if self._project:
            self._project.isDirty = 1
            if hasattr(child, '_url'):
                self._project.forgetChildByURL(child)
        child.project = None
        child.parent = None
        self.children = children
        self._wrapChildren()

    def serialize(self, writer):
        # just like parts, except there are children to consider.
        self._serializeHeader(writer)
        self._serializeChildren(writer)
        self._serializeTail(writer)

    def _serializeChildren(self, writer):
        children = self.getChildren()
        for part in children:
            part = UnwrapObject(part)
            part.serialize(writer)

    def set_name(self, name):
        self._name = name
        if hasattr(self, '_url') and self._url:
            # using lists to make it mutable
            parts = list(urlparse.urlparse(self._url))
            slashlocation = parts[2].rindex('/')
            parts[2] = parts[2][:slashlocation+1] + name
            if parts[2].find('.kpf') < 0:
                parts[2] = parts[2] + '.kpf'
            self._url = urlparse.urlunparse(tuple(parts))
        self.setStringAttribute('name', self._name)
        if self._project:
            self._project.isDirty = 1

    def set_url(self, url):
        # XXX should do URL validation
        self._url = url
        self.setStringAttribute('url', self._url)
        # Need to do the same checks as set_name.
        self._name = _url2basename(url)
        self.setStringAttribute('name', self._name)
        if self._project:
            self._project.isDirty = 1

    def get_url(self):
        return self._url

    def get_name(self):
        return self._name
    
    def clone(self):
        part = koPart.clone(self)
        for child in self.children:
            part.addChild(child.clone())
        return part
        
    def dump(self, indent):
        koPart.dump(self, indent)
        for child in self.children:
            child.dump(indent+2)

class koContainer(koContainerBase):
    def __init__(self):
        koContainerBase.__init__(self)

class koDialogPart(koContainer):
    _com_interfaces_ = [components.interfaces.koIPart_dialog,
                        components.interfaces.koIPart_SCC_file,
                        components.interfaces.koIPart_SCC_collection]
    _reg_desc_ = "Komodo GUI Builder Dialog part"
    _reg_contractid_ = "@activestate.com/koPart?type=dialog;1"
    _reg_clsid_ = "{80E175F0-401F-47d8-B317-E65D1223EFF0}"

    def __init__(self):
        koContainer.__init__(self)
        self.type = 'dialog'
        self.primaryInterface = 'koIPart_dialog'

    def getDragData(self):
        return self.getStringAttribute('url')

# See factory function below
class koFilePart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_file,
                        components.interfaces.koIPart_SCC_file
                       ]
    _reg_desc_ = "Komodo File"
    _reg_contractid_ = "@activestate.com/koPart?type=file;1"
    _reg_clsid_ = "{8FD46D97-B2F9-4875-8D36-82BFD7471A97}"
    
    def __init__(self):
        koPart.__init__(self)
        self.type = 'file'
        self.flavors.insert(0,'text/x-moz-url')
        self.primaryInterface = 'koIPart_file'
        
    def __str__(self):
        return "<koPart_file, url=%s>" % self._attributes.get('url', 'None')

    def getDragData(self):
        return self.getStringAttribute('url')

def File(url, name):
    """Construct a 'file' koIPart."""
    part = components.classes["@activestate.com/koPart?type=file;1"] \
                  .createInstance(components.interfaces.koIPart_file)
    part.setStringAttribute('url', url)
    part.setStringAttribute('name', name)
    return part


class koSnippetPart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_snippet]
    _reg_desc_ = "Komodo Snippet"
    _reg_contractid_ = "@activestate.com/koPart?type=snippet;1"
    _reg_clsid_ = "{D23CF82A-7E80-470d-8C6F-026193EF07F7}"

    def __init__(self):
        koPart.__init__(self)
        self.type = 'snippet'
        self.shortcut_command = 'cmd_insertNamedSnippet'
        self.shortcut_command_attribute = 'name'
        self.primaryInterface = 'koIPart_snippet'

    def getDragData(self):
        eol = '\n'
        # NOTE: IT IS IMPORTANT THAT IF UNICODE COMES IN, UNICODE GOES OUT!
        newtext = u'' # THE u IS IMPORTANT!
        text = self.value
        i = 0
        while i < len(text):
            if text[i] == '\\':
                i += 1
                if text[i] == 'n':
                    newtext += eol
                elif text[i] == 't':
                    newtext += '\t'
                elif text[i] == '\\':
                    newtext += '\\'
                else:
                    i -= 1
                    newtext == '\\'
            else:
                newtext += text[i]
            i += 1
        return newtext

class koCommandPart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_command]
    _reg_desc_ = "Komodo Command"
    _reg_contractid_ = "@activestate.com/koPart?type=command;1"
    _reg_clsid_ = "{AC02EC02-08B2-4c87-9861-D9917144A856}"

    def __init__(self):
        koPart.__init__(self)
        self.type = 'command'
        self.shortcut_command = 'cmd_runNamedCommand'
        self.shortcut_command_attribute = 'name'
        self.primaryInterface = 'koIPart_command'


class koTemplatePart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_template]
    _reg_desc_ = "Komodo Template"
    _reg_contractid_ = "@activestate.com/koPart?type=template;1"
    _reg_clsid_ = "{473BD963-632C-403e-953F-E7F2F2CA84A1}"

    def __init__(self):
        koPart.__init__(self)
        self.type = 'template'
        self.primaryInterface = 'koIPart_template'

class koURLPart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_URL]
    _reg_desc_ = "URL"
    _reg_contractid_ = "@activestate.com/koPart?type=URL;1"
    _reg_clsid_ = "{497C73A6-9405-4d0f-A472-8C7343AC5D60}"

    def __init__(self):
        koPart.__init__(self)
        self.type = 'URL'
        self.primaryInterface = 'koIPart_URL'

class koMacroPart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_macro]
    _reg_desc_ = "Macro"
    _reg_contractid_ = "@activestate.com/koPart?type=macro;1"
    _reg_clsid_ = "{4884D548-698E-45e2-9324-715F623CD300}"

    def __init__(self):
        koPart.__init__(self)
        self.type = 'macro'
        self.flavors.insert(0,'text/x-moz-url')
        self.shortcut_command = 'cmd_executeNamedMacro'
        self.shortcut_command_attribute = 'name'
        self.primaryInterface = 'koIPart_macro'

    def getDragData(self):
        return convertURL.LocalFileToURL(os.path.abspath(self.getStringAttribute('filename')))

    def evalAs(self, language, document, window, editor):
        #if editor:
        #    editor = editor.QueryInterface(components.interfaces.ISciMoz);
        #if window:
        #    window = window.QueryInterface(components.interfaces.nsIDOMWindowInternal);
        try:
            import komodo
        except ImportError:
            import new
            komodo = new.module('komodo')
            sys.modules['komodo'] = komodo
        komodo.document = document
        komodo.window = window
        komodo.editor = editor
        komodo.components = components
        #print "komodo.screen = ", komodo.screen
        fname = self.getAttribute('filename')
        if language == 'python':
            try:
                execfile(fname)
                return ''
            except:
                return ''.join(traceback.format_exception(*sys.exc_info()))
        elif language == 'perl':
            import perl
            try:
                fname = fname.replace('\\', '/')
                perl.eval('''\
                    use Python();
                    $komodo = Python::Import("komodo");
                    do "%s";
                    die $@ if $@;''' % fname)
            except:
                return ''.join(traceback.format_exception(*sys.exc_info()))                
            return ''

# See factory functions below
class koWebServicePart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_webservice]
    _reg_desc_ = "Komodo Web Service"
    _reg_contractid_ = "@activestate.com/koPart?type=webservice;1"
    _reg_clsid_ = "{00344B63-8DA5-4145-9D87-ABD13190FDC2}"
    
    def __init__(self):
        koPart.__init__(self)
        self.type = 'webservice'
        self.setStringAttribute('wsdl','');
        self.primaryInterface = 'koIPart_webservice'
        
    def getDragData(self):
        return self.getStringAttribute('wsdl')
        
# See factory functions below
class koFolderPart(koContainer):
    _com_interfaces_ = [components.interfaces.koIPart_folder,
                        components.interfaces.koIPart_SCC_directory
                       ]
    _reg_desc_ = "Komodo Folder"
    _reg_contractid_ = "@activestate.com/koPart?type=folder;1"
    _reg_clsid_ = "{7E774CEB-07F5-40da-AC37-954779B2F09F}"
    
    def __init__(self):
        koContainer.__init__(self)
        self.type = 'folder'
        self.primaryInterface = 'koIPart_folder'

    def importFromDisk(self, dirname, include, exclude, recursive,
                       useFolders, groupByType):
        #print dirname, include, exclude, recursive, useFolders
        if include != '':
            include_pats = include.split(';')
        else:
            include_pats = None
        if exclude != '':
            exclude_pats = exclude.split(';')
        else:
            exclude_pats = None
        if groupByType:
            _importFromDiskGroupByType(dirname, include_pats, exclude_pats,
                recursive,
                WrapObject(self, components.interfaces.koIContainer))
        else:
            if not recursive:
                names = os.listdir(os.path.abspath(dirname))
                names = _filterFiles(names, include_pats, exclude_pats)
                return names
            if useFolders:
                _importFromDiskNested(dirname, include_pats, exclude_pats,
                    WrapObject(self, components.interfaces.koIContainer))
            else:
                _importFromDiskFlat(dirname, include_pats, exclude_pats,
                    WrapObject(self, components.interfaces.koIContainer))

    def getDragData(self):
        if self.hasAttribute('url'):
            return self.getStringAttribute('url');
        return self.getStringAttribute('name');
    
    def getDragFlavors(self):
        if self.hasAttribute('url') and self.flavors[0] != 'text/x-moz-url':
            self.flavors.insert(0,'text/x-moz-url')
        return self.flavors        

def Folder(url, name):
    """Construct a koIPart_folder from URL and name."""
    part = components.classes["@activestate.com/koPart?type=folder;1"] \
                     .createInstance(components.interfaces.koIPart_folder)
    part.setStringAttribute('url', url)
    part.setStringAttribute('name', name)
    return part

def ExtensionFolder(extension):    
    """Construct a koIPart_folder for a particular file extension."""
    part = components.classes["@activestate.com/koPart?type=folder;1"] \
                     .createInstance(components.interfaces.koIPart_folder)
    part.type = 'extensionfolder'
    part.setStringAttribute('extension', extension)
    return part

def LanguageFolder(language):
    """Construct a koIPart_folder for a particular language type."""
    part = components.classes["@activestate.com/koPart?type=folder;1"] \
                     .createInstance(components.interfaces.koIPart_folder)
    part.type = 'languagefolder'
    part.setStringAttribute('language', language)
    return part

class koProjectRef(koFilePart):
    _com_interfaces_ = [components.interfaces.koIPart_ProjectRef]
    _reg_desc_ = "Project Reference"
    _reg_contractid_ = "@activestate.com/koPart?type=ProjectRef;1"
    _reg_clsid_ = "{0A027B8B-43B0-4da6-B26E-0426EA803208}"

    def __init__(self):
        koPart.__init__(self)
        self.type = 'ProjectRef'
        self.primaryInterface = 'koIPart_ProjectRef'
        
class koProject(koFolderPart):
    _com_interfaces_ = [components.interfaces.koIProject,
                        components.interfaces.koIPart_SCC_file,
                        components.interfaces.koIPart_SCC_collection
                       ]
    _reg_desc_ = "Komodo Project"
    _reg_contractid_ = "@activestate.com/koProject;1"
    _reg_clsid_ = "{39EE799F-D0F5-4b2d-A5A6-34151B2AC235}"
    
    def __init__(self):
        koFolderPart.__init__(self)
        self.type = 'project'
        self.isDirty = 0
        self._urlmap = weakref.WeakValueDictionary()
        self._project = self # cycle - need to investigate WeakReference
        self.primaryInterface = 'koIProject'
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                       .getService(components.interfaces.koILastErrorService)
        
    # Fill with a new empty, unnamed project 
    def create(self):
        self._name = _makeNewProjectName()
        self.prefset = components.classes["@activestate.com/koPreferenceSet;1"] \
                        .createInstance(components.interfaces.koIPreferenceSet)
        # XXX What should the ID of a project's prefset be?
        self.prefset.id = self._name
        self.setStringAttribute("kpf_version", KPF_VERSION_STRING)

    def _get_stream(self, path, mode):  # mode is 'r' or 'w'
        self.name = os.path.splitext(os.path.basename(path))[0]
        stream = components.classes["@activestate.com/koLocalFile;1"].\
             createInstance().QueryInterface(components.interfaces.koILocalFile)
        try:
            stream.init(path, mode)
        except (IOError, COMException), e:
            errstr = "Can't open project file %s because %s" % (path, e)
            logError(errstr)
            raise 
        return stream

    def load(self, url):
        self.lastErrorSvc.setLastError(0, '')
        self.set_url(url)
        fname = convertURL.LocalPathFromURL(self._url)
        if not os.path.exists(fname):
            self.lastErrorSvc.setLastError(0, 'File does not exist.')
            raise ServerException
            
        stream = self._get_stream(fname, 'r')
        self.children=[]
        try:
            rootNode = minidom.parse(stream)
        except Exception, e:
            self.lastErrorSvc.setLastError(0, "Project file is corrupt")
            logError(str(e))
            raise ServerException
       
        node = rootNode.childNodes[0]

        for attribute in node.attributes.keys():
            value = node.getAttribute(attribute)
            self.setStringAttribute(attribute, value)

        dirtyAtEnd = 0
        if 'kpf_version' not in node.attributes.keys():
            # We're dealing with Project files from Komodo 1.2 and before.
            # We want to get rid of the 'files' child of the 'project' node.

            # skip Text child Node.
            for childNode in node.childNodes:
                if childNode.nodeName == '#text':
                    continue
                assert childNode.tagName == 'files'
                node = childNode
                break
            # We'll upgrade on write
            self.setStringAttribute('kpf_version', KPF_VERSION_STRING)
            # Note that we're dirty since we changed the format.
            # XXX should notify the user somehow...
            dirtyAtEnd = 1
        
        for childNode in node.childNodes:
            if childNode.nodeName == '#text':
                continue
            if childNode.tagName == 'preference-set':
                # prefsets get special handling -- they're not really children, they're
                # more like very rich attributes
                self.prefset = dedomify(childNode, self)
            else:
                self.addChild(dedomify(childNode, self))
        
        if not self.prefset:
            self.prefset = components.classes["@activestate.com/koPreferenceSet;1"] \
                  .createInstance(components.interfaces.koIPreferenceSet)
            self.prefset.id = self._name # XXX
        self._loaded_from_url = self._url

        self.set_url(url)

        # this kicks off background status checking for the projects base path
        # the project url's will be added later, from partWrapper.js
        self._observerSvc.Notify(self,'file_project',os.path.dirname(self._url))

        self.isDirty = dirtyAtEnd

    def save(self):
        import tempfile
        # use a tempfile to avoid disasters if there's a failure on write.
        tempname = tempfile.mktemp()
        try:
            stream = self._get_stream(tempname, 'w')
            try:
                stream.write('<?xml version="1.0"?>'+newl)
                stream.write('<!-- Komodo Project File - DO NOT EDIT -->'+newl)
                self.serialize(stream)
            finally:
                stream.close()
            # if we get this far we have probably saved it correctly.
            shutil.copyfile(tempname, convertURL.LocalPathFromURL(self._url))
            self.isDirty = 0
            self._loaded_from_url = self._url
        finally:            
            os.unlink(tempname)

    def revert(self):
        self._url = self._loaded_from_url
        self.load(self._url)
        self.isDirty  = 0
        
    def dump(self, indent):
        print " "*indent + "project:"
        for x in self.__dict__.keys():
            print ' '*(indent+1) + x + ':' + str(self.__dict__[x])
        koFolderPart.dump(self, indent)

    def registerChildByURL(self, child):
        if child.hasAttribute('url'):
            url = child.getStringAttribute('url')
            self._urlmap[url] = child
        child = UnwrapObject(child) # this is necessary for the getattr to work.
        kids = getattr(child, 'children', None)
        if kids:
            # You don't just adopt someone, you adopt their extended family
            for _child in kids:
                self.registerChildByURL(_child)
        # some children don't-- that's ok.

    def forgetChildByURL(self, child):
        if child.hasAttribute('url'):
            url = child.getStringAttribute('url')
            if url in self._urlmap:
                del self._urlmap[url]

    def getChildByURL(self, url):
        return self._urlmap.get(url, None)

    def close(self):
        # this is primarily for removing watched files from the
        # file status service
        self.isDirty = 0
        for url in self._urlmap.keys():
            self._observerSvc.Notify(self,'file_removed',url)
        self._observerSvc.Notify(self,'file_removed',self._url)


#---- Utility routines

# Dedomification code -- converts minidom DOM nodes into koParts,
# koPreferenceSets, etc.

def dedomify_prefset(element, project):
    return NodeToPrefset(element)

dedomifiers = {
    'preference-set': dedomify_prefset,
}

def default_dedomifier(element, project):
    # This function needs to create a part or a subclass of a part,
    # given a DOM First, we try to find an interface for that particular
    # part. We need to find out if the element has children -- if yes,
    # then we'll treat it as a folder.
    hasNonTextChildNodes = len(filter(lambda x: x.nodeName != '#text',
                                      element.childNodes)) > 0
    try:
        interfaceId= "@activestate.com/koIPart_%s" % (element.tagName)
        el_interface = components.interfaces[interfaceId]
    except COMException, unused:
        # there is no special interface for this element type -- use
        # koIPart if has no children, koIContainerBase otherwise
        if hasNonTextChildNodes:
            el_interface = components.interfaces.koIContainerBase
        else:
            el_interface = components.interfaces.koIPart

    #  Now we look for an implementation of that interface
    try:
        componentId = "@activestate.com/koPart?type=%s;1" % element.tagName
        el_class = components.classes[componentId]
        part = el_class.createInstance(el_interface)
    except COMException, e:
        if hasNonTextChildNodes:
            part = components.classes['@activestate.com/koPart?type=folder;1']\
                             .createInstance(el_interface);
        else:
            part = components.classes['@activestate.com/koPart;1']\
                             .createInstance(el_interface);

    # at this point we have a class and an interface
    part.type = element.tagName
    for attribute in element.attributes.keys():
        value = element.getAttribute(attribute)
        # XXX need to decentralize this processing
        if (attribute == 'url' or attribute == 'startup') and value.find(':') == -1: # it is not an real URL -- it must be relative
            # URLs of files and "startup" attributes of Dialogs need to be
            # de-relativized.
            projecturl = project._url
            baseurl = projecturl[:projecturl.rindex('/')+1] # include the /
            # XXX This is where we need to be smarter if we do a better job at RelativizeURL in convertURL
            value = baseurl + value  # need to be smarter probably!
        part.setStringAttribute(attribute, value)
    part.value = ''
    for childNode in element.childNodes:
        if childNode.nodeName == "#text":
            if childNode.nodeValue != '\n':
                # for some reason we get spurious nodes like these.
                # XXX this is suspected to be a minidom bug.
                part.value += childNode.nodeValue
        elif childNode.tagName == 'preference-set':
            part.prefset = dedomify(childNode, project)
        else:
            childPart= dedomify(childNode, project)
            part.addChild(childPart)
    return part

def dedomify(element, project):
    return dedomifiers.get(element.tagName, default_dedomifier)(element,
                                                                project)


# Code used to group filenames in language groups.
def _initKoLanguageRegistryService():
    global koLanguageRegistryService
    if koLanguageRegistryService is None:
        koLanguageRegistryService = components.classes["@activestate.com/koLanguageRegistryService;1"] \
            .getService(components.interfaces.koILanguageRegistryService);

def mapNamesToLanguages(names):
    if koLanguageRegistryService is None: _initKoLanguageRegistryService()
    newmap = {}
    for name in names:
        language = koLanguageRegistryService.suggestLanguageForFile(name)
        if not language: language = 'Other'
        newmap.setdefault(language, []).append(name)
    return newmap


# Routines that do the actual importing.
def _importFromDiskGroupByType(dirname, include_pats, exclude_pats,
                               recursive, part):
    """Find all files matching include_pats, not matching exclude_pats,
    in dirname and subdirs.
    """

    # start with current directory contents
    allnames = os.listdir(os.path.abspath(dirname))
    allnames = _filterFiles(allnames, include_pats, exclude_pats)
    allnames = [os.path.join(dirname, name) for name in allnames]

    if recursive:
        # now look at children
        for subdirname in _dirwalker(dirname):
            names = os.listdir(os.path.abspath(subdirname))
            #print 'FOUND DIR', subdirname, "with children", names
            names = _filterFiles(names, include_pats, exclude_pats)
            names = [os.path.join(subdirname, name) for name in names]
            allnames.extend(names)

    # remove directories
    allnames = [name for name in allnames\
                if os.path.isfile(os.path.join(dirname, name))]
    languageMap = mapNamesToLanguages(allnames)
    languages = languageMap.keys()
    languages.sort()
    for language in languages:
        folder = part.getChildWithTypeAndStringAttribute('languagefolder',
                                                         'language', language)
        if not folder:
            folder = LanguageFolder(language)
            part.addChild(folder)
        for file in languageMap[language]:
            url = convertURL.LocalFileToURL(os.path.abspath(file))
            if not part.project.getChildByURL(url):
                basename = os.path.basename(file)
                filepart = File(url, basename)
                folder.addChild(filepart)


def _importFromDiskFlat(dirname, include_pats, exclude_pats, part):
    """Find all files matching include_pats, not matching exclude_pats,
    in dirname and subdirs.
    """
    # start with current directory contents
    allnames = os.listdir(os.path.abspath(dirname))
    allnames = _filterFiles(allnames, include_pats, exclude_pats)
    allnames = [os.path.join(dirname, name) for name in allnames]
    
    # now look at children
    for subdirname in _dirwalker(dirname):
        names = os.listdir(os.path.abspath(subdirname))
        names = _filterFiles(names, include_pats, exclude_pats)
        names = [os.path.join(subdirname, name) for name in names]
        allnames.extend(names)
    for name in allnames:
        url = convertURL.LocalFileToURL(os.path.abspath(name))
        if not part.project.getChildByURL(url):
            basename = os.path.basename(name)
            filepart = File(url, basename)
            part.addChild(filepart)


def _importFromDiskNested(dirname, include_pats, exclude_pats, part):
    # start with current directory contents
    allnames = os.listdir(os.path.abspath(dirname))
    allnames = _filterFiles(allnames, include_pats, exclude_pats)
    # remove directories
    subdirnames = [os.path.join(dirname, name) for name in allnames\
                if os.path.isdir(os.path.join(dirname, name))]
    allnames = [os.path.join(dirname, name) for name in allnames\
                if os.path.isfile(os.path.join(dirname, name))]
    for name in allnames:
        url = convertURL.LocalFileToURL(os.path.abspath(name))
        if not part.project.getChildByURL(url):
            basename = os.path.basename(name)
            filepart = File(url, basename)
            part.addChild(filepart)
    for subdirname in subdirnames:
        fullsubdirname = os.path.join(dirname, subdirname)
        url = convertURL.LocalFileToURL(os.path.abspath(fullsubdirname))
        subdirpart = part.project.getChildByURL(url)
        if not subdirpart:
            basename = os.path.basename(fullsubdirname)
            subdirpart = Folder(url, basename)
            part.addChild(subdirpart)
        _importFromDiskNested(fullsubdirname, include_pats, exclude_pats,
                              subdirpart)
    
def _dirwalker(dirname):
    """Generate all subdirectories of the given directory."""
    contents = os.listdir(dirname)
    for f in contents:
        fullpath = os.path.join(dirname, f)
        if os.path.isdir(fullpath):
            yield fullpath
            for subdirname in _dirwalker(fullpath):
                yield subdirname

def _filterFiles(names, include_pats, exclude_pats):
    """Filter the names array for filenames which match any of the
    patterns in include_pats and match none of the patterns in
    exclude_pats -- returns a list containing a selection of the input
    names.
    
    if include_pats is None then all files match are considered by
    default
    
    if exclude_pats is None then no files are excluded explicitly
    """
    oknames = []
    for name in names:
        try:
            if include_pats is None:
                oknames.append(name)
                raise 'done'
            for pat in include_pats:
                if fnmatch.fnmatch(name, pat):
                    oknames.append(name)
                    raise 'done'
        except 'done':
            if exclude_pats:
                try:
                    for pat in exclude_pats:
                        if fnmatch.fnmatch(name, pat):
                            oknames.remove(name)
                            raise 'done'
                except 'done':
                    continue
    return oknames

