#!python
# Copyright (c) 2000-2002 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

# The implementation of the Komodo Run Output Window Terminal

import threading
import process
import time

from koSciMozEvents import SciMozEvents
from koEditorController import CommandHandler
from xpcom import components
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject
from koLogger import logWarning, logError, logDebug


#---- globals

# 1==buffer output to scintilla as an optimization (old, no longer need
#    for performance, XXX keep code around for now -- should drop before
#    final release)
# 0==do not do buffering (default)
_WRITE_BUFFERING = 0


#---- implementation

class _KoRunTerminalFile(process.IOBuffer):
    """A std handler for the Run output window.
    I.e. one each for <stdin>, <stdout>, and <stderr>.
    """
    _com_interfaces_ = [components.interfaces.koIFile]

    def __init__(self, name, scintilla, terminal):
        #print "[%s] _KoRunTerminalFile.__init__(name=%r)" % (name, name)
        self._scintilla = scintilla
        self._terminal = terminal
        process.IOBuffer.__init__(self, self._terminal.mutex,
                                  self._terminal.stateChange, name=name)

        if _WRITE_BUFFERING:
            self._buf = ""
            self._closed = 0
            self._lastWriteInTime = time.time()
            self._lastWriteOutTime = time.time()
            self._writeInDelay = 0.100 # write out if have not received for 100ms
            self._writeOutDelay = 1.000 # write out if have not for 1s
            self._bufLock = threading.RLock()
            self._bufEvent = threading.Condition()
            self._bufferer = threading.Thread(target=self._buffering)
            self._bufferer.start()

    if _WRITE_BUFFERING:
        def _buffering(self):
            while 1:
                t = time.time()
                if self._closed:
                    #print "[%s] _KoRunTerminalFile._buffering: flush buffer "\
                    #      "(closed), write %r" % (self._name, self._buf)
                    self._bufLock.acquire()
                    s, self._buf = self._buf, ""
                    self._bufLock.release()
                    if s:
                        self.__reallyDoWrite(s)
                    break
                elif self._buf:
                    if self._buf.find('\n') != -1 or\
                       t - self._lastWriteOutTime >= self._writeOutDelay or\
                       t - self._lastWriteInTime >= self._writeInDelay:
                        #if self._buf.find('\n') != -1:
                        #    reason = "newline"
                        #elif t - self._lastWriteOutTime >= self._writeOutDelay:
                        #    reason = "write out delay"
                        #elif t - self._lastWriteInTime >= self._writeInDelay:
                        #    reason = "write in delay"
                        #print "[%s] _KoRunTerminalFile._buffering: flush "\
                        #      "buffer (%s), write %r"\
                        #      % (self._name, reason, self._buf)
                        self._bufLock.acquire()
                        s, self._buf = self._buf, ""
                        self._bufLock.release()
                        if s:
                            self.__reallyDoWrite(s)
                        # Now wait until next event.
                        #print "[%s] _KoRunTerminalFile._buffering: wait for "\
                        #      "next event" % self._name
                        self._bufEvent.acquire()
                        self._bufEvent.wait()
                        self._bufEvent.release()
                        #print "[%s] _KoRunTerminalFile._buffering: done "\
                        #      "waiting for next event" % self._name
                    else:
                        # Wait until should next write.
                        delay = min(t - self._lastWriteOutTime,
                                    t - self._lastWriteInTime)
                        #print "[%s] _KoRunTerminalFile._buffering: sleep "\
                        #      "for %fs" % (self._name, delay)
                        time.sleep(delay)
                else:
                    # Now wait until next event.
                    #print "[%s] _KoRunTerminalFile._buffering: wait for "\
                    #      "next event" % self._name
                    self._bufEvent.acquire()
                    self._bufEvent.wait()
                    self._bufEvent.release()
                    #print "[%s] _KoRunTerminalFile._buffering: done waiting "\
                    #      "for next event" % self._name
            #print "[%s] _KoRunTerminalFile._buffering: done" % self._name

    def _doWrite(self, s):
        #print "[%s] _KoRunTerminalFile._doWrite: write in: %r"\
        #      % (self._name, s)
        if _WRITE_BUFFERING:
            self._bufLock.acquire()
            self._buf += s
            self._lastWriteInTime = time.time()
            self._bufLock.release()
            self._bufEvent.acquire()
            self._bufEvent.notify()
            self._bufEvent.release()
        else:
            self.__reallyDoWrite(s)

    if _WRITE_BUFFERING:
        def close(self):
            #print "[%s] _KoRunTerminalFile.close()" % self._name
            self._bufLock.acquire()
            self._closed = 1
            self._bufLock.release()
            self._bufEvent.acquire()
            self._bufEvent.notify()
            self._bufEvent.release()
            self._bufferer.join()
            return process.IOBuffer.close(self)

    def __reallyDoWrite(self, s):
        if _WRITE_BUFFERING:
            self._bufLock.acquire()
        #print "[%s] _KoRunTerminalFile.__reallyDoWrite: write out: %r"\
        #      % (self._name, s)

        self._scintilla.readOnly = 0
        if self._name == "<stderr>":
            start = self._scintilla.length
            self._scintilla.addText(len(s), s)
            # we want to set all of the non-indicator bits.  This means 5 LSBs, which appears to be the default
            self._scintilla.startStyling(start, 0x1f);  
            self._scintilla.setStyling(len(s), self._scintilla.SCE_C_COMMENT);
        else:
            self._scintilla.addText(len(s), s)
        self._scintilla.gotoPos(self._scintilla.length)
        self._scintilla.readOnly = 1

        if _WRITE_BUFFERING:
            self._lastWriteOutTime = time.time()
            self._bufLock.release()

    def _doRead(self, n):
        """Pop 'n' bytes from the terminal in-buffer and return them."""
        #print "[%s] _KoRunTerminalFile._doRead(n=%r)" % (self._name, n)
        if n < 0:
            idx = len(self._terminal.inBuffer)
        else:
            idx = min(n, len(self._terminal.inBuffer))
        retval = self._terminal.inBuffer[:idx]
        self._terminal.inBuffer = self._terminal.inBuffer[idx:]
        return retval

    def _doReadline(self, n):
        """Pop the front line (or n bytes of it, whichever is less) from
        the internal buffer and return it.
        """
        #print "[%s] _KoRunTerminalFile._doReadline(n=%r)" % (self._name, n)
        idx = self._terminal.inBuffer.find('\n')
        if idx == -1:
            idx = len(self._terminal.inBuffer)
        else:
            idx += 1 # include the '\n'
        if n is not None:
            idx = min(idx, n) 
        retval = self._terminal.inBuffer[:idx]
        self._terminal.inBuffer = self._terminal.inBuffer[idx:]
        return retval

    def _haveLine(self):
        #print "[%s] _KoRunTerminalFile._haveLine()" % self._name
        return len(self._terminal.inBuffer) > 0

    def _haveNumBytes(self, n):
        #print "[%s] _KoRunTerminalFile._haveNumBytes(n=%r)"\
        #      % (self._name, n)
        # Note: Checking for >0 rather than >=n because don't want to
        #       do any buffering.
        return len(self._terminal.inBuffer) > 0


class _KoRunTerminalCommandHandler(CommandHandler):
    _com_interfaces_ = components.interfaces.nsIController

    def __init__(self, scintilla, terminal):
        self.wrapper = None  #XXX make CommandHandler happy
        self._terminal = terminal  #XXX creates a cycle!
        self._scintilla = scintilla

    def _do_cmd_newline(self):
        #print "XXX [%s] _KoRunTerminalCommandHandler._do_cmd_newline()" % id(self)
        if self._terminal.active:
            self._terminal.mutex.acquire()
            # .newLine() will cause cause the terminal's onCharAdded
            # handler to be called, adding the EOL marker to lineBuffer,
            # so copy lineBuffer beforehand, and clear it afterward.
            #XXX Do we have to explicitly use '\n' and not just take the
            #    native newline terminator added by scintilla.newLine()?
            self._terminal.inBuffer = self._terminal.lineBuffer + '\n'
            #print "inBuffer = ", self._terminal.lineBuffer
            self._scintilla.newLine()
            self._terminal.lineBuffer = ''
            self._terminal.mutex.release()
            # Notify the terminal in-file of the new input.
            self._terminal.stateChange.acquire()
            self._terminal.stateChange.notify()
            self._terminal.stateChange.release()

    def _do_cmd_backSmart(self):
        #print "XXX [%s] _KoRunTerminalCommandHandler._do_cmd_backSmart()" % id(self)
        if self._terminal.active and self._terminal.lineBuffer:
            self._terminal.mutex.acquire()
            self._scintilla.readOnly = 0
            self._scintilla.selectionStart = self._scintilla.length - 1
            self._scintilla.selectionEnd = self._scintilla.length
            self._scintilla.replaceSel('')
            self._scintilla.readOnly = 1
            self._terminal.lineBuffer = self._terminal.lineBuffer[:-1]
            self._terminal.mutex.release()


class KoRunTerminal(SciMozEvents):
    """This acts as the handler for events on a <scintilla> window (e.g.
    the output2 window).

    This manages a <scintilla> terminal window (generally for
    communication with external processes). The embedded Scintilla is
    kept in a 'readOnly' state unless currently interacting with the
    child process (self.active is True). XXX Perhaps there is a
    better API for that (perhaps call it 'interactive')?

    Input is buffered in line chunks, i.e. user input is forwarded to
    the child a line at a time when <Enter> is pressed. This means that
    one can use backspace on the current line only.
    """
    _com_interfaces_ = [components.interfaces.koIRunTerminal,
                        components.interfaces.ISciMozEvents]
    _reg_clsid_ = "{A077FDDF-DBAD-4e49-9090-4390C084893B}"
    _reg_contractid_ = "@activestate.com/koRunTerminal;1"
    _reg_desc_ = "Run Output Window Terminal Handler"

    def init(self, scintilla):
        #print "XXX [%s] KoRunTerminal.init(...)" % id(self)
        self._unproxiedScintilla = scintilla
        self._scintilla = getProxyForObject(None, components.interfaces.ISciMoz,
                                            self._unproxiedScintilla,
                                            PROXY_ALWAYS | PROXY_SYNC)
        self._scintilla.styleSetFore(self._scintilla.SCE_C_COMMENT, 0x0000bb);
        self.mutex = threading.RLock()
        self.stateChange = threading.Condition()
        self.controller = _KoRunTerminalCommandHandler(scintilla, self)

        # Initially NOT interacting with a child process.
        self.active = 0

    def startSession(self):
        self.clear()
        self.stdin = _KoRunTerminalFile("<stdin>", self._scintilla, self)
        self.stdout = _KoRunTerminalFile("<stdout>", self._scintilla, self)
        self.stderr = _KoRunTerminalFile("<stderr>", self._scintilla, self)
        self.active = 1

    def endSession(self):
        self.active = 0
        #XXX Do they have to be closed manually here?
        self.stdin = None
        self.stdout = None
        self.stderr = None
    
    def get_controller(self):
        #XXX necessary?
        return self.controller

    def clear(self):
        #print "XXX KoRunTerminal.clear()"
        self.mutex.acquire()
        self._scintilla.readOnly = 0
        self._scintilla.text = ''
        self.lineBuffer = ''    # Current line input.
        self.inBuffer = ''      # All unread input except the current line.
        self._scintilla.gotoPos(self._scintilla.length)
        self._scintilla.readOnly = 1
        self.mutex.release()

    def onCharAdded(self, char):
        #print "XXX [%s] KoRunTerminal.onCharAdded(char=%r)" % (id(self), char) 
        if self.active:
            self.mutex.acquire()
            self._scintilla.readOnly = 0
            self._scintilla.insertText(self._scintilla.length, chr(char))
            self._scintilla.gotoPos(self._scintilla.length)
            self._scintilla.readOnly = 1
            self.lineBuffer += chr(char)
            self.mutex.release()

    def onZoom(self):
        s = self._scintilla
        if s.getMarginWidthN(0):
            s.setMarginWidthN(0, s.textWidth(0, str(s.lineCount))+5)
