# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

from xpcom import components
from koLogger import logWarning
import string

command_map = {
    'cmd_selectHome' : 'vCHomeExtend',
    'cmd_selectEnd' : 'lineEndExtend',
    'cmd_selectCharPrevious' : 'charLeftExtend',
    'cmd_selectCharNext' : 'charRightExtend',
    'cmd_pageUp' : 'pageUp',
    'cmd_pageDown' : 'pageDown',
    'cmd_selectPageUp' : 'pageUpExtend',
    'cmd_selectPageDown' : 'pageDownExtend',
    'cmd_selectLinePrevious' : 'lineUpExtend',
    'cmd_selectLineNext' : 'lineDownExtend',
    'cmd_undo' : 'undo',
    'cmd_redo' : 'redo',
    'cmd_cut' : 'cut',
    'cmd_copy' : 'copy',
    'cmd_paste' : 'paste',
    'cmd_selectWordLeft' : 'wordLeftExtend',
    'cmd_selectWordRight' : 'wordRightExtend',
    'cmd_wordPartLeftExtend' : 'wordPartLeftExtend',
    'cmd_wordPartRightExtend' : 'wordPartRightExtend',
    'cmd_deleteWordLeft' : 'delWordLeft',
    'cmd_deleteWordRight' : 'delWordRight',
    'cmd_documentHome' : 'documentStart',
    'cmd_documentEnd' : 'documentEnd',
    'cmd_selectDocumentHome' : 'documentStartExtend',
    'cmd_selectDocumentEnd' : 'documentEndExtend',
    'cmd_editSelectAll' : 'selectAll',
    'cmd_delete' : 'clear',
    'cmd_back' : 'deleteBack',
    'cmd_lineScrollUp' : 'lineScrollUp',
    'cmd_lineScrollDown' : 'lineScrollDown',
    'cmd_lineCut' : 'lineCut',
    'cmd_lineDelete' : 'lineDelete',
#    'cmd_lineTranspose' : 'lineTranspose',  # commented out because its undo behavior is wrong.
    'cmd_fontZoomIn' : 'zoomIn',
    'cmd_fontZoomOut' : 'zoomOut',
    'cmd_toggleOvertype' : 'editToggleOvertype',
    'cmd_newline' : 'newLine',
}

undoable_commands = {
    'cmd_lineCut' : 1,
    'cmd_lineDelete' : 1,
    'cmd_lineTranspose' : 1,
    'cmd_deleteWordLeft' : 1,
    'cmd_deleteWordRight' : 1,
}

class koScintillaController:
    _com_interfaces_ = components.interfaces.ISciMozController
    _reg_clsid_ = "{726cc885-6d17-48af-b8a6-c9b759f1fe6b}"
    _reg_contractid_ = "@ActiveState.com/scintilla/controller;1"

    def init(self, scimoz):
        self.scimoz = scimoz
        self._lastcutposition = None
        # A cached collection of boolean attributes.
        # No key == "dunno if boolean"
        # Key == None == definitely not bool.
        # Else - key == scimoz attribute name.
        self.bool_attributes = {}

    def isCommandEnabled( self, command_name ):
        # Result: boolean
        # In: param0: wstring
        meth = getattr(self, "_is_%s_enabled" % (str(command_name),), None)
        if meth is None:
            # Handle the fact it may be a bool property that can be toggled.
            # (and cache the fact, to speed things up 2nd and later times)
            attr = self.bool_attributes.get(command_name, 0)
            if attr == 0: # not in map
                attr = prop_name = command_map.get(command_name)
                if attr is not None:
                    attr = getattr(self.scimoz, attr)
                    if type(attr) == type(0):
                        # Cache for next time.
                        self.bool_attributes[command_name] = prop_name
                        # And to avoid re-fetching, get out here!
                        return attr
                    # Not an attribute we can use - that is OK.
                    attr = None

            if attr is None:
                # No custom function, and no scintilla integer property of that name.
                # Assume OK.
                rc = 1
            else:
                # We have a remembered property name - use it.
                rc = getattr(self.scimoz, attr)
        else:
            # Custom function - do it
            rc = meth()
        return rc

    def supportsCommand( self, command_name ):
        # Result: boolean
        # In: param0: wstring

        return command_name in command_map or hasattr(self, '_do_'+command_name)

    def _DoTheSciMozWiggle(self):
        # this code is necessary because of a bug in scintilla, by which
        # the column number is not always kept up to date.
        # This shows up in up and down arrow moving to the "old" columnar position
        # after finds, line cuts, and other actions.
        # to fix this, we wiggle - a bit to the left, a bit to the right
        scimoz = self.scimoz
        # this is the wiggle needed to set lastX
        c = scimoz.currentPos
        a = scimoz.anchor
        scimoz.currentPos = c
        scimoz.anchor = c
        if scimoz.currentPos > 0:
            scimoz.charLeft()
            scimoz.charRight()
        else:
            scimoz.charRight()
            scimoz.charLeft()
        scimoz.currentPos = c
        scimoz.anchor = a

    def doCommand( self, command_name ):
        # Result: void - None
        # In: param0: wstring
        sm = self.scimoz
        old_sel_exists = sm.selectionEnd != sm.selectionStart
        currentPos = sm.currentPos
        if command_name == 'cmd_undo':
            self._lastcutposition = None
        if command_name == 'cmd_copy':
            # If there is a selection, we just do the usual cut
            if sm.selectionStart != sm.selectionEnd:
                sm.copy()
                return
            # if there's no selection, we copy the current line, being careful to leave
            # the cursor in its original position.
            # but should we?  Most of the time one will want to place the line somewhere else.
            # Hmm -- I think I'll leave the cursor at the beginning of the copied line for now.
            oldCurrentPos = sm.currentPos
            lineStart = sm.lineFromPosition(sm.currentPos)
            sm.selectionStart = sm.positionFromLine(lineStart)
            sm.selectionEnd = sm.positionFromLine(lineStart+1)
            sm.copy()
            #sm.currentPos = sm.selectionStart = sm.selectionEnd = oldCurrentPos
            sm.currentPos = sm.selectionEnd = sm.selectionStart
            self._DoTheSciMozWiggle()
        elif command_name == 'cmd_cut':
            # If there is a selection, we just do the usual cut
            if sm.selectionStart != sm.selectionEnd:
                sm.cut()
                self._lastcutposition = None
                return
            # if there's no selection, we get to do our fancy cut.
            # If the last cut position was different that the current position, this is the 'first cut' -- e.g. a lineCut

            if sm.currentPos == sm.textLength: return # Do nothing at end of file
            if self._lastcutposition != sm.currentPos:
                # first cut -- lineCut
                sm.lineCut()
                self._DoTheSciMozWiggle()
                self._lastcutposition = sm.currentPos
                self._numCutLines = 1
                return
            # If we're here, it's because someone did Cut twice w/o a selection -- we need to accumulate the
            # lines into the cut buffer
            lineStart = sm.lineFromPosition(sm.currentPos)
            start = sm.positionFromLine(lineStart);
            sm.selectionStart = start
            sm.selectionEnd = start
            # we undo the previous cut and redo a "bigger" cut so that the undo stack is sane.
            sm.undo();
            end = sm.positionFromLine(lineStart + 1 + self._numCutLines)
            self._numCutLines += 1
            sm.selectionStart = start
            sm.selectionEnd = end
            sm.cut();
            self._DoTheSciMozWiggle()
            return
        methname= '_do_'+command_name
        attr = getattr(self, methname, None)
        if attr is None:
            attr = getattr(sm, command_map[command_name])
        # If we fetch an attribute and it is a number..
        if type(attr)==type(0): # Assume boolean
            setattr(sm, command_map[command_name], not attr)
        # Usually it will be a method..
        elif callable(attr):
            attr()
        # and hopefully this will never happen!
        else:
            raise TypeError, "The command map entry '%s' yielded a '%r' - dunno what to do with it" % (command_name, attr)
        # We only need to send a command update when the selection changes _shape_, not location
        # (ie, only when it changes from no selection to selection, or vice-versa.
        new_sel_exists = sm.selectionEnd != sm.selectionStart
        if old_sel_exists != new_sel_exists:
            sm.sendUpdateCommands("select")

    def onEvent( self, param0 ):
        # Result: void - None
        # In: param0: wstring
        logWarning("Scintilla controller caught an event: " % param0)

    def _has_sel(self):
        return self.scimoz.selectionEnd != self.scimoz.selectionStart
    # Specific command handlers
    def _is_cmd_undo_enabled(self):
        return self.scimoz.canUndo()
    def _is_cmd_redo_enabled(self):
        return self.scimoz.canRedo()
    def _is_cmd_paste_enabled(self):
        return self.scimoz.canPaste()
    def _is_cmd_cut_enabled(self):
        return not self.scimoz.readOnly
    def _is_cmd_copy_enabled(self):
        return self.scimoz.textLength != 0

    def _is_cmd_pasteAndSelect_enabled(self):
        return self.scimoz.canPaste()
    def _do_cmd_pasteAndSelect(self):
        start = self.scimoz.currentPos
        self.scimoz.paste()
        self.scimoz.anchor = start

    def _do_cmd_endOfWord(self):
        sm = self.scimoz
        sm.beginUndoAction()
        curChar = chr(sm.getCharAt(sm.currentPos)) 
        nextChar = chr(sm.getCharAt(sm.currentPos+1))# XXX check for end of buffer
        if charClass.get(curChar, 'special') != charClass.get(nextChar, 'special'):
            sm.charRight()
        while charClass.get(curChar, 'special') == charClass.get(nextChar, 'special'):
            sm.charRight()
            curChar = nextChar
            nextChar = chr(sm.getCharAt(sm.currentPos))
        sm.endUndoAction()
        return 1

    def _do_cmd_endOfWordExtend(self):
        sm = self.scimoz
        sm.beginUndoAction()
        curChar = chr(sm.getCharAt(sm.currentPos)) 
        nextChar = chr(sm.getCharAt(sm.currentPos+1))# XXX check for end of buffer
        if charClass.get(curChar, 'special') != charClass.get(nextChar, 'special'):
            sm.charRightExtend()
        while charClass.get(curChar, 'special') == charClass.get(nextChar, 'special'):
            sm.charRightExtend()
            curChar = nextChar
            nextChar = chr(sm.getCharAt(sm.currentPos))
        sm.endUndoAction()
        return 1

    def _do_cmd_beginningOfWord(self):
        sm = self.scimoz
        sm.beginUndoAction()
        curChar = chr(sm.getCharAt(sm.currentPos)) 
        nextChar = chr(sm.getCharAt(sm.currentPos-1))# XXX check for end of buffer
        if charClass.get(curChar, 'special') != charClass.get(nextChar, 'special'):
            sm.charLeft()
        curChar = chr(sm.getCharAt(sm.currentPos)) 
        nextChar = chr(sm.getCharAt(sm.currentPos-1))# XXX check for end of buffer
        while charClass.get(curChar, 'special') == charClass.get(nextChar, 'special'):
            sm.charLeft()
            curChar = nextChar
            nextChar = chr(sm.getCharAt(sm.currentPos-1))
        sm.endUndoAction()
        return 1

    def _do_cmd_beginningOfWordExtend(self):
        sm = self.scimoz
        sm.beginUndoAction()
        curChar = chr(sm.getCharAt(sm.currentPos)) 
        nextChar = chr(sm.getCharAt(sm.currentPos-1))# XXX check for end of buffer
        if charClass.get(curChar, 'special') != charClass.get(nextChar, 'special'):
            sm.charLeftExtend()
        curChar = chr(sm.getCharAt(sm.currentPos)) 
        nextChar = chr(sm.getCharAt(sm.currentPos-1))# XXX check for end of buffer
        while charClass.get(curChar, 'special') == charClass.get(nextChar, 'special'):
            sm.charLeftExtend()
            curChar = nextChar
            nextChar = chr(sm.getCharAt(sm.currentPos-1))
        sm.endUndoAction()
        return 1

    def _do_cmd_lineTranspose(self):
        sm = self.scimoz
        sm.beginUndoAction()
        sm.lineTranspose();
        sm.endUndoAction()
        return 1
    
    def _do_cmd_join(self):
        sm = self.scimoz
        sm.beginUndoAction()
        sm.lineEnd()
        sm.wordRightExtend()
        sm.replaceSel(' ')
        sm.charLeft()
        sm.endUndoAction()

    def _do_cmd_linePrevious(self):
        sm = self.scimoz
        sm.currentPos = min(sm.currentPos, sm.anchor)
        sm.lineUp()
        
    def _do_cmd_lineNext(self):
        sm = self.scimoz
        sm.currentPos = max(sm.currentPos, sm.anchor)
        sm.lineDown()
        
    def _do_cmd_left(self):
        sm = self.scimoz
        sm.currentPos = min(sm.currentPos, sm.anchor)
        sm.charLeft()
        
    def _do_cmd_right(self):
        sm = self.scimoz
        sm.currentPos = max(sm.currentPos, sm.anchor)
        sm.charRight()

    def _do_cmd_wordLeft(self):
        sm = self.scimoz
        sm.currentPos = min(sm.currentPos, sm.anchor)
        sm.wordLeft()
        
    def _do_cmd_wordRight(self):
        sm = self.scimoz
        sm.currentPos = max(sm.currentPos, sm.anchor)
        sm.wordRight()
        
    def _do_cmd_wordPartLeft(self):
        sm = self.scimoz
        sm.currentPos = min(sm.currentPos, sm.anchor)
        sm.wordPartLeft()
        
    def _do_cmd_wordPartRight(self):
        sm = self.scimoz
        sm.currentPos = max(sm.currentPos, sm.anchor)
        sm.wordPartRight()

    def _do_cmd_home(self):
        sm = self.scimoz
        sm.currentPos = min(sm.currentPos, sm.anchor)
        sm.vCHome()  # XXX should make this a pref.
        
    def _do_cmd_end(self):
        sm = self.scimoz
        sm.currentPos = max(sm.currentPos, sm.anchor)
        sm.lineEnd()

charClass = {}
for x in string.letters + string.digits + '_':
    charClass[x] = 'alpha'
for x in string.whitespace:
    charClass[x] = 'whitespace'
