# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

# Scintilla event handler
from xpcom import components, _xpcom
from xpcom.file import LocalFileToURL
from koLogger import logWarning
import traceback
import sys
import code
import urllib, urlparse, os
import threading, time
import re

def MakePoint(pos):
    return pos >> 16, pos & 0xFF

class ScintillaFile:
    def __init__(self, scintilla):
        self.scintilla = scintilla
    def write(self, text):
        self.scintilla.selectionStart = 0
        self.scintilla.selectionEnd = self.scintilla.length
        self.scintilla.replaceSel(text)


from koSciMozEvents import SciMozEvents

class TailObserver:
    _com_interfaces_ = [components.interfaces.koITailObserver]
    _reg_desc_ = "Observer for Tail Results"
    _reg_contractid_ = "@activestate.com/koTailObserver;1"
    _reg_clsid_ = "{5D1E04BB-5E00-4ee8-A5BC-13E272338626}"

    def init(self, scin):
        self._scin = scin

    def notify(self, text):
        # hide the selection and do a replacement so the scrollbar
        # doesn't jump around.  Ideally, we would only tag on the new
        # text.

        # Portably replace line endings with the native format.  This
        # is done so that text copied to the clipboard will have the
        # expected format, but will interfere with View EOL Markers
        # once that is implemented for file watching.        
        nativeText = re.sub("\r\n|\r|\n", os.linesep, text)
        
        self._scin.hideSelection(0)
        self._scin.selectionStart = 0
        self._scin.selectionEnd = self._scin.length
        self._scin.readOnly = 0
        self._scin.replaceSel(nativeText)
        self._scin.readOnly = 1
        self._scin.selectionStart = self._scin.length
        self._scin.selectionEnd = self._scin.length
        self._scin.gotoPos(self._scin.length)
        self._scin.hideSelection(1)

class Tail(SciMozEvents):
    _com_interfaces_ = [components.interfaces.ISciMozEvents,
                        components.interfaces.koITail]
    _reg_desc_ = "Tail Dash Eff"
    _reg_contractid_ = "@activestate.com/koTail;1"
    _reg_clsid_ = "{21996CE3-A454-49e8-8751-FE4993861221}"

    def init(self, scintilla, url):
        self._fileObj = None
        self._do_init(scintilla, url)

    def initFile(self, scintilla, fileObj):
        self._fileObj = fileObj
        url = LocalFileToURL(fileObj.file.path).spec
        self._do_init(scintilla, url)

    def _do_init(self, scintilla, url):
        scintilla.readOnly = 0
        scintilla.text = ''
        scintilla.setMarginTypeN (0, scintilla.SC_MARGIN_NUMBER);
        scintilla.setMarginWidthN (0, 50);
        scintilla.setMarginTypeN (1, 0);
        scintilla.setMarginWidthN (1, 0);
        scintilla.setMarginTypeN (2, 0);
        scintilla.setMarginWidthN (2, 0);
        scintilla.readOnly = 1
        self.url = url

        observer = components.classes["@activestate.com/koTailObserver;1"].createInstance(components.interfaces.koITailObserver)

        scintilla = scintilla.QueryInterface(components.interfaces.ISciMozLite)
        # Create a proxy so we don't call scintilla on the wrong thread.
        # As we create the proxy on this main thread, we must pass "PROXY_ALWAYS",
        # otherwise it will decide we really dont need a proxy at all.
        scintilla = _xpcom.getProxyForObject(None, components.interfaces.ISciMozLite,
                                                     scintilla, _xpcom.PROXY_SYNC | _xpcom.PROXY_ALWAYS)
        observer.init(scintilla)

        self._observer = observer

        # This is scaring me.  We use this object to convert one of
        # Mozilla's funky file:///c/ URIs to a path (Python wants a file:/C|/ URL)
        self.hacky_url_ob = components.classes["@mozilla.org/network/standard-url;1"] \
          .createInstance(components.interfaces.nsIFileURL)

        # start up the monitoring thread
        self._die_count = None
        self._thread = threading.Thread(target=self._run)
        self._thread.start()

        # For the terminate functions, we can NOT call 
        # -- self._thread.join()
        # or we _may_ dead-lock:
        # * Observer has notification, and it calling via
        #   the proxy.  Proxy is waiting on main thread to become
        #   free.
        # * Main UI thread calls tail.terminate.
        # If we attempt to "join()" in this situation, we
        # are in a dead-lock - observer thread waiting on
        # main thread, main thread waiting on observer.
        #
        # Not joining means the thread may live longer than we expect, and
        # indeed may end up trying to perform operations on Scintilla
        # later than we expect.
        # Whatever happens, we can not afford to dead-lock!
    def terminate(self):
        self._die_count = 0 # die asap
    def terminateSoon(self, secs):
        # Give the thread a few seconds to die, to any last-minute
        # output is grabbed.
        self._die_count = secs

    def _run(self):
        #if __debug__: print "Tail._run(): thread started"
        orig_text = ""
        while self._die_count != 0:

            # it would be nice if we did not have to read the
            # entire file each time, but we cannot use readline() or
            # we will be unable to handle the case where the file
            # is truncated.

            scheme = urlparse.urlparse(self.url)[0]

            # Ahh - screw this - can't create URIFile objects on other threads :-(
            # And Python doesnt like Mozilla file:// URIs :-( :-(
            if scheme == "file": #  or scheme == "chrome":
                # infile = URIFile(self.url)
                self.hacky_url_ob.spec = self.url
                infile = open(self.hacky_url_ob.file.path)
            else:
                infile = urllib.urlopen(self.url)

            data = infile.read()  
            if orig_text != data:
                self._observer.notify(data)
                orig_text = data
                
            time.sleep(1)
            if self._die_count: # neither None or zero.
                self._die_count -= 1
        # And drop everything!
        self._observer = None
        self._fileObj = None # If a koTempFile, must remove reference.
        #if __debug__: print "Tail._run(): thread stopped"
 
