#!python
# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

from xpcom import components, nsError, ServerException
from koLogger import logDebug, logError, logWarning
from koLintResult import *
from koLintResults import koLintResults
from xpcom.server.enumerator import *
import os, sys, re, string
import process
import koprocessutils

endWordSet = ";} \t\n"

class KoTclCompileLinter:
    _com_interfaces_ = [components.interfaces.koILinter]
    _reg_desc_ = "Komodo Tcl Linter"
    _reg_clsid_ = "{842EC3B2-B79B-44a2-9417-04CEA5C02BF1}"
    _reg_contractid_ = "@activestate.com/koTclLinter;1"

    def __init__(self):
        self._linterArgvCache = None
        self._tclProRe = None

        self.sysUtils = components.classes["@activestate.com/koSysUtils;1"].\
            getService(components.interfaces.koISysUtils)
        self.prefService = components.classes["@activestate.com/koPrefService;1"].\
                getService(components.interfaces.koIPrefService)
        self.infoSvc = components.classes["@activestate.com/koInfoService;1"].\
                       getService(components.interfaces.koIInfoService)

        koDirs = components.classes["@activestate.com/koDirs;1"].\
                 getService(components.interfaces.koIDirs)
        self.internalLinter = os.path.join(koDirs.installDir, "tcl",
                                           "komodo_tcl_lint.tcl")

        languageService = components.classes[\
                "@activestate.com/koLanguage?language=Tcl;1"].\
                getService(components.interfaces.koILanguage)

        self.interp = languageService.getLanguageService(components.interfaces.koIInterpreterLanguageService)

        self.tclInfoEx = components.classes["@activestate.com/koTclProInfoEx;1"] \
                 .getService(components.interfaces.koITclInfoEx)
        self.goodInstallation = 0 # Indicates if a good (present and licensed) installation has been selected in
                                  # self.tclInfoEx

        
    def _getLinterArgv(self):
        if self._linterArgvCache:
            return self._linterArgvCache

        # 'goodInstallation' takes advantage of the fact that the
        # linter will not change.
        
        linterArgv = []
        # XXX In the future we might want to allow usage of a
        #     'tclDefaultLinter' preference so that users can specify their
        #     own linters independent of the chosen tcl shell.
        pref = "tclshDefaultInterpreter"
        tclsh = None
        if self.prefService.prefs.hasStringPref(pref) and \
           self.prefService.prefs.getStringPref(pref):
            tclsh = self.prefService.prefs.getStringPref(pref)
            if not self.goodInstallation:
                self.tclInfoEx.installationPath = self.tclInfoEx.getInstallationPathFromBinary(tclsh)
        else:
            if self.goodInstallation:
                tclsh = self.tclInfoEx.tclsh_path
            elif self.tclInfoEx.selectDefault():
                tclsh = self.tclInfoEx.tclsh_path

        if not tclsh:
            raise ServerException, \
                  (nsError.NS_ERROR_NOT_AVAILABLE,
                   "Could not find a suitable Tcl executable for linting.")
        else:
            linterArgv.append(tclsh)
            if self.goodInstallation or self.tclInfoEx.haveLicense:
                self.goodInstallation = 1
            else:
                raise ServerException, (nsError.NS_ERROR_ILLEGAL_VALUE,
                                        "License required for Tcl linting")
        
        if linterArgv[0].find('procheck') != -1:
            # This is specific to procheck and derivatives
            linterArgv.append("-onepass")
        else:
            # Assume that it's a tclsh that requires komodo_tcl_lint.tcl
            if os.path.isfile(self.internalLinter):
                linterArgv += [self.internalLinter, "-onepass"]
            else:
                logError("Unable to find internal Komodo Tcl linter")
                return None

        #logDebug("found Tcl linter '%s'" % (linterArgv))
        # Store so we only do this lookup once
        self._linterArgvCache = linterArgv
        return linterArgv

    def lint(self, text, encoding, cwd):
        argv = self._getLinterArgv() # This will also do license checking
        env = koprocessutils.getUserEnv()
        cwd = cwd or None
        p = process.ProcessOpen(argv, cwd=cwd, env=env, mode='b')
        p.stdin.write(text)
        p.stdin.close()
        lOutput = p.stdout.readlines()
        p.close()
	#print '======'
	#print lOutput
	#print '<<<<<<'

        if lOutput[0].startswith("TclPro") or \
           lOutput[0].startswith("scanning"):
            # The open source TclPro has the header stripped,
            # so we need to check for both "TclPro" and "scanning"
            lInput = text.split('\n')
            return self._TclPro_lint(lInput, lOutput)
        else:
            logError("unrecognized lint output format:\n%s" % lOutput[0])
            raise ServerException, \
                  (nsError.NS_ERROR_NOT_AVAILABLE,
                   "unrecognized lint output format.")

    def _TclPro_lint(self, lInput, lOutput):
        if self._tclProRe == None:
            self._tclProRe = re.compile(r'^[^:]+:(\d+) \((\w+)\) (.*)$')
        # TclPro output has the following header info:
        ###################
        #TclPro -- Version 1.4.1
        #Copyright (C) Ajuba Solutions 1998-2001. All rights reserved.
        #This product is registered to: <user name>
        #
        #scanning: <filename>
        #checking: <filename>
        #<filename>:<lineNo> (<errorName>) <errorText>
        #<commandWithError>
        #<caret at col in command where error is>
        #[repeat above three until no more errors]
        ###################
        # The 'checking' line exists when the -onepass option isn't used
        i = 1
        lastCol = -1
        lastLine = -1
        results = koLintResults()
        numLines = len(lOutput)
        while i < numLines:
            match = re.search(self._tclProRe, lOutput[i])
            if not match:
                # We expect the first few lines to not match, while we
                # skip the header lines
                i += 1
                # But if too many lines aren't matching, then we must
                # unrecognizable input
                if i > 8:
                    logWarning("couldn't recognized Tcl linter output format")
                    break
                continue
            if i+2 >= numLines:
                logError("unexpected return data format: short data")
                return                    
            cmd = lOutput[i+1].rstrip()
            col = lOutput[i+2].find('^')
            if col == -1:
                logError("unexpected return data format: no caret")
                return
            lineNo = match.group(1)
            # lInput is zero-based, whereas text display lines start at one,
            # so we adjust lineNo down one here.
            lineNo = int(lineNo) - 1
            if lineNo != lastLine:
                lastLine = lineNo
                lastCol = 0

            startCol = lInput[lineNo].find(cmd, lastCol)
            if startCol == -1:
                # It's possible that we are throwing multiple warnings on
                # the same command.  Let's check again, starting at col 0.
                # If something was found, ignore multiple warnings in the
                # same line, otherwise log a problem
                if lastCol == 0 or \
                   lInput[lineNo].find(cmd, 0) == -1:
                    # Looks like we can't figure something out...
                    logDebug('undecipherable lint output: %s,%s "%s" in "%s" last %s' \
                             % (lineNo+1, col, cmd, lInput[lineNo], lastCol))
                i += 3
                continue

            #print 'TclLint: %s,%s "%s" in "%s"' \
            #      % (lineNo+1, startCol, cmd, lInput[lineNo])

            result = KoLintResult()
            msg = match.group(2)
            result.description = match.group(3).strip()
            result.lineStart = lineNo+1
            # All our warnings stay on one line
            result.lineEnd = result.lineStart
            cmdLen = len(cmd)
            if msg.startswith("warn") or msg.startswith("nonPort"):
                result.severity = result.SEV_WARNING
                # In the warning case, go to the next char that we think
                # ends the word (not precise), starting at the caret.
                endCol = col
                while endCol < cmdLen and cmd[endCol] not in endWordSet:
                    # might want to handle \ escapes
                    endCol += 1
                # readjust for where the command started
                endCol += startCol
                # and bump up the startCol to where the caret indicated
                startCol += col
            else:
                result.severity = result.SEV_ERROR
                # In the error case, highlight the entire command
                # The linter should only give one error per command
                endCol = startCol + cmdLen
            # Adjust by 1 as columns start at 1 for the text display
            result.columnStart = startCol + 1
            result.columnEnd = endCol + 1
            results.addResult(result)

            # Set lastCol to properly handle multiple similar code-blocks
            # with errors on the same line.
            lastCol = startCol+1
            i += 3
            # end while
        return results

