# Copyright (c) 2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.

import os
import sys
from xpcom import components
from koLogger import logError
import urllib
import shutil

userPersonalDirectoryName = "My Templates"

def resolvePath(path):
    "Resolve a path through symlinks and Windows shortcuts"
    if sys.platform.startswith('win') and (os.path.splitext(path)[1]== '.lnk'):
        from win32com.shell import shell, shellcon
        import pythoncom
        from win32com.storagecon import STGM_READ

        try:
            shellLink = pythoncom.CoCreateInstance(shell.CLSID_ShellLink, None, pythoncom.CLSCTX_INPROC_SERVER, shell.IID_IShellLink)
            persistFile = shellLink.QueryInterface(pythoncom.IID_IPersistFile)
            persistFile.Load(path, STGM_READ)
            shellLink.Resolve(0, shell.SLR_ANY_MATCH | shell.SLR_NO_UI)
            fname, findData = shellLink.GetPath(0)
            return shellLink.GetPath(shell.SLGP_UNCPRIORITY)[0]
        except Exception, e:
            logError('exception thrown when trying to resolve shortcut:' + str(e))
            raise e
    elif os.path.islink(path):        
        linkpath = os.readlink(path)
        if os.path.isabs(linkpath):
            return linkpath
        else:
            return os.path.join(os.path.dirname(path), linkpath)

    return path

def cmp_TemplateNames(name1, name2):
    if name1 == name2 == userPersonalDirectoryName: # User's personal directory should come first
        return 0
    elif name1 == userPersonalDirectoryName:
        return -1
    elif name2 == userPersonalDirectoryName:
        return 1
    
    simpleCmp = cmp(name1.lower(), name2.lower())
    if simpleCmp == 0:
        if sys.platform.startswith('win'):
            return 0            
        else:
            return cmp(name1, name2)
    else:
        return simpleCmp

class koFileTemplate:
    _com_interfaces_ = [components.interfaces.koIHierarchyItem, components.interfaces.koIFileTemplate]
    _reg_contractid_ = "@activestate.com/koFileTemplate;1"
    _reg_clsid_ = "{b334c0e2-76c8-4f5d-a7d6-a96dcf5203e7}"

    def __init__(self):
        self.container = 0
        self.available_types = components.interfaces.koIHierarchyItem.ITEM_HAS_STRING
        
    def set_fileName(self, file):
        self.fileName = file
        self.fileURI = 'file:' + urllib.pathname2url(os.path.abspath(self.fileName))
        self.name = os.path.basename(self.fileName)
        if self.name == '':
            self.name = self.fileName

        if os.path.splitext(self.name)[0]:
            self.name = os.path.splitext(self.name)[0]
        
    def __repr__(self):
        return '<koFileTemplate fileName="%s" name="%s">' % (self.fileName, self.name)

    def repr(self):
        return repr(self)

    def get_item_string(self):
        return self.fileName

    def get_language(self):
        assert 0, 'Trying to get language of file template'

# This class is no longer used. It was originally used to insert a hierarchy of blank
# language templates but we now have an actual empty file for each language
                
class koBlankLanguageTemplate:
    _com_interfaces_ = [components.interfaces.koIHierarchyItem, components.interfaces.koIBlankLanguageTemplate]
    _reg_contractid_ = "@activestate.com/koBlankLanguageTemplate;1"
    _reg_clsid_ = "{83d9e7f2-7ec3-404f-b489-fdd1ab47df01}"
 
    def __init__(self):
        self.container = 0        

    def set_language(self, language):
        self.language = language
        self.name = language

    def __repr__(self):
        return '<koBlankLanguageTemplate language="%s" name="%s">' % (self.language, self.name)

    def repr(self):
        return repr(self)
    
class koTemplateDirectoryContainer:
    _com_interfaces_ = [components.interfaces.koIHierarchyItem]
    _reg_contractid_ = "@activestate.com/koTemplateDirectoryContainer;1"
    _reg_clsid_ = "{60b21b08-d1e2-4bbd-b72d-06fc7abd2b64}"
    
    def __init__(self, directory, maxDepth = -1):
        self.directory = directory
        self.container = 1
        self.available_types = components.interfaces.koIHierarchyItem.ITEM_HAS_STRING
        self.name = os.path.basename(self.directory)
        if self.name == '':
            self.name = self.directory

        self.maxDepth = maxDepth

    def __repr__(self):
        return '<koTemplateDirectoryContainer directory="%s" name="%s" maxDepth="%d">' % (self.directory, self.name, self.maxDepth)
    
    def getChildren(self):
        children = []
        try:
            fileList = os.listdir(self.directory)
        except Exception, e: # e.g. missing removable media or user directory
            pass
        else:
            for child in fileList:
                originalPath = os.path.join(self.directory, child)
                try:
                    fullpath = resolvePath(originalPath)
                except Exception, e:
                    logError('could not resolve "%s" [%s], skipping file' % (originalPath,e))
                else:
                    dir = os.path.isdir(fullpath)
                    if dir:
                        if self.maxDepth != 0:
                            children.append(koTemplateDirectoryContainer(fullpath, self.maxDepth - 1))
                    else:
                        template = koFileTemplate()
                        template.set_fileName(fullpath)
                        children.append(template)

        children.sort(lambda x,y: cmp_TemplateNames(x.name, y.name))
        return children

class koTemplateSimpleContainer:
    _com_interfaces_ = [components.interfaces.koIHierarchyItem]
    _reg_contractid_ = "@activestate.com/koTemplateSimpleContainer;1"
    _reg_clsid_ = "{c3ebb896-dec9-431c-8bc5-214388af8776}"
    
    def __init__(self, name, children):
        self.name = name
        self.children = children
        self.container = 1

    def getChildren(self):
        return self.children

class koTemplateMergedContainer:
    _com_interfaces_ = [components.interfaces.koIHierarchyItem]
    _reg_contractid_ = "@activestate.com/koTemplateMergedContainer;1"
    _reg_clsid_ = "{07ab1858-5f81-4319-acbf-f3acfaab1a3e}"

    def __init__(self, userContainer, defaultContainer, maxDepth):
        self.userContainer = userContainer
        self.defaultContainer = defaultContainer
        self.name = userContainer.name
        self.maxDepth = maxDepth
        self.container = 1

        userContainer.maxDepth = self.maxDepth
        defaultContainer.maxDepth = self.maxDepth

    def __repr__(self):
        return '<koTemplateMergedContainer maxDepth="%d">' % self.maxDepth
        
    def getChildren(self):
        def makeUserChildFromDefaultChild(userContainer, defaultChild):
            userDirectory = os.path.join(userContainer.directory, defaultChild.name)
            return koTemplateDirectoryContainer(userDirectory)
        
        userChildren = self.userContainer.getChildren()
        defaultChildren = self.defaultContainer.getChildren()

        u = 0
        d = 0
        children = []
        # Merge the two lists
        while u < len(userChildren) and d < len(defaultChildren):
            uChild = userChildren[u]
            dChild = defaultChildren[d]
            
            cmpNames = cmp_TemplateNames(uChild.name, dChild.name)        
            if cmpNames == -1:
                children.append(uChild)
                u += 1
            elif cmpNames == 0:
                if uChild.container:
                    if dChild.container:
                        children.append(koTemplateMergedContainer(uChild, dChild, self.maxDepth - 1))
                    else:
                        children.append(uChild)
                u += 1
                d += 1
            else:
                # Need to merge here
                if dChild.container:
                    children.append(koTemplateMergedContainer(makeUserChildFromDefaultChild(self.userContainer, dChild),
                                                              dChild, self.maxDepth - 1))
                else:                    
                    children.append(dChild)
                d += 1

        if u < len(userChildren):
            children += userChildren[u:]
        else:
            for dChild in defaultChildren[d:]:
                if dChild.container:
                    children.append(koTemplateMergedContainer(makeUserChildFromDefaultChild(self.userContainer, dChild),
                                                              dChild, self.maxDepth - 1))
                else:
                    children.append(dChild)

        return children

def toTemplateHierarchy(languageHierarchy):
    if languageHierarchy.container:
        children = [toTemplateHierarchy(i) for i in languageHierarchy.getChildren()]
        return koTemplateSimpleContainer(languageHierarchy.name, children)
    else:
        template = koBlankLanguageTemplate()
        template.set_language(languageHierarchy.item_string)
        return template

# This class is no longer used. It was originally used to insert a hiearchy of blank
# language templates but we now have an actual empty file for each language

class koTemplateRoot(koTemplateMergedContainer):
    _com_interfaces_ = [components.interfaces.koIHierarchyItem]
    _reg_contractid_ = "@activestate.com/koTemplateRoot;1"
    _reg_clsid_ = "{69390984-2d51-4416-9f62-c15b237a990f}"
        
    def __init__(self, userDirectory, defaultDirectory, maxDepth):
        koTemplateMergedContainer.__init__(self, userDirectory, defaultDirectory, maxDepth)
        self.languageRegistryService = components.classes["@activestate.com/koLanguageRegistryService;1"].\
                                  getService()
        self.blanks = toTemplateHierarchy(self.languageRegistryService.getLanguageHierarchy())
        self.blanks.name = "Blank"
    
    def getChildren(self):
        children = [self.blanks] + koTemplateMergedContainer.getChildren(self)
        return children

    
class koTemplateManager:
    _com_interfaces_ = [components.interfaces.koITemplateManager]
    _reg_contractid_ = "@activestate.com/koTemplateManager;1"
    _reg_clsid_ = "{60476817-f0b7-4b72-8d11-e676ee4e3eb0}"

    def __init__(self):
        self.koDirs = components.classes["@activestate.com/koDirs;1"].getService()

    def init(self):
        userDirectory = self.getUserTemplateDirectory()
        userPersonalDirectory = os.path.join(self.getUserTemplateDirectory(), userPersonalDirectoryName)
        defaultDirectory = self.getDefaultTemplateDirectory()
        destReadme = os.path.join(userDirectory, 'readme.txt')
        srcReadme = os.path.join(defaultDirectory, 'readme.txt')
        
        if not os.path.exists(destReadme):
            try:
                shutil.copyfile(srcReadme, destReadme)
            except Exception, e:
                logDebug("Could not copy readme file to user template directory: %r [%r]" % (e, e.args))

        if not os.path.exists(userPersonalDirectory):
            # Only add the user's personal directory if they haven't already been mucking about

            contents = os.listdir(userDirectory)
            ourFiles = ['readme.txt']
            for ourFile in ourFiles:
                try:
                    contents.remove(ourFile)
                except ValueError:            
                    pass

            if len(contents) == 0: # Looks like only our files are here
                try:
                    os.mkdir(userPersonalDirectory)
                except Exception, e:
                    logDebug("Could not create the personal template directory: %r [%r]" % (e, e.args))

    def getUserTemplateDirectory(self):
        templateDir = os.path.join(self.koDirs.userDataDir, 'templates')
        if os.path.exists(templateDir):
            if not os.path.isdir(templateDir):
                raise 'expected "%s" to be a directory' % templateDir
        else:
            os.mkdir(templateDir)
        return templateDir
    
    def getDefaultTemplateDirectory(self):
        return os.path.join(self.koDirs.installDir,'default-templates')
        
    
    def getTemplateHierarchy(self):
        userDirectory = koTemplateDirectoryContainer(self.getUserTemplateDirectory())
        defaultDirectory = koTemplateDirectoryContainer(self.getDefaultTemplateDirectory())
        
        return koTemplateMergedContainer(userDirectory, defaultDirectory, 5)
