#!python
# Copyright (c) 2000-2001 ActiveState Tool Corporation.
# See the file LICENSE.txt for licensing information.


from xpcom import components, nsError, ServerException
from koLintResult import *
from koLintResults import koLintResults
from koLogger import logError, logWarning, logDebug
from xpcom.server.enumerator import *
import os, sys, re
import tempfile
import process


class KoXPCShellLinter:
    _com_interfaces_ = [components.interfaces.koILinter]
    _reg_desc_ = "Komodo XPCShell JavaScript Linter"
    _reg_clsid_ = "{111FBEA1-7CA3-4858-B040-E51CF5A20CE9}"
    _reg_contractid_ = "@activestate.com/koXPCShellLinter;1"

    def __init__(self):
        self.infoSvc = components.classes["@activestate.com/koInfoService;1"].\
            getService(components.interfaces.koIInfoService)
        self.isDebugBuild = self.infoSvc.buildType == "debug"

    def lint(self, text, encoding, cwd):
        # copy file-to-lint to a temp file
        jsfilename = tempfile.mktemp() + '.js'
        # convert to UNIX line terminators before splitting
        datalines = re.sub("\r\n|\r", "\n", text).split("\n")
        # escape the jscode for inclusion in a JS string
        jscode = text
        jscode = jscode.replace('\\', '\\\\')
        jscode = jscode.replace('\n', '\\n')
        jscode = jscode.replace('\r', '\\r')
        jscode = jscode.replace('"', '\\"')
        fout = open(jsfilename, 'w')
        fout.write("Script(\""+ jscode + "\").compile();")
        fout.close()

        # lint the temp file
        cmd = 'xpcshell -w -s %s' % jsfilename
        cwd = cwd or None
        p = process.ProcessOpen(cmd, cwd=cwd)
        warnLines = p.stderr.readlines()
        p.close()
        os.unlink(jsfilename)

        # strip off some debugging output from 'xpcshell' for debug builds
        if self.isDebugBuild:
            # Expect to see this:
            #   Adding parser accelerators ...
            #   Done.
            #   nNCL: registering deferred (0)
            #   <useful lines ...>
            #   CanUnload_enumerate: skipping native
            if warnLines[0].startswith("Adding parser accelerators ...") and\
               warnLines[1].startswith("Done.") and\
               warnLines[2].startswith("nNCL: registering deferred"):
                warnLines = warnLines[3:]
            else:
                msg = "unexpected output when parsing JS syntax check "\
                    "output: %s\n" % warnLines
                logError(msg)
                raise ServerException, (nsError.NS_ERROR_UNEXPECTED, msg)

        # 'xpcshell' error reports come in 4 line chunks that look like
        # this:
        #    <filename>:8: SyntaxError: missing ; before statement:
        #    <filename>:8: ar asdf = 1;
        #
        #    <filename>:8: ...^
        # There is one exception: if the file is only one line then
        # the third blank line is not there. THerefore we will strip
        # empty lines and parse 3 line chunks.
        strippedWarnLines = [line for line in warnLines if line.strip()]

        # Parse out the xpcshell lint results
        results = koLintResults()
        counter = 0  # count index in 3 line groups
        firstLineRe = re.compile("^%s:(?P<lineNo>\d+): (?P<desc>.*?):?\s*$" %\
            re.escape(jsfilename))
        lastLineRe = re.compile("^%s:(?P<lineNo>\d+): (?P<dots>\.*)\^$" %\
            re.escape(jsfilename))
        for line in strippedWarnLines:
            if counter == 0 and line.startswith(jsfilename):
                # first line: get the error description and line number
                firstLineMatch = firstLineRe.search(line.strip())
                if firstLineMatch:
                    lineNo = int(firstLineMatch.group("lineNo"))
                    desc = firstLineMatch.group("desc")
                else:
                    msg = "Unexpected output when parsing JS syntax check "\
                        "output: '%s'\n" % line
                    logError(msg)
                    raise ServerException, (nsError.NS_ERROR_UNEXPECTED, msg)
            elif counter == 2:
                # get the column of the error
                lastLineMatch = lastLineRe.search(line.strip())
                if lastLineMatch:
                    numDots = len(lastLineMatch.group("dots"))
                else:
                    msg = "Unexpected output when parsing JS syntax check "\
                        "output: '%s'\n" % line
                    logError(msg)
                    raise ServerException, (nsError.NS_ERROR_UNEXPECTED, msg)
                # build lint result object
                result = KoLintResult()
                result.description = desc
                if lineNo >= len(datalines):
                    # if the error is on the last line, work back to the last
                    # character of the first nonblank line so we can display
                    # the error somewhere
                    while len(datalines[lineNo - 1]) == 0:
                        lineNo -= 1
                    result.columnEnd = len(datalines[lineNo - 1])
                    result.columnStart = result.columnEnd - 1
                else:
                    result.columnStart = numDots + 1
                    result.columnEnd = result.columnStart + 1
                result.lineStart = lineNo
                result.lineEnd = lineNo
                result.severity = result.SEV_ERROR
                # This always results in a lint result spanning a single
                # character, which, given the squiggly reporting scheme is
                # almost invisible. Workaround: set the result to be the
                # whole line and append the column number to the description.
                result.description += " (on column %d)" % result.columnStart
                result.columnStart = 1
                result.columnEnd = len(datalines[lineNo-1])+1
                results.addResult(result)
            counter = (counter + 1) % 3

        return results


