
from xpcom import components, ServerException, COMException, nsError
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC
from koLogger import logError, logDebug, logWarning
import os
import threading

class PrefException(Exception):
    def __init__(self, prefid, msg):
        self.prefid = prefid
        self.msg = msg
        Exception.__init__(self)

# Utility for auto-registering an invocation class.
def RegisterAsInvocation(klass, compMgr, location, registryLocation, componentType):
    registry, key = _GetInvocationsRegistryAndKey()
    key = registry.addSubtree(key, klass._invocation_name_)
    registry.setString(key, "Description", klass._invocation_description_)
    registry.setString(key, "ContractID", klass._reg_contractid_)
    registry.setString(key, "chrome_category", klass._invocation_chrome_category_)
    registry.setString(key, "chrome_instance", klass._invocation_chrome_instance_)
    #logDebug("  class %s also registered as invocation '%s'" % (klass.__name__, klass._invocation_name_))

def _GetInvocationsRegistryAndKey():    
    registry = components.classes["@mozilla.org/registry;1"].getService(components.interfaces.nsIRegistry)
    key = registry.addSubtree(registry.Common, "software/ActiveState/Komodo/Invocations")
    return registry, key

def checkFileExists(prefset, pref, msg_if_empty = None):
    fname = prefset.getStringPref(pref)
    if msg_if_empty and not fname:
        raise PrefException(pref, msg_if_empty)
    if not os.path.isfile(fname):
        message = "File '%s' does exist" % fname
        raise PrefException(pref, message)

def environmentStringsToDict(existing):
    ret = {}
    for i in range(len(existing)):
        try:
            key, value = existing[i].split("=", 1)
            ret[key] = value;
        except ValueError:
            logDebug("error on value ",repr(existing[i]))
            pass
    return ret

def environmentDictToStrings(env_dict):
    return ["%s=%s"  % (item[0], item[1]) for item in env_dict.items()]

def _safewrite(file, data):
    if file is not None:
        try:
            file.write(data)
        except COMException, details:
            logError("Writing to file '%s' failed: %s" % (file, details))
    
class ProcessIOHandler:
    _com_interfaces_ = "koIProcessHandlerEventSink"
    def __init__(self, stdin, stdout, stderr, process_spawner):
        self.terminate_event = threading.Event()
        self.stdout = stdout
        self.stderr = stderr
        self.stdin = stdin
        self.process_spawner = process_spawner
        if stdin:
            threading.Thread(target=self._localInputThread).start()
        proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"].\
            getService(components.interfaces.nsIProxyObjectManager)
        observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._observerSvc = proxyMgr.getProxyForObject(None,
            components.interfaces.nsIObserverService,
            observerSvc, PROXY_ALWAYS | PROXY_SYNC)

    def relayOutput( self, output, size, output_type):
        # Result: void (None)
        # In: param0: string
        # In: param1: int32
        # In: param2: int32
        #print "Relay output", `output`
        if output_type == 1:
            _safewrite(self.stdout, output)
        else:
            _safewrite(self.stderr, output)
        #print "Relay output end", `output`

    def signalProcessEnd( self, exitCode ):
        # Result: void (None)
        # Tell the command-loop to break out with a stop request.
        # This will also set the state was we come out.

        try:
            #XXX Should eventually pass exitCode out with notification
            self._observerSvc.Notify(None, "invocation_terminated", None)
        except COMException, e:
            # do nothing if nobody listening
            pass
        # get the input thread to quit properly
        self.terminate_event.set()
        _safewrite(self.stdin,'')

    def _localInputThread(self):
        while 1:
            data = self.stdin.read(1024)
            if not data and self.terminate_event.isSet():
                #logDebug("Invocation input termination set")
                break
            if data:
                try:
                    self.process_spawner.sendInputLine(data, len(data))
                except COMException, details:
                    # process has quit
                    #logDebug("Invocation input caught exeption sending data")
                    break
        try:
            self.process_spawner.send_EOI()
        except COMException, details:
            pass
        #logDebug("Invocation input thread dieing")
