#!/usr/bin/env python

# Copyright (c) 2002 Trent Mick
# 
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
    Show full path of commands.

    Usage:
        which [<options>...] [<command-name>...]

    Options:
        -h, --help      Print this help and exit.
        -V, --version   Print the version info and exit.

        -a, --all       Print *all* matching paths.
        -v, --verbose   Print out near misses on stderr. [default]
        -q, --quiet     Do not print out near misses.

        -p <altpath>, --path=<altpath>
                        An alternative path (list of directories) may
                        be specified for searching.

    Show the full path to the program that would be run for each given
    command name, if any. Which, like GNU's which, returns the number of
    failed arguments, or -1 when no <command-name> was given.

    Near misses include duplicates, non-regular files and (on Un*x)
    files without executable access.
"""
#REQUIREMENTS:
#   - Python >= 2.2 (because of the use of generators)
#

from __future__ import generators
import os
import sys
import getopt
import stat


#---- exceptions

class WhichError(Exception):
    pass


#---- global data

_version_ = (0, 7, 0)
gVerbose = 0


#---- internal support stuff

def _getRegisteredExecutable(exeName):
    """Windows allow application paths to be registered in the registry."""
    registered = None
    if sys.platform.startswith('win'):
        if os.path.splitext(exeName)[1].lower() != '.exe':
            exeName += '.exe'
        import _winreg
        try:
            registered = _winreg.QueryValue(_winreg.HKEY_LOCAL_MACHINE,
                "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\"\
                + exeName)
        except _winreg.error:
            pass
        if registered and not os.path.exists(registered):
            registered = None
    return None

def _samefile(fname1, fname2):
    if sys.platform.startswith('win'):
        return ( os.path.normpath(os.path.normcase(fname1)) ==\
            os.path.normpath(os.path.normcase(fname2)) )
    else:
        return os.path.samefile(fname1, fname2)

def _cull(potential, matches):
    """Cull inappropriate matches. Possible reasons:
        - a duplicate of a previous match
        - not a disk file
        - not executable (non-Windows)
    If 'potential' is approved it is returned and added to 'matches'.
    Otherwise, None is returned.
    """
    global gVerbose
    for match in matches:  # don't yield duplicates
        if _samefile(potential, match):
            if gVerbose:
                sys.stderr.write("duplicate: %s\n" % potential)
            return None
    else:
        if not stat.S_ISREG(os.stat(potential).st_mode):
            if gVerbose:
                sys.stderr.write("not a regular file: %s\n" % potential)
        elif not os.access(potential, os.X_OK):
            if gVerbose:
                sys.stderr.write("no executable access: %s\n" % potential)
        else:
            matches.append(potential)
            return potential

        
#---- module API

def which(command, path=None):
    """Generate a list of found apps for the given 'command' on the path.
    
    "path" (optional) an alternate search path, instead of using the PATH
        environment variable. This is a list, e.g. ["C:\\TEMP",...].
    """
    matches = []
    if path is None:
        path = os.environ.get("PATH", "").split(os.pathsep)
        if sys.platform.startswith("win"):
            path.insert(0, os.curdir)  # implied by Windows shell
    # Windows has the concept of a list of extensions (PATHEXT env var).
    exts = []
    if sys.platform.startswith("win"):
        exts = os.environ.get("PATHEXT", "").split(os.pathsep)
        # If '.exe' is not in exts then obviously this is Win9x and or a
        # bogus PATHEXT, then use a reasonable default.
        for ext in exts:
            if ext.lower() == ".exe":
                break
        else:
            exts = ['.COM', '.EXE', '.BAT']
    # File name cannot have path separators because PATH lookup does not
    # work that way.
    if os.sep in command or os.altsep and os.altsep in command:
        pass
    else:
        for dirName in path:
            # On windows the dirName *could* be quoted, drop the quotes
            if sys.platform.startswith("win") and len(dirName) >= 2\
               and dirName[0] == '"' and dirName[-1] == '"':
                dirName = dirName[1:-1]
            for ext in ['']+exts:
                absName = os.path.abspath(
                    os.path.normpath(os.path.join(dirName, command+ext)))
                if os.path.isfile(absName):
                    match = _cull(absName, matches)
                    if match:
                        yield match
        absName = _getRegisteredExecutable(command)
        if absName is not None:
            match = _cull(absName, matches)
            if match:
                yield match


#---- mainline

def main(argv):
    global gVerbose, _version_
    all = 0
    gVerbose = 1 # be verbose through the cmdln interface, by default.
    try:
        optlist, args = getopt.getopt(argv[1:], 'haVvqp:',
            ['help', 'all', 'version', 'verbose', 'quiet', 'path'])
    except getopt.GetoptError, msg:
        print "which: error: %s. Your invocation was: %s\n" % (msg, argv)
        print __doc__
        return 1
    altpath = None
    for opt, optarg in optlist:
        if opt in ('-h', '--help'):
            print __doc__
            return 0
        elif opt in ('-V', '--version'):
            print "Which %s" % '.'.join([str(i) for i in _version_])
            return 0
        elif opt in ('-a', '--all'):
            all = 1
        elif opt in ('-v', '--verbose'):
            gVerbose = 1
        elif opt in ('-q', '--quiet'):
            gVerbose = 0
        elif opt in ('-p', '--path'):
            altpath = optarg.split(os.pathsep)

    if len(args) == 0:
        return -1

    failures = 0
    for arg in args:
        #print "debug: search for %r" % arg
        npaths = 0
        for path in which(arg, altpath):
            print path
            npaths += 1
            if not all:
                break
        if not npaths:
            failures += 1
    return failures


if __name__ == "__main__":
    sys.exit( main(sys.argv) )


